# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hangups/test/test_pblite.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hangups/test/test_pblite.proto',
  package='',
  syntax='proto2',
  serialized_pb=_b('\n\x1ehangups/test/test_pblite.proto\"0\n\x13TestRequiredMessage\x12\x19\n\x11test_required_int\x18\x01 \x02(\x05\"\xe8\x03\n\x0bTestMessage\x12\x10\n\x08test_int\x18\x01 \x01(\x05\x12\x19\n\x11test_repeated_int\x18\x02 \x03(\x05\x12\x13\n\x0btest_string\x18\x03 \x01(\t\x12\x1c\n\x14test_repeated_string\x18\x04 \x03(\t\x12(\n\ttest_enum\x18\x05 \x01(\x0e\x32\x15.TestMessage.TestEnum\x12\x31\n\x12test_repeated_enum\x18\x06 \x03(\x0e\x32\x15.TestMessage.TestEnum\x12;\n\x15test_embedded_message\x18\x07 \x01(\x0b\x32\x1c.TestMessage.EmbeddedMessage\x12\x44\n\x1etest_repeated_embedded_message\x18\x08 \x03(\x0b\x32\x1c.TestMessage.EmbeddedMessage\x12\x12\n\ntest_bytes\x18\t \x01(\x0c\x12\x1b\n\x13test_repeated_bytes\x18\n \x03(\x0c\x1a,\n\x0f\x45mbeddedMessage\x12\x19\n\x11test_embedded_int\x18\x01 \x01(\x05\":\n\x08TestEnum\x12\n\n\x06TEST_0\x10\x00\x12\n\n\x06TEST_1\x10\x01\x12\n\n\x06TEST_2\x10\x02\x12\n\n\x06TEST_3\x10\x03')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_TESTMESSAGE_TESTENUM = _descriptor.EnumDescriptor(
  name='TestEnum',
  full_name='TestMessage.TestEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TEST_0', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_1', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_2', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_3', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=515,
  serialized_end=573,
)
_sym_db.RegisterEnumDescriptor(_TESTMESSAGE_TESTENUM)


_TESTREQUIREDMESSAGE = _descriptor.Descriptor(
  name='TestRequiredMessage',
  full_name='TestRequiredMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='test_required_int', full_name='TestRequiredMessage.test_required_int', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=82,
)


_TESTMESSAGE_EMBEDDEDMESSAGE = _descriptor.Descriptor(
  name='EmbeddedMessage',
  full_name='TestMessage.EmbeddedMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='test_embedded_int', full_name='TestMessage.EmbeddedMessage.test_embedded_int', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=469,
  serialized_end=513,
)

_TESTMESSAGE = _descriptor.Descriptor(
  name='TestMessage',
  full_name='TestMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='test_int', full_name='TestMessage.test_int', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_repeated_int', full_name='TestMessage.test_repeated_int', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_string', full_name='TestMessage.test_string', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_repeated_string', full_name='TestMessage.test_repeated_string', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_enum', full_name='TestMessage.test_enum', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_repeated_enum', full_name='TestMessage.test_repeated_enum', index=5,
      number=6, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_embedded_message', full_name='TestMessage.test_embedded_message', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_repeated_embedded_message', full_name='TestMessage.test_repeated_embedded_message', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_bytes', full_name='TestMessage.test_bytes', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_repeated_bytes', full_name='TestMessage.test_repeated_bytes', index=9,
      number=10, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TESTMESSAGE_EMBEDDEDMESSAGE, ],
  enum_types=[
    _TESTMESSAGE_TESTENUM,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=573,
)

_TESTMESSAGE_EMBEDDEDMESSAGE.containing_type = _TESTMESSAGE
_TESTMESSAGE.fields_by_name['test_enum'].enum_type = _TESTMESSAGE_TESTENUM
_TESTMESSAGE.fields_by_name['test_repeated_enum'].enum_type = _TESTMESSAGE_TESTENUM
_TESTMESSAGE.fields_by_name['test_embedded_message'].message_type = _TESTMESSAGE_EMBEDDEDMESSAGE
_TESTMESSAGE.fields_by_name['test_repeated_embedded_message'].message_type = _TESTMESSAGE_EMBEDDEDMESSAGE
_TESTMESSAGE_TESTENUM.containing_type = _TESTMESSAGE
DESCRIPTOR.message_types_by_name['TestRequiredMessage'] = _TESTREQUIREDMESSAGE
DESCRIPTOR.message_types_by_name['TestMessage'] = _TESTMESSAGE

TestRequiredMessage = _reflection.GeneratedProtocolMessageType('TestRequiredMessage', (_message.Message,), dict(
  DESCRIPTOR = _TESTREQUIREDMESSAGE,
  __module__ = 'hangups.test.test_pblite_pb2'
  # @@protoc_insertion_point(class_scope:TestRequiredMessage)
  ))
_sym_db.RegisterMessage(TestRequiredMessage)

TestMessage = _reflection.GeneratedProtocolMessageType('TestMessage', (_message.Message,), dict(

  EmbeddedMessage = _reflection.GeneratedProtocolMessageType('EmbeddedMessage', (_message.Message,), dict(
    DESCRIPTOR = _TESTMESSAGE_EMBEDDEDMESSAGE,
    __module__ = 'hangups.test.test_pblite_pb2'
    # @@protoc_insertion_point(class_scope:TestMessage.EmbeddedMessage)
    ))
  ,
  DESCRIPTOR = _TESTMESSAGE,
  __module__ = 'hangups.test.test_pblite_pb2'
  # @@protoc_insertion_point(class_scope:TestMessage)
  ))
_sym_db.RegisterMessage(TestMessage)
_sym_db.RegisterMessage(TestMessage.EmbeddedMessage)


# @@protoc_insertion_point(module_scope)
