# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hangups/hangouts.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hangups/hangouts.proto',
  package='',
  syntax='proto2',
  serialized_pb=_b('\n\x16hangups/hangouts.proto\"\\\n\x13\x44oNotDisturbSetting\x12\x16\n\x0e\x64o_not_disturb\x18\x01 \x01(\x08\x12\x1c\n\x14\x65xpiration_timestamp\x18\x02 \x01(\x04\x12\x0f\n\x07version\x18\x03 \x01(\x04\"B\n\x14NotificationSettings\x12*\n\x0c\x64nd_settings\x18\x01 \x01(\x0b\x32\x14.DoNotDisturbSetting\"\x1c\n\x0e\x43onversationId\x12\n\n\x02id\x18\x01 \x01(\t\"1\n\rParticipantId\x12\x0f\n\x07gaia_id\x18\x01 \x01(\t\x12\x0f\n\x07\x63hat_id\x18\x02 \x01(\t\"?\n\x0c\x44\x65viceStatus\x12\x0e\n\x06mobile\x18\x01 \x01(\x08\x12\x0f\n\x07\x64\x65sktop\x18\x02 \x01(\x08\x12\x0e\n\x06tablet\x18\x03 \x01(\x08\"J\n\x08LastSeen\x12 \n\x18last_seen_timestamp_usec\x18\x01 \x01(\x04\x12\x1c\n\x14usec_since_last_seen\x18\x02 \x01(\x04\"\x98\x01\n\x08Presence\x12\x11\n\treachable\x18\x01 \x01(\x08\x12\x11\n\tavailable\x18\x02 \x01(\x08\x12$\n\rdevice_status\x18\x06 \x01(\x0b\x32\r.DeviceStatus\x12\"\n\x0cmood_message\x18\t \x01(\x0b\x32\x0c.MoodMessage\x12\x1c\n\tlast_seen\x18\n \x01(\x0b\x32\t.LastSeen\"N\n\x0ePresenceResult\x12\x1f\n\x07user_id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12\x1b\n\x08presence\x18\x02 \x01(\x0b\x32\t.Presence\"7\n\x10\x43lientIdentifier\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x11\n\theader_id\x18\x02 \x01(\t\"e\n\x13\x43lientPresenceState\x12%\n\nidentifier\x18\x01 \x01(\x0b\x32\x11.ClientIdentifier\x12\'\n\x05state\x18\x02 \x01(\x0e\x32\x18.ClientPresenceStateType\"~\n\x0eUserEventState\x12\x1f\n\x07user_id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12\x1b\n\x13\x63lient_generated_id\x18\x02 \x01(\t\x12.\n\x12notification_level\x18\x03 \x01(\x0e\x32\x12.NotificationLevel\"T\n\nFormatting\x12\x0c\n\x04\x62old\x18\x01 \x01(\x08\x12\x0e\n\x06italic\x18\x02 \x01(\x08\x12\x15\n\rstrikethrough\x18\x03 \x01(\x08\x12\x11\n\tunderline\x18\x04 \x01(\x08\"\x1f\n\x08LinkData\x12\x13\n\x0blink_target\x18\x01 \x01(\t\"r\n\x07Segment\x12\x1a\n\x04type\x18\x01 \x02(\x0e\x32\x0c.SegmentType\x12\x0c\n\x04text\x18\x02 \x01(\t\x12\x1f\n\nformatting\x18\x03 \x01(\x0b\x32\x0b.Formatting\x12\x1c\n\tlink_data\x18\x04 \x01(\x0b\x32\t.LinkData\"\x89\x03\n\tPlusPhoto\x12\'\n\tthumbnail\x18\x01 \x01(\x0b\x32\x14.PlusPhoto.Thumbnail\x12\x1b\n\x13owner_obfuscated_id\x18\x02 \x01(\t\x12\x10\n\x08\x61lbum_id\x18\x03 \x01(\t\x12\x10\n\x08photo_id\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x06 \x01(\t\x12\x1c\n\x14original_content_url\x18\n \x01(\t\x12(\n\nmedia_type\x18\r \x01(\x0e\x32\x14.PlusPhoto.MediaType\x12\x11\n\tstream_id\x18\x0e \x03(\t\x1aP\n\tThumbnail\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x11\n\timage_url\x18\x04 \x01(\t\x12\x10\n\x08width_px\x18\n \x01(\x04\x12\x11\n\theight_px\x18\x0b \x01(\x04\"X\n\tMediaType\x12\x16\n\x12MEDIA_TYPE_UNKNOWN\x10\x00\x12\x14\n\x10MEDIA_TYPE_PHOTO\x10\x01\x12\x1d\n\x19MEDIA_TYPE_ANIMATED_PHOTO\x10\x04\"\x83\x01\n\x05Place\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1b\n\x07\x61\x64\x64ress\x18\x18 \x01(\x0b\x32\n.EmbedItem\x12\x17\n\x03geo\x18\x19 \x01(\x0b\x32\n.EmbedItem\x12)\n\x14representative_image\x18\xb9\x01 \x01(\x0b\x32\n.EmbedItem\"\xf3\x02\n\tEmbedItem\x12\x17\n\x04type\x18\x01 \x03(\x0e\x32\t.ItemType\x12\n\n\x02id\x18\x02 \x01(\t\x12!\n\nplus_photo\x18\x95\x81\x97\r \x01(\x0b\x32\n.PlusPhoto\x12\x18\n\x05place\x18\xe8\xcf\x8a\x11 \x01(\x0b\x32\x06.Place\x12\x33\n\x0epostal_address\x18\xe2\xbb\x95\x11 \x01(\x0b\x32\x18.EmbedItem.PostalAddress\x12\x35\n\x0fgeo_coordinates\x18\xed\x9d\xc2\x11 \x01(\x0b\x32\x19.EmbedItem.GeoCoordinates\x12\"\n\x05image\x18\xc9\xca\x99\x13 \x01(\x0b\x32\x10.EmbedItem.Image\x1a\'\n\rPostalAddress\x12\x16\n\x0estreet_address\x18# \x01(\t\x1a\x35\n\x0eGeoCoordinates\x12\x10\n\x08latitude\x18$ \x01(\x01\x12\x11\n\tlongitude\x18% \x01(\x01\x1a\x14\n\x05Image\x12\x0b\n\x03url\x18\x01 \x01(\t\",\n\nAttachment\x12\x1e\n\nembed_item\x18\x01 \x01(\x0b\x32\n.EmbedItem\"L\n\x0eMessageContent\x12\x19\n\x07segment\x18\x01 \x03(\x0b\x32\x08.Segment\x12\x1f\n\nattachment\x18\x02 \x03(\x0b\x32\x0b.Attachment\".\n\x0f\x45ventAnnotation\x12\x0c\n\x04type\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t\"]\n\x0b\x43hatMessage\x12$\n\nannotation\x18\x02 \x03(\x0b\x32\x10.EventAnnotation\x12(\n\x0fmessage_content\x18\x03 \x01(\x0b\x32\x0f.MessageContent\"`\n\x10MembershipChange\x12#\n\x04type\x18\x01 \x01(\x0e\x32\x15.MembershipChangeType\x12\'\n\x0fparticipant_ids\x18\x03 \x03(\x0b\x32\x0e.ParticipantId\"8\n\x12\x43onversationRename\x12\x10\n\x08new_name\x18\x01 \x01(\t\x12\x10\n\x08old_name\x18\x02 \x01(\t\"]\n\x0cHangoutEvent\x12%\n\nevent_type\x18\x01 \x01(\x0e\x32\x11.HangoutEventType\x12&\n\x0eparticipant_id\x18\x02 \x03(\x0b\x32\x0e.ParticipantId\"\xc5\x01\n\x0fOTRModification\x12+\n\x0eold_otr_status\x18\x01 \x01(\x0e\x32\x13.OffTheRecordStatus\x12+\n\x0enew_otr_status\x18\x02 \x01(\x0e\x32\x13.OffTheRecordStatus\x12+\n\x0eold_otr_toggle\x18\x03 \x01(\x0e\x32\x13.OffTheRecordToggle\x12+\n\x0enew_otr_toggle\x18\x04 \x01(\x0e\x32\x13.OffTheRecordToggle\"E\n\x0cHashModifier\x12\x11\n\tupdate_id\x18\x01 \x01(\t\x12\x11\n\thash_diff\x18\x02 \x01(\x04\x12\x0f\n\x07version\x18\x04 \x01(\x04\"\xe2\x05\n\x05\x45vent\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12!\n\tsender_id\x18\x02 \x01(\x0b\x32\x0e.ParticipantId\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12)\n\x10self_event_state\x18\x04 \x01(\x0b\x32\x0f.UserEventState\x12 \n\x0bsource_type\x18\x06 \x01(\x0e\x32\x0b.SourceType\x12\"\n\x0c\x63hat_message\x18\x07 \x01(\x0b\x32\x0c.ChatMessage\x12,\n\x11membership_change\x18\t \x01(\x0b\x32\x11.MembershipChange\x12\x30\n\x13\x63onversation_rename\x18\n \x01(\x0b\x32\x13.ConversationRename\x12$\n\rhangout_event\x18\x0b \x01(\x0b\x32\r.HangoutEvent\x12\x10\n\x08\x65vent_id\x18\x0c \x01(\t\x12\x1c\n\x14\x65xpiration_timestamp\x18\r \x01(\x04\x12*\n\x10otr_modification\x18\x0e \x01(\x0b\x32\x10.OTRModification\x12\x1f\n\x17\x61\x64vances_sort_timestamp\x18\x0f \x01(\x08\x12\'\n\notr_status\x18\x10 \x01(\x0e\x32\x13.OffTheRecordStatus\x12\x11\n\tpersisted\x18\x11 \x01(\x08\x12$\n\x0bmedium_type\x18\x14 \x01(\x0b\x32\x0f.DeliveryMedium\x12\x1e\n\nevent_type\x18\x17 \x01(\x0e\x32\n.EventType\x12\x15\n\revent_version\x18\x18 \x01(\x04\x12$\n\rhash_modifier\x18\x1a \x01(\x0b\x32\r.HashModifier\x12\x46\n\x1fgroup_link_sharing_modification\x18\x1f \x01(\x0b\x32\x1d.GroupLinkSharingModification\"V\n\rUserReadState\x12&\n\x0eparticipant_id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12\x1d\n\x15latest_read_timestamp\x18\x02 \x01(\x04\"^\n\x0e\x44\x65liveryMedium\x12(\n\x0bmedium_type\x18\x01 \x01(\x0e\x32\x13.DeliveryMediumType\x12\"\n\x0cphone_number\x18\x02 \x01(\x0b\x32\x0c.PhoneNumber\"Y\n\x14\x44\x65liveryMediumOption\x12(\n\x0f\x64\x65livery_medium\x18\x01 \x01(\x0b\x32\x0f.DeliveryMedium\x12\x17\n\x0f\x63urrent_default\x18\x02 \x01(\x08\"\xa8\x03\n\x15UserConversationState\x12\x1b\n\x13\x63lient_generated_id\x18\x02 \x01(\t\x12\'\n\x0fself_read_state\x18\x07 \x01(\x0b\x32\x0e.UserReadState\x12#\n\x06status\x18\x08 \x01(\x0e\x32\x13.ConversationStatus\x12.\n\x12notification_level\x18\t \x01(\x0e\x32\x12.NotificationLevel\x12\x1f\n\x04view\x18\n \x03(\x0e\x32\x11.ConversationView\x12\"\n\ninviter_id\x18\x0b \x01(\x0b\x32\x0e.ParticipantId\x12\x18\n\x10invite_timestamp\x18\x0c \x01(\x04\x12\x16\n\x0esort_timestamp\x18\r \x01(\x04\x12\x18\n\x10\x61\x63tive_timestamp\x18\x0e \x01(\x04\x12,\n\x0finvite_affinity\x18\x0f \x01(\x0e\x32\x13.InvitationAffinity\x12\x35\n\x16\x64\x65livery_medium_option\x18\x11 \x03(\x0b\x32\x15.DeliveryMediumOption\"\xdc\x01\n\x1b\x43onversationParticipantData\x12\x1a\n\x02id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12\x15\n\rfallback_name\x18\x02 \x01(\t\x12,\n\x11invitation_status\x18\x03 \x01(\x0e\x32\x11.InvitationStatus\x12*\n\x10participant_type\x18\x05 \x01(\x0e\x32\x10.ParticipantType\x12\x30\n\x15new_invitation_status\x18\x06 \x01(\x0e\x32\x11.InvitationStatus\"\xcb\x04\n\x0c\x43onversation\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12\x1f\n\x04type\x18\x02 \x01(\x0e\x32\x11.ConversationType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x37\n\x17self_conversation_state\x18\x04 \x01(\x0b\x32\x16.UserConversationState\x12\"\n\nread_state\x18\x08 \x03(\x0b\x32\x0e.UserReadState\x12\x1a\n\x12has_active_hangout\x18\t \x01(\x08\x12\'\n\notr_status\x18\n \x01(\x0e\x32\x13.OffTheRecordStatus\x12\'\n\notr_toggle\x18\x0b \x01(\x0e\x32\x13.OffTheRecordToggle\x12&\n\x1e\x63onversation_history_supported\x18\x0c \x01(\x08\x12+\n\x13\x63urrent_participant\x18\r \x03(\x0b\x32\x0e.ParticipantId\x12\x36\n\x10participant_data\x18\x0e \x03(\x0b\x32\x1c.ConversationParticipantData\x12\"\n\x0cnetwork_type\x18\x12 \x03(\x0e\x32\x0c.NetworkType\x12*\n\x13\x66orce_history_state\x18\x13 \x01(\x0e\x32\r.ForceHistory\x12:\n\x19group_link_sharing_status\x18\x16 \x01(\x0e\x32\x17.GroupLinkSharingStatus\"\x1c\n\tEasterEgg\x12\x0f\n\x07message\x18\x01 \x01(\t\"`\n\x10\x42lockStateChange\x12&\n\x0eparticipant_id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12$\n\x0fnew_block_state\x18\x02 \x01(\x0e\x32\x0b.BlockState\"l\n\x05Photo\x12\x10\n\x08photo_id\x18\x01 \x01(\t\x12%\n\x1d\x64\x65lete_albumless_source_photo\x18\x02 \x01(\x08\x12\x0f\n\x07user_id\x18\x03 \x01(\t\x12\x19\n\x11is_custom_user_id\x18\x04 \x01(\x08\"&\n\rExistingMedia\x12\x15\n\x05photo\x18\x01 \x01(\x0b\x32\x06.Photo\"\xd0\x01\n\x12\x45ventRequestHeader\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12\x1b\n\x13\x63lient_generated_id\x18\x02 \x01(\x04\x12)\n\x0c\x65xpected_otr\x18\x03 \x01(\x0e\x32\x13.OffTheRecordStatus\x12(\n\x0f\x64\x65livery_medium\x18\x04 \x01(\x0b\x32\x0f.DeliveryMedium\x12\x1e\n\nevent_type\x18\x05 \x01(\x0e\x32\n.EventType\"\xb9\x01\n\rClientVersion\x12\x1c\n\tclient_id\x18\x01 \x01(\x0e\x32\t.ClientId\x12$\n\nbuild_type\x18\x02 \x01(\x0e\x32\x10.ClientBuildType\x12\x15\n\rmajor_version\x18\x03 \x01(\t\x12\x19\n\x11version_timestamp\x18\x04 \x01(\x04\x12\x19\n\x11\x64\x65vice_os_version\x18\x05 \x01(\t\x12\x17\n\x0f\x64\x65vice_hardware\x18\x06 \x01(\t\"|\n\rRequestHeader\x12&\n\x0e\x63lient_version\x18\x01 \x01(\x0b\x32\x0e.ClientVersion\x12,\n\x11\x63lient_identifier\x18\x02 \x01(\x0b\x32\x11.ClientIdentifier\x12\x15\n\rlanguage_code\x18\x04 \x01(\t\"\x96\x01\n\x0eResponseHeader\x12\x1f\n\x06status\x18\x01 \x01(\x0e\x32\x0f.ResponseStatus\x12\x19\n\x11\x65rror_description\x18\x02 \x01(\t\x12\x11\n\tdebug_url\x18\x03 \x01(\t\x12\x18\n\x10request_trace_id\x18\x04 \x01(\t\x12\x1b\n\x13\x63urrent_server_time\x18\x05 \x01(\x04\"\xcf\x02\n\x06\x45ntity\x12\x1a\n\x02id\x18\t \x01(\x0b\x32\x0e.ParticipantId\x12\x1b\n\x08presence\x18\x08 \x01(\x0b\x32\t.Presence\x12%\n\nproperties\x18\n \x01(\x0b\x32\x11.EntityProperties\x12%\n\x0b\x65ntity_type\x18\r \x01(\x0e\x32\x10.ParticipantType\x12\x38\n\x16had_past_hangout_state\x18\x10 \x01(\x0e\x32\x18.Entity.PastHangoutState\"\x83\x01\n\x10PastHangoutState\x12\x1e\n\x1aPAST_HANGOUT_STATE_UNKNOWN\x10\x00\x12\'\n#PAST_HANGOUT_STATE_HAD_PAST_HANGOUT\x10\x01\x12&\n\"PAST_HANGOUT_STATE_NO_PAST_HANGOUT\x10\x02\"\xff\x01\n\x10\x45ntityProperties\x12\x1a\n\x04type\x18\x01 \x01(\x0e\x32\x0c.ProfileType\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x12\n\nfirst_name\x18\x03 \x01(\t\x12\x11\n\tphoto_url\x18\x04 \x01(\t\x12\r\n\x05\x65mail\x18\x05 \x03(\t\x12\r\n\x05phone\x18\x06 \x03(\t\x12\x17\n\x0fin_users_domain\x18\n \x01(\x08\x12\x17\n\x06gender\x18\x0b \x01(\x0e\x32\x07.Gender\x12)\n\x10photo_url_status\x18\x0c \x01(\x0e\x32\x0f.PhotoUrlStatus\x12\x17\n\x0f\x63\x61nonical_email\x18\x0f \x01(\t\"\xb4\x01\n\x11\x43onversationState\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12#\n\x0c\x63onversation\x18\x02 \x01(\x0b\x32\r.Conversation\x12\x15\n\x05\x65vent\x18\x03 \x03(\x0b\x32\x06.Event\x12\x39\n\x18\x65vent_continuation_token\x18\x05 \x01(\x0b\x32\x17.EventContinuationToken\"g\n\x16\x45ventContinuationToken\x12\x10\n\x08\x65vent_id\x18\x01 \x01(\t\x12\"\n\x1astorage_continuation_token\x18\x02 \x01(\x0c\x12\x17\n\x0f\x65vent_timestamp\x18\x03 \x01(\x04\"a\n\x10\x45ntityLookupSpec\x12\x0f\n\x07gaia_id\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\r\n\x05phone\x18\x04 \x01(\t\x12\x1e\n\x16\x63reate_offnetwork_gaia\x18\x06 \x01(\x08\"X\n\x10\x43onfigurationBit\x12\x35\n\x16\x63onfiguration_bit_type\x18\x01 \x01(\x0e\x32\x15.ConfigurationBitType\x12\r\n\x05value\x18\x02 \x01(\x08\"W\n\x11RichPresenceState\x12\x42\n\x1fget_rich_presence_enabled_state\x18\x03 \x03(\x0b\x32\x19.RichPresenceEnabledState\"L\n\x18RichPresenceEnabledState\x12\x1f\n\x04type\x18\x01 \x01(\x0e\x32\x11.RichPresenceType\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\"(\n\x11\x44\x65sktopOffSetting\x12\x13\n\x0b\x64\x65sktop_off\x18\x01 \x01(\x08\"7\n\x0f\x44\x65sktopOffState\x12\x13\n\x0b\x64\x65sktop_off\x18\x01 \x01(\x08\x12\x0f\n\x07version\x18\x02 \x01(\x04\":\n\nDndSetting\x12\x16\n\x0e\x64o_not_disturb\x18\x01 \x01(\x08\x12\x14\n\x0ctimeout_secs\x18\x02 \x01(\x04\"T\n\x14PresenceStateSetting\x12\x14\n\x0ctimeout_secs\x18\x01 \x01(\x04\x12&\n\x04type\x18\x02 \x01(\x0e\x32\x18.ClientPresenceStateType\"1\n\x0bMoodMessage\x12\"\n\x0cmood_content\x18\x01 \x01(\x0b\x32\x0c.MoodContent\"(\n\x0bMoodContent\x12\x19\n\x07segment\x18\x01 \x03(\x0b\x32\x08.Segment\"1\n\x0bMoodSetting\x12\"\n\x0cmood_message\x18\x01 \x01(\x0b\x32\x0c.MoodMessage\"/\n\tMoodState\x12\"\n\x0cmood_setting\x18\x04 \x01(\x0b\x32\x0c.MoodSetting\"w\n\x0c\x44\x65leteAction\x12\x1f\n\x17\x64\x65lete_action_timestamp\x18\x01 \x01(\x04\x12$\n\x1c\x64\x65lete_upper_bound_timestamp\x18\x02 \x01(\x04\x12 \n\x0b\x64\x65lete_type\x18\x03 \x01(\x0e\x32\x0b.DeleteType\"3\n\tInviteeID\x12\x0f\n\x07gaia_id\x18\x01 \x01(\t\x12\x15\n\rfallback_name\x18\x04 \x01(\t\"4\n\x07\x43ountry\x12\x13\n\x0bregion_code\x18\x01 \x01(\t\x12\x14\n\x0c\x63ountry_code\x18\x02 \x01(\x04\"n\n\x13\x44\x65sktopSoundSetting\x12(\n\x13\x64\x65sktop_sound_state\x18\x01 \x01(\x0e\x32\x0b.SoundState\x12-\n\x18\x64\x65sktop_ring_sound_state\x18\x02 \x01(\x0e\x32\x0b.SoundState\"Z\n\tPhoneData\x12\x15\n\x05phone\x18\x01 \x03(\x0b\x32\x06.Phone\x12\x36\n\x17\x63\x61ller_id_settings_mask\x18\x03 \x01(\x0e\x32\x15.CallerIdSettingsMask\"\xdc\x01\n\x05Phone\x12\"\n\x0cphone_number\x18\x01 \x01(\x0b\x32\x0c.PhoneNumber\x12\x14\n\x0cgoogle_voice\x18\x02 \x01(\x08\x12\x35\n\x13verification_status\x18\x03 \x01(\x0e\x32\x18.PhoneVerificationStatus\x12\x14\n\x0c\x64iscoverable\x18\x04 \x01(\x08\x12;\n\x16\x64iscoverability_status\x18\x05 \x01(\x0e\x32\x1b.PhoneDiscoverabilityStatus\x12\x0f\n\x07primary\x18\x06 \x01(\x08\"\xb1\x01\n\x08I18nData\x12\x17\n\x0fnational_number\x18\x01 \x01(\t\x12\x1c\n\x14international_number\x18\x02 \x01(\t\x12\x14\n\x0c\x63ountry_code\x18\x03 \x01(\x04\x12\x13\n\x0bregion_code\x18\x04 \x01(\t\x12\x10\n\x08is_valid\x18\x05 \x01(\x08\x12\x31\n\x11validation_result\x18\x06 \x01(\x0e\x32\x16.PhoneValidationResult\"9\n\x0bPhoneNumber\x12\x0c\n\x04\x65\x31\x36\x34\x18\x01 \x01(\t\x12\x1c\n\ti18n_data\x18\x02 \x01(\x0b\x32\t.I18nData\">\n\x19SuggestedContactGroupHash\x12\x13\n\x0bmax_results\x18\x01 \x01(\x04\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"Y\n\x10SuggestedContact\x12\x17\n\x06\x65ntity\x18\x01 \x01(\x0b\x32\x07.Entity\x12,\n\x11invitation_status\x18\x02 \x01(\x0e\x32\x11.InvitationStatus\"_\n\x15SuggestedContactGroup\x12\x14\n\x0chash_matched\x18\x01 \x01(\x08\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\"\n\x07\x63ontact\x18\x03 \x03(\x0b\x32\x11.SuggestedContact\"K\n\x1cGroupLinkSharingModification\x12+\n\nnew_status\x18\x01 \x01(\x0e\x32\x17.GroupLinkSharingStatus\"\xb2\x08\n\x0bStateUpdate\x12/\n\x13state_update_header\x18\x01 \x01(\x0b\x32\x12.StateUpdateHeader\x12#\n\x0c\x63onversation\x18\r \x01(\x0b\x32\r.Conversation\x12>\n\x19\x63onversation_notification\x18\x02 \x01(\x0b\x32\x19.ConversationNotificationH\x00\x12\x30\n\x12\x65vent_notification\x18\x03 \x01(\x0b\x32\x12.EventNotificationH\x00\x12\x33\n\x12\x66ocus_notification\x18\x04 \x01(\x0b\x32\x15.SetFocusNotificationH\x00\x12\x35\n\x13typing_notification\x18\x05 \x01(\x0b\x32\x16.SetTypingNotificationH\x00\x12X\n\x1fnotification_level_notification\x18\x06 \x01(\x0b\x32-.SetConversationNotificationLevelNotificationH\x00\x12\x42\n\x1creply_to_invite_notification\x18\x07 \x01(\x0b\x32\x1a.ReplyToInviteNotificationH\x00\x12\x38\n\x16watermark_notification\x18\x08 \x01(\x0b\x32\x16.WatermarkNotificationH\x00\x12:\n\x11view_modification\x18\x0b \x01(\x0b\x32\x1d.ConversationViewModificationH\x00\x12\x39\n\x17\x65\x61ster_egg_notification\x18\x0c \x01(\x0b\x32\x16.EasterEggNotificationH\x00\x12?\n\x1aself_presence_notification\x18\x0e \x01(\x0b\x32\x19.SelfPresenceNotificationH\x00\x12\x38\n\x13\x64\x65lete_notification\x18\x0f \x01(\x0b\x32\x19.DeleteActionNotificationH\x00\x12\x36\n\x15presence_notification\x18\x10 \x01(\x0b\x32\x15.PresenceNotificationH\x00\x12\x30\n\x12\x62lock_notification\x18\x11 \x01(\x0b\x32\x12.BlockNotificationH\x00\x12P\n!notification_setting_notification\x18\x13 \x01(\x0b\x32#.SetNotificationSettingNotificationH\x00\x12Y\n(rich_presence_enabled_state_notification\x18\x14 \x01(\x0b\x32%.RichPresenceEnabledStateNotificationH\x00\x42\x0e\n\x0cstate_update\"\xb1\x01\n\x11StateUpdateHeader\x12/\n\x13\x61\x63tive_client_state\x18\x01 \x01(\x0e\x32\x12.ActiveClientState\x12\x18\n\x10request_trace_id\x18\x03 \x01(\t\x12\x34\n\x15notification_settings\x18\x04 \x01(\x0b\x32\x15.NotificationSettings\x12\x1b\n\x13\x63urrent_server_time\x18\x05 \x01(\x04\"1\n\x0b\x42\x61tchUpdate\x12\"\n\x0cstate_update\x18\x01 \x03(\x0b\x32\x0c.StateUpdate\"?\n\x18\x43onversationNotification\x12#\n\x0c\x63onversation\x18\x01 \x01(\x0b\x32\r.Conversation\"*\n\x11\x45ventNotification\x12\x15\n\x05\x65vent\x18\x01 \x01(\x0b\x32\x06.Event\"\xae\x01\n\x14SetFocusNotification\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12!\n\tsender_id\x18\x02 \x01(\x0b\x32\x0e.ParticipantId\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x18\n\x04type\x18\x04 \x01(\x0e\x32\n.FocusType\x12\x1c\n\x06\x64\x65vice\x18\x05 \x01(\x0e\x32\x0c.FocusDevice\"\x92\x01\n\x15SetTypingNotification\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12!\n\tsender_id\x18\x02 \x01(\x0b\x32\x0e.ParticipantId\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x19\n\x04type\x18\x04 \x01(\x0e\x32\x0b.TypingType\"\x8e\x01\n,SetConversationNotificationLevelNotification\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12!\n\x05level\x18\x02 \x01(\x0e\x32\x12.NotificationLevel\x12\x11\n\ttimestamp\x18\x04 \x01(\x04\"g\n\x19ReplyToInviteNotification\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12 \n\x04type\x18\x02 \x01(\x0e\x32\x12.ReplyToInviteType\"\x83\x01\n\x15WatermarkNotification\x12!\n\tsender_id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12\x1d\n\x15latest_read_timestamp\x18\x03 \x01(\x04\"\x92\x01\n\x1c\x43onversationViewModification\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12#\n\x08old_view\x18\x02 \x01(\x0e\x32\x11.ConversationView\x12#\n\x08new_view\x18\x03 \x01(\x0e\x32\x11.ConversationView\"\x84\x01\n\x15\x45\x61sterEggNotification\x12!\n\tsender_id\x18\x01 \x01(\x0b\x32\x0e.ParticipantId\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12\x1e\n\neaster_egg\x18\x03 \x01(\x0b\x32\n.EasterEgg\"\x83\x02\n\x18SelfPresenceNotification\x12\x33\n\x15\x63lient_presence_state\x18\x01 \x01(\x0b\x32\x14.ClientPresenceState\x12\x34\n\x16\x64o_not_disturb_setting\x18\x03 \x01(\x0b\x32\x14.DoNotDisturbSetting\x12/\n\x13\x64\x65sktop_off_setting\x18\x04 \x01(\x0b\x32\x12.DesktopOffSetting\x12+\n\x11\x64\x65sktop_off_state\x18\x05 \x01(\x0b\x32\x10.DesktopOffState\x12\x1e\n\nmood_state\x18\x06 \x01(\x0b\x32\n.MoodState\"j\n\x18\x44\x65leteActionNotification\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\x12$\n\rdelete_action\x18\x02 \x01(\x0b\x32\r.DeleteAction\"9\n\x14PresenceNotification\x12!\n\x08presence\x18\x01 \x03(\x0b\x32\x0f.PresenceResult\"B\n\x11\x42lockNotification\x12-\n\x12\x62lock_state_change\x18\x01 \x03(\x0b\x32\x11.BlockStateChange\"\x87\x01\n\"SetNotificationSettingNotification\x12,\n\x11\x63onfiguration_bit\x18\x01 \x03(\x0b\x32\x11.ConfigurationBit\x12\x33\n\x15\x64\x65sktop_sound_setting\x18\x02 \x01(\x0b\x32\x14.DesktopSoundSetting\"f\n$RichPresenceEnabledStateNotification\x12>\n\x1brich_presence_enabled_state\x18\x01 \x03(\x0b\x32\x19.RichPresenceEnabledState\"<\n\x10\x43onversationSpec\x12(\n\x0f\x63onversation_id\x18\x01 \x01(\x0b\x32\x0f.ConversationId\"H\n\x11OffnetworkAddress\x12$\n\x04type\x18\x01 \x01(\x0e\x32\x16.OffnetworkAddressType\x12\r\n\x05\x65mail\x18\x03 \x01(\t\"O\n\x0c\x45ntityResult\x12&\n\x0blookup_spec\x18\x01 \x01(\x0b\x32\x11.EntityLookupSpec\x12\x17\n\x06\x65ntity\x18\x02 \x03(\x0b\x32\x07.Entity\"\x8b\x01\n\x0e\x41\x64\x64UserRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x1e\n\ninvitee_id\x18\x03 \x03(\x0b\x32\n.InviteeID\x12\x31\n\x14\x65vent_request_header\x18\x05 \x01(\x0b\x32\x13.EventRequestHeader\"Z\n\x0f\x41\x64\x64UserResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1d\n\rcreated_event\x18\x05 \x01(\x0b\x32\x06.Event\"\xaf\x01\n\x19\x43reateConversationRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x1f\n\x04type\x18\x02 \x01(\x0e\x32\x11.ConversationType\x12\x1b\n\x13\x63lient_generated_id\x18\x03 \x01(\x04\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\x1e\n\ninvitee_id\x18\x05 \x03(\x0b\x32\n.InviteeID\"\x8d\x01\n\x1a\x43reateConversationResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12#\n\x0c\x63onversation\x18\x02 \x01(\x0b\x32\r.Conversation\x12 \n\x18new_conversation_created\x18\x07 \x01(\x08\"\x93\x01\n\x19\x44\x65leteConversationRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12$\n\x1c\x64\x65lete_upper_bound_timestamp\x18\x03 \x01(\x04\"l\n\x1a\x44\x65leteConversationResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12$\n\rdelete_action\x18\x02 \x01(\x0b\x32\r.DeleteAction\"\x84\x01\n\x10\x45\x61sterEggRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12\x1e\n\neaster_egg\x18\x03 \x01(\x0b\x32\n.EasterEgg\"P\n\x11\x45\x61sterEggResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"\x8c\x02\n\x16GetConversationRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12,\n\x11\x63onversation_spec\x18\x02 \x01(\x0b\x32\x11.ConversationSpec\x12%\n\x1dinclude_conversation_metadata\x18\x03 \x01(\x08\x12\x15\n\rinclude_event\x18\x04 \x01(\x08\x12#\n\x1bmax_events_per_conversation\x18\x06 \x01(\x04\x12\x39\n\x18\x65vent_continuation_token\x18\x07 \x01(\x0b\x32\x17.EventContinuationToken\"s\n\x17GetConversationResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12.\n\x12\x63onversation_state\x18\x02 \x01(\x0b\x32\x12.ConversationState\"l\n\x14GetEntityByIdRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12,\n\x11\x62\x61tch_lookup_spec\x18\x03 \x03(\x0b\x32\x11.EntityLookupSpec\"\x80\x01\n\x15GetEntityByIdResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x17\n\x06\x65ntity\x18\x02 \x03(\x0b\x32\x07.Entity\x12$\n\rentity_result\x18\x03 \x03(\x0b\x32\r.EntityResult\"r\n\x1eGetGroupConversationUrlRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\"k\n\x1fGetGroupConversationUrlResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1e\n\x16group_conversation_url\x18\x02 \x01(\t\"\xa8\x03\n\x1bGetSuggestedEntitiesRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x11\n\tmax_count\x18\x04 \x01(\x04\x12-\n\tfavorites\x18\x08 \x01(\x0b\x32\x1a.SuggestedContactGroupHash\x12=\n\x19\x63ontacts_you_hangout_with\x18\t \x01(\x0b\x32\x1a.SuggestedContactGroupHash\x12>\n\x1aother_contacts_on_hangouts\x18\n \x01(\x0b\x32\x1a.SuggestedContactGroupHash\x12\x32\n\x0eother_contacts\x18\x0b \x01(\x0b\x32\x1a.SuggestedContactGroupHash\x12\x36\n\x12\x64ismissed_contacts\x18\x0c \x01(\x0b\x32\x1a.SuggestedContactGroupHash\x12\x34\n\x10pinned_favorites\x18\r \x01(\x0b\x32\x1a.SuggestedContactGroupHash\"\x99\x03\n\x1cGetSuggestedEntitiesResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x17\n\x06\x65ntity\x18\x02 \x03(\x0b\x32\x07.Entity\x12)\n\tfavorites\x18\x04 \x01(\x0b\x32\x16.SuggestedContactGroup\x12\x39\n\x19\x63ontacts_you_hangout_with\x18\x05 \x01(\x0b\x32\x16.SuggestedContactGroup\x12:\n\x1aother_contacts_on_hangouts\x18\x06 \x01(\x0b\x32\x16.SuggestedContactGroup\x12.\n\x0eother_contacts\x18\x07 \x01(\x0b\x32\x16.SuggestedContactGroup\x12\x32\n\x12\x64ismissed_contacts\x18\x08 \x01(\x0b\x32\x16.SuggestedContactGroup\x12\x30\n\x10pinned_favorites\x18\t \x01(\x0b\x32\x16.SuggestedContactGroup\"<\n\x12GetSelfInfoRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\"\xed\x03\n\x13GetSelfInfoResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1c\n\x0bself_entity\x18\x02 \x01(\x0b\x32\x07.Entity\x12\x16\n\x0eis_known_minor\x18\x03 \x01(\x08\x12\'\n\tdnd_state\x18\x05 \x01(\x0b\x32\x14.DoNotDisturbSetting\x12/\n\x13\x64\x65sktop_off_setting\x18\x06 \x01(\x0b\x32\x12.DesktopOffSetting\x12\x1e\n\nphone_data\x18\x07 \x01(\x0b\x32\n.PhoneData\x12,\n\x11\x63onfiguration_bit\x18\x08 \x03(\x0b\x32\x11.ConfigurationBit\x12+\n\x11\x64\x65sktop_off_state\x18\t \x01(\x0b\x32\x10.DesktopOffState\x12\x18\n\x10google_plus_user\x18\n \x01(\x08\x12\x33\n\x15\x64\x65sktop_sound_setting\x18\x0b \x01(\x0b\x32\x14.DesktopSoundSetting\x12/\n\x13rich_presence_state\x18\x0c \x01(\x0b\x32\x12.RichPresenceState\x12!\n\x0f\x64\x65\x66\x61ult_country\x18\x13 \x01(\x0b\x32\x08.Country\"\x86\x01\n\x14QueryPresenceRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12&\n\x0eparticipant_id\x18\x02 \x03(\x0b\x32\x0e.ParticipantId\x12\x1e\n\nfield_mask\x18\x03 \x03(\x0e\x32\n.FieldMask\"k\n\x15QueryPresenceResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12(\n\x0fpresence_result\x18\x02 \x03(\x0b\x32\x0f.PresenceResult\"\x96\x01\n\x11RemoveUserRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12&\n\x0eparticipant_id\x18\x03 \x01(\x0b\x32\x0e.ParticipantId\x12\x31\n\x14\x65vent_request_header\x18\x05 \x01(\x0b\x32\x13.EventRequestHeader\"]\n\x12RemoveUserResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1d\n\rcreated_event\x18\x04 \x01(\x0b\x32\x06.Event\"\x88\x01\n\x19RenameConversationRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x10\n\x08new_name\x18\x03 \x01(\t\x12\x31\n\x14\x65vent_request_header\x18\x05 \x01(\x0b\x32\x13.EventRequestHeader\"e\n\x1aRenameConversationResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1d\n\rcreated_event\x18\x04 \x01(\x0b\x32\x06.Event\"a\n\x15SearchEntitiesRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\r\n\x05query\x18\x03 \x01(\t\x12\x11\n\tmax_count\x18\x04 \x01(\x04\"[\n\x16SearchEntitiesResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x17\n\x06\x65ntity\x18\x02 \x03(\x0b\x32\x07.Entity\"!\n\x08Location\x12\x15\n\x05place\x18\x01 \x01(\x0b\x32\x06.Place\"\xa9\x02\n\x16SendChatMessageRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12$\n\nannotation\x18\x05 \x03(\x0b\x32\x10.EventAnnotation\x12(\n\x0fmessage_content\x18\x06 \x01(\x0b\x32\x0f.MessageContent\x12&\n\x0e\x65xisting_media\x18\x07 \x01(\x0b\x32\x0e.ExistingMedia\x12\x31\n\x14\x65vent_request_header\x18\x08 \x01(\x0b\x32\x13.EventRequestHeader\x12\x1f\n\x07user_id\x18\t \x01(\x0b\x32\x0e.ParticipantId\x12\x1b\n\x08location\x18\n \x01(\x0b\x32\t.Location\"b\n\x17SendChatMessageResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1d\n\rcreated_event\x18\x06 \x01(\x0b\x32\x06.Event\"\x9c\x01\n\x16ModifyOTRStatusRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\'\n\notr_status\x18\x03 \x01(\x0e\x32\x13.OffTheRecordStatus\x12\x31\n\x14\x65vent_request_header\x18\x05 \x01(\x0b\x32\x13.EventRequestHeader\"b\n\x17ModifyOTRStatusResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1d\n\rcreated_event\x18\x04 \x01(\x0b\x32\x06.Event\"v\n\x1fSendOffnetworkInvitationRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12+\n\x0finvitee_address\x18\x02 \x01(\x0b\x32\x12.OffnetworkAddress\"L\n SendOffnetworkInvitationResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\"{\n\x16SetActiveClientRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x11\n\tis_active\x18\x02 \x01(\x08\x12\x10\n\x08\x66ull_jid\x18\x03 \x01(\t\x12\x14\n\x0ctimeout_secs\x18\x04 \x01(\x04\"C\n\x17SetActiveClientResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\"E\n\x1bSetConversationLevelRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\"H\n\x1cSetConversationLevelResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\"\x9e\x01\n\'SetConversationNotificationLevelRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12!\n\x05level\x18\x03 \x01(\x0e\x32\x12.NotificationLevel\"g\n(SetConversationNotificationLevelResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"\x93\x01\n\x0fSetFocusRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12\x18\n\x04type\x18\x03 \x01(\x0e\x32\n.FocusType\x12\x14\n\x0ctimeout_secs\x18\x04 \x01(\r\"O\n\x10SetFocusResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"\xba\x01\n!SetGroupLinkSharingEnabledRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x31\n\x14\x65vent_request_header\x18\x02 \x01(\x0b\x32\x13.EventRequestHeader\x12:\n\x19group_link_sharing_status\x18\x04 \x01(\x0e\x32\x17.GroupLinkSharingStatus\"\x9a\x01\n\"SetGroupLinkSharingEnabledResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x1d\n\rcreated_event\x18\x02 \x01(\x0b\x32\x06.Event\x12+\n\x14updated_conversation\x18\x03 \x01(\x0b\x32\r.Conversation\"\xea\x01\n\x12SetPresenceRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x35\n\x16presence_state_setting\x18\x02 \x01(\x0b\x32\x15.PresenceStateSetting\x12 \n\x0b\x64nd_setting\x18\x03 \x01(\x0b\x32\x0b.DndSetting\x12/\n\x13\x64\x65sktop_off_setting\x18\x05 \x01(\x0b\x32\x12.DesktopOffSetting\x12\"\n\x0cmood_setting\x18\x08 \x01(\x0b\x32\x0c.MoodSetting\"?\n\x13SetPresenceResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\"\x7f\n\x10SetTypingRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12\x19\n\x04type\x18\x03 \x01(\x0e\x32\x0b.TypingType\"P\n\x11SetTypingResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"\x7f\n\x17SyncAllNewEventsRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x1b\n\x13last_sync_timestamp\x18\x02 \x01(\x04\x12\x1f\n\x17max_response_size_bytes\x18\x08 \x01(\x04\"\x8c\x01\n\x18SyncAllNewEventsResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x16\n\x0esync_timestamp\x18\x02 \x01(\x04\x12.\n\x12\x63onversation_state\x18\x03 \x03(\x0b\x32\x12.ConversationState\"\xc8\x01\n\x1eSyncRecentConversationsRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12\x1c\n\x14last_event_timestamp\x18\x02 \x01(\x04\x12\x19\n\x11max_conversations\x18\x03 \x01(\x04\x12#\n\x1bmax_events_per_conversation\x18\x04 \x01(\x04\x12 \n\x0bsync_filter\x18\x05 \x03(\x0e\x32\x0b.SyncFilter\"\xb7\x01\n\x1fSyncRecentConversationsResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader\x12\x16\n\x0esync_timestamp\x18\x02 \x01(\x04\x12.\n\x12\x63onversation_state\x18\x03 \x03(\x0b\x32\x12.ConversationState\x12\"\n\x1a\x63ontinuation_end_timestamp\x18\x04 \x01(\x04\"\x87\x01\n\x16UpdateWatermarkRequest\x12&\n\x0erequest_header\x18\x01 \x01(\x0b\x32\x0e.RequestHeader\x12(\n\x0f\x63onversation_id\x18\x02 \x01(\x0b\x32\x0f.ConversationId\x12\x1b\n\x13last_read_timestamp\x18\x03 \x01(\x04\"C\n\x17UpdateWatermarkResponse\x12(\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeader*\x7f\n\x11\x41\x63tiveClientState\x12!\n\x1d\x41\x43TIVE_CLIENT_STATE_NO_ACTIVE\x10\x00\x12!\n\x1d\x41\x43TIVE_CLIENT_STATE_IS_ACTIVE\x10\x01\x12$\n ACTIVE_CLIENT_STATE_OTHER_ACTIVE\x10\x02*U\n\tFocusType\x12\x16\n\x12\x46OCUS_TYPE_UNKNOWN\x10\x00\x12\x16\n\x12\x46OCUS_TYPE_FOCUSED\x10\x01\x12\x18\n\x14\x46OCUS_TYPE_UNFOCUSED\x10\x02*_\n\x0b\x46ocusDevice\x12\x1c\n\x18\x46OCUS_DEVICE_UNSPECIFIED\x10\x00\x12\x18\n\x14\x46OCUS_DEVICE_DESKTOP\x10\x14\x12\x18\n\x13\x46OCUS_DEVICE_MOBILE\x10\xac\x02*o\n\nTypingType\x12\x17\n\x13TYPING_TYPE_UNKNOWN\x10\x00\x12\x17\n\x13TYPING_TYPE_STARTED\x10\x01\x12\x16\n\x12TYPING_TYPE_PAUSED\x10\x02\x12\x17\n\x13TYPING_TYPE_STOPPED\x10\x03*\xae\x01\n\x17\x43lientPresenceStateType\x12!\n\x1d\x43LIENT_PRESENCE_STATE_UNKNOWN\x10\x00\x12\x1e\n\x1a\x43LIENT_PRESENCE_STATE_NONE\x10\x01\x12&\n\"CLIENT_PRESENCE_STATE_DESKTOP_IDLE\x10\x1e\x12(\n$CLIENT_PRESENCE_STATE_DESKTOP_ACTIVE\x10(*n\n\x11NotificationLevel\x12\x1e\n\x1aNOTIFICATION_LEVEL_UNKNOWN\x10\x00\x12\x1c\n\x18NOTIFICATION_LEVEL_QUIET\x10\n\x12\x1b\n\x17NOTIFICATION_LEVEL_RING\x10\x1e*X\n\x0bSegmentType\x12\x15\n\x11SEGMENT_TYPE_TEXT\x10\x00\x12\x1b\n\x17SEGMENT_TYPE_LINE_BREAK\x10\x01\x12\x15\n\x11SEGMENT_TYPE_LINK\x10\x02*i\n\x08ItemType\x12\x13\n\x0fITEM_TYPE_THING\x10\x00\x12\x19\n\x14ITEM_TYPE_PLUS_PHOTO\x10\xf9\x01\x12\x14\n\x0fITEM_TYPE_PLACE\x10\xcf\x02\x12\x17\n\x12ITEM_TYPE_PLACE_V2\x10\xd4\x02*Y\n\x14MembershipChangeType\x12\x1f\n\x1bMEMBERSHIP_CHANGE_TYPE_JOIN\x10\x01\x12 \n\x1cMEMBERSHIP_CHANGE_TYPE_LEAVE\x10\x02*\xeb\x01\n\x10HangoutEventType\x12\x1e\n\x1aHANGOUT_EVENT_TYPE_UNKNOWN\x10\x00\x12\x1c\n\x18HANGOUT_EVENT_TYPE_START\x10\x01\x12\x1a\n\x16HANGOUT_EVENT_TYPE_END\x10\x02\x12\x1b\n\x17HANGOUT_EVENT_TYPE_JOIN\x10\x03\x12\x1c\n\x18HANGOUT_EVENT_TYPE_LEAVE\x10\x04\x12\"\n\x1eHANGOUT_EVENT_TYPE_COMING_SOON\x10\x05\x12\x1e\n\x1aHANGOUT_EVENT_TYPE_ONGOING\x10\x06*~\n\x12OffTheRecordToggle\x12!\n\x1dOFF_THE_RECORD_TOGGLE_UNKNOWN\x10\x00\x12!\n\x1dOFF_THE_RECORD_TOGGLE_ENABLED\x10\x01\x12\"\n\x1eOFF_THE_RECORD_TOGGLE_DISABLED\x10\x02*\x8a\x01\n\x12OffTheRecordStatus\x12!\n\x1dOFF_THE_RECORD_STATUS_UNKNOWN\x10\x00\x12(\n$OFF_THE_RECORD_STATUS_OFF_THE_RECORD\x10\x01\x12\'\n#OFF_THE_RECORD_STATUS_ON_THE_RECORD\x10\x02*%\n\nSourceType\x12\x17\n\x13SOURCE_TYPE_UNKNOWN\x10\x00*\xc2\x03\n\tEventType\x12\x16\n\x12\x45VENT_TYPE_UNKNOWN\x10\x00\x12#\n\x1f\x45VENT_TYPE_REGULAR_CHAT_MESSAGE\x10\x01\x12\x12\n\x0e\x45VENT_TYPE_SMS\x10\x02\x12\x18\n\x14\x45VENT_TYPE_VOICEMAIL\x10\x03\x12\x17\n\x13\x45VENT_TYPE_ADD_USER\x10\x04\x12\x1a\n\x16\x45VENT_TYPE_REMOVE_USER\x10\x05\x12\"\n\x1e\x45VENT_TYPE_CONVERSATION_RENAME\x10\x06\x12\x16\n\x12\x45VENT_TYPE_HANGOUT\x10\x07\x12\x19\n\x15\x45VENT_TYPE_PHONE_CALL\x10\x08\x12\x1f\n\x1b\x45VENT_TYPE_OTR_MODIFICATION\x10\t\x12\x1c\n\x18\x45VENT_TYPE_PLAN_MUTATION\x10\n\x12\x12\n\x0e\x45VENT_TYPE_MMS\x10\x0b\x12\x1c\n\x18\x45VENT_TYPE_DEPRECATED_12\x10\x0c\x12\x1d\n\x19\x45VENT_TYPE_OBSERVED_EVENT\x10\r\x12.\n*EVENT_TYPE_GROUP_LINK_SHARING_MODIFICATION\x10\x0e*p\n\x10\x43onversationType\x12\x1d\n\x19\x43ONVERSATION_TYPE_UNKNOWN\x10\x00\x12 \n\x1c\x43ONVERSATION_TYPE_ONE_TO_ONE\x10\x01\x12\x1b\n\x17\x43ONVERSATION_TYPE_GROUP\x10\x02*\x94\x01\n\x12\x43onversationStatus\x12\x1f\n\x1b\x43ONVERSATION_STATUS_UNKNOWN\x10\x00\x12\x1f\n\x1b\x43ONVERSATION_STATUS_INVITED\x10\x01\x12\x1e\n\x1a\x43ONVERSATION_STATUS_ACTIVE\x10\x02\x12\x1c\n\x18\x43ONVERSATION_STATUS_LEFT\x10\x03*n\n\x10\x43onversationView\x12\x1d\n\x19\x43ONVERSATION_VIEW_UNKNOWN\x10\x00\x12\x1b\n\x17\x43ONVERSATION_VIEW_INBOX\x10\x01\x12\x1e\n\x1a\x43ONVERSATION_VIEW_ARCHIVED\x10\x02*\x8d\x01\n\x12\x44\x65liveryMediumType\x12\x1b\n\x17\x44\x45LIVERY_MEDIUM_UNKNOWN\x10\x00\x12\x19\n\x15\x44\x45LIVERY_MEDIUM_BABEL\x10\x01\x12 \n\x1c\x44\x45LIVERY_MEDIUM_GOOGLE_VOICE\x10\x02\x12\x1d\n\x19\x44\x45LIVERY_MEDIUM_LOCAL_SMS\x10\x03*d\n\x12InvitationAffinity\x12\x1b\n\x17INVITE_AFFINITY_UNKNOWN\x10\x00\x12\x18\n\x14INVITE_AFFINITY_HIGH\x10\x01\x12\x17\n\x13INVITE_AFFINITY_LOW\x10\x02*m\n\x0fParticipantType\x12\x1c\n\x18PARTICIPANT_TYPE_UNKNOWN\x10\x00\x12\x19\n\x15PARTICIPANT_TYPE_GAIA\x10\x02\x12!\n\x1dPARTICIPANT_TYPE_GOOGLE_VOICE\x10\x03*p\n\x10InvitationStatus\x12\x1d\n\x19INVITATION_STATUS_UNKNOWN\x10\x00\x12\x1d\n\x19INVITATION_STATUS_PENDING\x10\x01\x12\x1e\n\x1aINVITATION_STATUS_ACCEPTED\x10\x02*?\n\x0c\x46orceHistory\x12\x19\n\x15\x46ORCE_HISTORY_UNKNOWN\x10\x00\x12\x14\n\x10\x46ORCE_HISTORY_NO\x10\x01*^\n\x0bNetworkType\x12\x18\n\x14NETWORK_TYPE_UNKNOWN\x10\x00\x12\x16\n\x12NETWORK_TYPE_BABEL\x10\x01\x12\x1d\n\x19NETWORK_TYPE_GOOGLE_VOICE\x10\x02*U\n\nBlockState\x12\x17\n\x13\x42LOCK_STATE_UNKNOWN\x10\x00\x12\x15\n\x11\x42LOCK_STATE_BLOCK\x10\x01\x12\x17\n\x13\x42LOCK_STATE_UNBLOCK\x10\x02*x\n\x11ReplyToInviteType\x12 \n\x1cREPLY_TO_INVITE_TYPE_UNKNOWN\x10\x00\x12\x1f\n\x1bREPLY_TO_INVITE_TYPE_ACCEPT\x10\x01\x12 \n\x1cREPLY_TO_INVITE_TYPE_DECLINE\x10\x02*\xca\x01\n\x08\x43lientId\x12\x15\n\x11\x43LIENT_ID_UNKNOWN\x10\x00\x12\x15\n\x11\x43LIENT_ID_ANDROID\x10\x01\x12\x11\n\rCLIENT_ID_IOS\x10\x02\x12\x14\n\x10\x43LIENT_ID_CHROME\x10\x03\x12\x17\n\x13\x43LIENT_ID_WEB_GPLUS\x10\x05\x12\x17\n\x13\x43LIENT_ID_WEB_GMAIL\x10\x06\x12\x19\n\x15\x43LIENT_ID_ULTRAVIOLET\x10\r\x12\x1a\n\x16\x43LIENT_ID_WEB_HANGOUTS\x10,*g\n\x0f\x43lientBuildType\x12\x16\n\x12\x42UILD_TYPE_UNKNOWN\x10\x00\x12\x1d\n\x19\x42UILD_TYPE_PRODUCTION_WEB\x10\x01\x12\x1d\n\x19\x42UILD_TYPE_PRODUCTION_APP\x10\x03*\x90\x01\n\x0eResponseStatus\x12\x1b\n\x17RESPONSE_STATUS_UNKNOWN\x10\x00\x12\x16\n\x12RESPONSE_STATUS_OK\x10\x01\x12$\n RESPONSE_STATUS_UNEXPECTED_ERROR\x10\x03\x12#\n\x1fRESPONSE_STATUS_INVALID_REQUEST\x10\x04*q\n\x0ePhotoUrlStatus\x12\x1c\n\x18PHOTO_URL_STATUS_UNKNOWN\x10\x00\x12 \n\x1cPHOTO_URL_STATUS_PLACEHOLDER\x10\x01\x12\x1f\n\x1bPHOTO_URL_STATUS_USER_PHOTO\x10\x02*@\n\x06Gender\x12\x12\n\x0eGENDER_UNKNOWN\x10\x00\x12\x0f\n\x0bGENDER_MALE\x10\x01\x12\x11\n\rGENDER_FEMALE\x10\x02*>\n\x0bProfileType\x12\x15\n\x11PROFILE_TYPE_NONE\x10\x00\x12\x18\n\x14PROFILE_TYPE_ES_USER\x10\x01*\x81\x0c\n\x14\x43onfigurationBitType\x12\"\n\x1e\x43ONFIGURATION_BIT_TYPE_UNKNOWN\x10\x00\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_1\x10\x01\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_2\x10\x02\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_3\x10\x03\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_4\x10\x04\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_5\x10\x05\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_6\x10\x06\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_7\x10\x07\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_8\x10\x08\x12$\n CONFIGURATION_BIT_TYPE_UNKNOWN_9\x10\t\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_10\x10\n\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_11\x10\x0b\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_12\x10\x0c\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_13\x10\r\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_14\x10\x0e\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_15\x10\x0f\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_16\x10\x10\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_17\x10\x11\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_18\x10\x12\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_19\x10\x13\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_20\x10\x14\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_21\x10\x15\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_22\x10\x16\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_23\x10\x17\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_24\x10\x18\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_25\x10\x19\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_26\x10\x1a\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_27\x10\x1b\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_28\x10\x1c\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_29\x10\x1d\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_30\x10\x1e\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_31\x10\x1f\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_32\x10 \x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_33\x10!\x12@\n<CONFIGURATION_BIT_TYPE_DESKTOP_AUTO_EMOJI_CONVERSION_ENABLED\x10\"\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_35\x10#\x12%\n!CONFIGURATION_BIT_TYPE_UNKNOWN_36\x10$\x12\x37\n3CONFIGURATION_BIT_TYPE_DESKTOP_COMPACT_MODE_ENABLED\x10&*\xff\x01\n\x10RichPresenceType\x12\x1e\n\x1aRICH_PRESENCE_TYPE_UNKNOWN\x10\x00\x12$\n RICH_PRESENCE_TYPE_IN_CALL_STATE\x10\x01\x12 \n\x1cRICH_PRESENCE_TYPE_UNKNOWN_3\x10\x03\x12 \n\x1cRICH_PRESENCE_TYPE_UNKNOWN_4\x10\x04\x12 \n\x1cRICH_PRESENCE_TYPE_UNKNOWN_5\x10\x05\x12\x1d\n\x19RICH_PRESENCE_TYPE_DEVICE\x10\x02\x12 \n\x1cRICH_PRESENCE_TYPE_LAST_SEEN\x10\x06*\x9d\x01\n\tFieldMask\x12\x18\n\x14\x46IELD_MASK_REACHABLE\x10\x01\x12\x18\n\x14\x46IELD_MASK_AVAILABLE\x10\x02\x12\x13\n\x0f\x46IELD_MASK_MOOD\x10\x03\x12\x16\n\x12\x46IELD_MASK_IN_CALL\x10\x06\x12\x15\n\x11\x46IELD_MASK_DEVICE\x10\x07\x12\x18\n\x14\x46IELD_MASK_LAST_SEEN\x10\n*B\n\nDeleteType\x12\x17\n\x13\x44\x45LETE_TYPE_UNKNOWN\x10\x00\x12\x1b\n\x17\x44\x45LETE_TYPE_UPPER_BOUND\x10\x01*V\n\nSyncFilter\x12\x17\n\x13SYNC_FILTER_UNKNOWN\x10\x00\x12\x15\n\x11SYNC_FILTER_INBOX\x10\x01\x12\x18\n\x14SYNC_FILTER_ARCHIVED\x10\x02*N\n\nSoundState\x12\x17\n\x13SOUND_STATE_UNKNOWN\x10\x00\x12\x12\n\x0eSOUND_STATE_ON\x10\x01\x12\x13\n\x0fSOUND_STATE_OFF\x10\x02*a\n\x14\x43\x61llerIdSettingsMask\x12#\n\x1f\x43\x41LLER_ID_SETTINGS_MASK_UNKNOWN\x10\x00\x12$\n CALLER_ID_SETTINGS_MASK_PROVIDED\x10\x01*h\n\x17PhoneVerificationStatus\x12%\n!PHONE_VERIFICATION_STATUS_UNKNOWN\x10\x00\x12&\n\"PHONE_VERIFICATION_STATUS_VERIFIED\x10\x01*\x86\x01\n\x1aPhoneDiscoverabilityStatus\x12(\n$PHONE_DISCOVERABILITY_STATUS_UNKNOWN\x10\x00\x12>\n:PHONE_DISCOVERABILITY_STATUS_OPTED_IN_BUT_NOT_DISCOVERABLE\x10\x02*@\n\x15PhoneValidationResult\x12\'\n#PHONE_VALIDATION_RESULT_IS_POSSIBLE\x10\x00*_\n\x15OffnetworkAddressType\x12#\n\x1fOFFNETWORK_ADDRESS_TYPE_UNKNOWN\x10\x00\x12!\n\x1dOFFNETWORK_ADDRESS_TYPE_EMAIL\x10\x01*\x84\x01\n\x16GroupLinkSharingStatus\x12%\n!GROUP_LINK_SHARING_STATUS_UNKNOWN\x10\x00\x12 \n\x1cGROUP_LINK_SHARING_STATUS_ON\x10\x01\x12!\n\x1dGROUP_LINK_SHARING_STATUS_OFF\x10\x02')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_ACTIVECLIENTSTATE = _descriptor.EnumDescriptor(
  name='ActiveClientState',
  full_name='ActiveClientState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ACTIVE_CLIENT_STATE_NO_ACTIVE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE_CLIENT_STATE_IS_ACTIVE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE_CLIENT_STATE_OTHER_ACTIVE', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20008,
  serialized_end=20135,
)
_sym_db.RegisterEnumDescriptor(_ACTIVECLIENTSTATE)

ActiveClientState = enum_type_wrapper.EnumTypeWrapper(_ACTIVECLIENTSTATE)
_FOCUSTYPE = _descriptor.EnumDescriptor(
  name='FocusType',
  full_name='FocusType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FOCUS_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FOCUS_TYPE_FOCUSED', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FOCUS_TYPE_UNFOCUSED', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20137,
  serialized_end=20222,
)
_sym_db.RegisterEnumDescriptor(_FOCUSTYPE)

FocusType = enum_type_wrapper.EnumTypeWrapper(_FOCUSTYPE)
_FOCUSDEVICE = _descriptor.EnumDescriptor(
  name='FocusDevice',
  full_name='FocusDevice',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FOCUS_DEVICE_UNSPECIFIED', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FOCUS_DEVICE_DESKTOP', index=1, number=20,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FOCUS_DEVICE_MOBILE', index=2, number=300,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20224,
  serialized_end=20319,
)
_sym_db.RegisterEnumDescriptor(_FOCUSDEVICE)

FocusDevice = enum_type_wrapper.EnumTypeWrapper(_FOCUSDEVICE)
_TYPINGTYPE = _descriptor.EnumDescriptor(
  name='TypingType',
  full_name='TypingType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPING_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPING_TYPE_STARTED', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPING_TYPE_PAUSED', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPING_TYPE_STOPPED', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20321,
  serialized_end=20432,
)
_sym_db.RegisterEnumDescriptor(_TYPINGTYPE)

TypingType = enum_type_wrapper.EnumTypeWrapper(_TYPINGTYPE)
_CLIENTPRESENCESTATETYPE = _descriptor.EnumDescriptor(
  name='ClientPresenceStateType',
  full_name='ClientPresenceStateType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CLIENT_PRESENCE_STATE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_PRESENCE_STATE_NONE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_PRESENCE_STATE_DESKTOP_IDLE', index=2, number=30,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_PRESENCE_STATE_DESKTOP_ACTIVE', index=3, number=40,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20435,
  serialized_end=20609,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPRESENCESTATETYPE)

ClientPresenceStateType = enum_type_wrapper.EnumTypeWrapper(_CLIENTPRESENCESTATETYPE)
_NOTIFICATIONLEVEL = _descriptor.EnumDescriptor(
  name='NotificationLevel',
  full_name='NotificationLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NOTIFICATION_LEVEL_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOTIFICATION_LEVEL_QUIET', index=1, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOTIFICATION_LEVEL_RING', index=2, number=30,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20611,
  serialized_end=20721,
)
_sym_db.RegisterEnumDescriptor(_NOTIFICATIONLEVEL)

NotificationLevel = enum_type_wrapper.EnumTypeWrapper(_NOTIFICATIONLEVEL)
_SEGMENTTYPE = _descriptor.EnumDescriptor(
  name='SegmentType',
  full_name='SegmentType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SEGMENT_TYPE_TEXT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEGMENT_TYPE_LINE_BREAK', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEGMENT_TYPE_LINK', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20723,
  serialized_end=20811,
)
_sym_db.RegisterEnumDescriptor(_SEGMENTTYPE)

SegmentType = enum_type_wrapper.EnumTypeWrapper(_SEGMENTTYPE)
_ITEMTYPE = _descriptor.EnumDescriptor(
  name='ItemType',
  full_name='ItemType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ITEM_TYPE_THING', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ITEM_TYPE_PLUS_PHOTO', index=1, number=249,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ITEM_TYPE_PLACE', index=2, number=335,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ITEM_TYPE_PLACE_V2', index=3, number=340,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20813,
  serialized_end=20918,
)
_sym_db.RegisterEnumDescriptor(_ITEMTYPE)

ItemType = enum_type_wrapper.EnumTypeWrapper(_ITEMTYPE)
_MEMBERSHIPCHANGETYPE = _descriptor.EnumDescriptor(
  name='MembershipChangeType',
  full_name='MembershipChangeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MEMBERSHIP_CHANGE_TYPE_JOIN', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEMBERSHIP_CHANGE_TYPE_LEAVE', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=20920,
  serialized_end=21009,
)
_sym_db.RegisterEnumDescriptor(_MEMBERSHIPCHANGETYPE)

MembershipChangeType = enum_type_wrapper.EnumTypeWrapper(_MEMBERSHIPCHANGETYPE)
_HANGOUTEVENTTYPE = _descriptor.EnumDescriptor(
  name='HangoutEventType',
  full_name='HangoutEventType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_START', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_END', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_JOIN', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_LEAVE', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_COMING_SOON', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HANGOUT_EVENT_TYPE_ONGOING', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=21012,
  serialized_end=21247,
)
_sym_db.RegisterEnumDescriptor(_HANGOUTEVENTTYPE)

HangoutEventType = enum_type_wrapper.EnumTypeWrapper(_HANGOUTEVENTTYPE)
_OFFTHERECORDTOGGLE = _descriptor.EnumDescriptor(
  name='OffTheRecordToggle',
  full_name='OffTheRecordToggle',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OFF_THE_RECORD_TOGGLE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFF_THE_RECORD_TOGGLE_ENABLED', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFF_THE_RECORD_TOGGLE_DISABLED', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=21249,
  serialized_end=21375,
)
_sym_db.RegisterEnumDescriptor(_OFFTHERECORDTOGGLE)

OffTheRecordToggle = enum_type_wrapper.EnumTypeWrapper(_OFFTHERECORDTOGGLE)
_OFFTHERECORDSTATUS = _descriptor.EnumDescriptor(
  name='OffTheRecordStatus',
  full_name='OffTheRecordStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OFF_THE_RECORD_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFF_THE_RECORD_STATUS_OFF_THE_RECORD', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFF_THE_RECORD_STATUS_ON_THE_RECORD', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=21378,
  serialized_end=21516,
)
_sym_db.RegisterEnumDescriptor(_OFFTHERECORDSTATUS)

OffTheRecordStatus = enum_type_wrapper.EnumTypeWrapper(_OFFTHERECORDSTATUS)
_SOURCETYPE = _descriptor.EnumDescriptor(
  name='SourceType',
  full_name='SourceType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SOURCE_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=21518,
  serialized_end=21555,
)
_sym_db.RegisterEnumDescriptor(_SOURCETYPE)

SourceType = enum_type_wrapper.EnumTypeWrapper(_SOURCETYPE)
_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='EventType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_REGULAR_CHAT_MESSAGE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_SMS', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_VOICEMAIL', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_ADD_USER', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_REMOVE_USER', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_CONVERSATION_RENAME', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_HANGOUT', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_PHONE_CALL', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_OTR_MODIFICATION', index=9, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_PLAN_MUTATION', index=10, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_MMS', index=11, number=11,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_DEPRECATED_12', index=12, number=12,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_OBSERVED_EVENT', index=13, number=13,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVENT_TYPE_GROUP_LINK_SHARING_MODIFICATION', index=14, number=14,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=21558,
  serialized_end=22008,
)
_sym_db.RegisterEnumDescriptor(_EVENTTYPE)

EventType = enum_type_wrapper.EnumTypeWrapper(_EVENTTYPE)
_CONVERSATIONTYPE = _descriptor.EnumDescriptor(
  name='ConversationType',
  full_name='ConversationType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_TYPE_ONE_TO_ONE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_TYPE_GROUP', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22010,
  serialized_end=22122,
)
_sym_db.RegisterEnumDescriptor(_CONVERSATIONTYPE)

ConversationType = enum_type_wrapper.EnumTypeWrapper(_CONVERSATIONTYPE)
_CONVERSATIONSTATUS = _descriptor.EnumDescriptor(
  name='ConversationStatus',
  full_name='ConversationStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_STATUS_INVITED', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_STATUS_ACTIVE', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_STATUS_LEFT', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22125,
  serialized_end=22273,
)
_sym_db.RegisterEnumDescriptor(_CONVERSATIONSTATUS)

ConversationStatus = enum_type_wrapper.EnumTypeWrapper(_CONVERSATIONSTATUS)
_CONVERSATIONVIEW = _descriptor.EnumDescriptor(
  name='ConversationView',
  full_name='ConversationView',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_VIEW_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_VIEW_INBOX', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVERSATION_VIEW_ARCHIVED', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22275,
  serialized_end=22385,
)
_sym_db.RegisterEnumDescriptor(_CONVERSATIONVIEW)

ConversationView = enum_type_wrapper.EnumTypeWrapper(_CONVERSATIONVIEW)
_DELIVERYMEDIUMTYPE = _descriptor.EnumDescriptor(
  name='DeliveryMediumType',
  full_name='DeliveryMediumType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DELIVERY_MEDIUM_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DELIVERY_MEDIUM_BABEL', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DELIVERY_MEDIUM_GOOGLE_VOICE', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DELIVERY_MEDIUM_LOCAL_SMS', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22388,
  serialized_end=22529,
)
_sym_db.RegisterEnumDescriptor(_DELIVERYMEDIUMTYPE)

DeliveryMediumType = enum_type_wrapper.EnumTypeWrapper(_DELIVERYMEDIUMTYPE)
_INVITATIONAFFINITY = _descriptor.EnumDescriptor(
  name='InvitationAffinity',
  full_name='InvitationAffinity',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INVITE_AFFINITY_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVITE_AFFINITY_HIGH', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVITE_AFFINITY_LOW', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22531,
  serialized_end=22631,
)
_sym_db.RegisterEnumDescriptor(_INVITATIONAFFINITY)

InvitationAffinity = enum_type_wrapper.EnumTypeWrapper(_INVITATIONAFFINITY)
_PARTICIPANTTYPE = _descriptor.EnumDescriptor(
  name='ParticipantType',
  full_name='ParticipantType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PARTICIPANT_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PARTICIPANT_TYPE_GAIA', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PARTICIPANT_TYPE_GOOGLE_VOICE', index=2, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22633,
  serialized_end=22742,
)
_sym_db.RegisterEnumDescriptor(_PARTICIPANTTYPE)

ParticipantType = enum_type_wrapper.EnumTypeWrapper(_PARTICIPANTTYPE)
_INVITATIONSTATUS = _descriptor.EnumDescriptor(
  name='InvitationStatus',
  full_name='InvitationStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INVITATION_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVITATION_STATUS_PENDING', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVITATION_STATUS_ACCEPTED', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22744,
  serialized_end=22856,
)
_sym_db.RegisterEnumDescriptor(_INVITATIONSTATUS)

InvitationStatus = enum_type_wrapper.EnumTypeWrapper(_INVITATIONSTATUS)
_FORCEHISTORY = _descriptor.EnumDescriptor(
  name='ForceHistory',
  full_name='ForceHistory',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FORCE_HISTORY_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FORCE_HISTORY_NO', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22858,
  serialized_end=22921,
)
_sym_db.RegisterEnumDescriptor(_FORCEHISTORY)

ForceHistory = enum_type_wrapper.EnumTypeWrapper(_FORCEHISTORY)
_NETWORKTYPE = _descriptor.EnumDescriptor(
  name='NetworkType',
  full_name='NetworkType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NETWORK_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NETWORK_TYPE_BABEL', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NETWORK_TYPE_GOOGLE_VOICE', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=22923,
  serialized_end=23017,
)
_sym_db.RegisterEnumDescriptor(_NETWORKTYPE)

NetworkType = enum_type_wrapper.EnumTypeWrapper(_NETWORKTYPE)
_BLOCKSTATE = _descriptor.EnumDescriptor(
  name='BlockState',
  full_name='BlockState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BLOCK_STATE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BLOCK_STATE_BLOCK', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BLOCK_STATE_UNBLOCK', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23019,
  serialized_end=23104,
)
_sym_db.RegisterEnumDescriptor(_BLOCKSTATE)

BlockState = enum_type_wrapper.EnumTypeWrapper(_BLOCKSTATE)
_REPLYTOINVITETYPE = _descriptor.EnumDescriptor(
  name='ReplyToInviteType',
  full_name='ReplyToInviteType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REPLY_TO_INVITE_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPLY_TO_INVITE_TYPE_ACCEPT', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPLY_TO_INVITE_TYPE_DECLINE', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23106,
  serialized_end=23226,
)
_sym_db.RegisterEnumDescriptor(_REPLYTOINVITETYPE)

ReplyToInviteType = enum_type_wrapper.EnumTypeWrapper(_REPLYTOINVITETYPE)
_CLIENTID = _descriptor.EnumDescriptor(
  name='ClientId',
  full_name='ClientId',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_ANDROID', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_IOS', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_CHROME', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_WEB_GPLUS', index=4, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_WEB_GMAIL', index=5, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_ULTRAVIOLET', index=6, number=13,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ID_WEB_HANGOUTS', index=7, number=44,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23229,
  serialized_end=23431,
)
_sym_db.RegisterEnumDescriptor(_CLIENTID)

ClientId = enum_type_wrapper.EnumTypeWrapper(_CLIENTID)
_CLIENTBUILDTYPE = _descriptor.EnumDescriptor(
  name='ClientBuildType',
  full_name='ClientBuildType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BUILD_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BUILD_TYPE_PRODUCTION_WEB', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BUILD_TYPE_PRODUCTION_APP', index=2, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23433,
  serialized_end=23536,
)
_sym_db.RegisterEnumDescriptor(_CLIENTBUILDTYPE)

ClientBuildType = enum_type_wrapper.EnumTypeWrapper(_CLIENTBUILDTYPE)
_RESPONSESTATUS = _descriptor.EnumDescriptor(
  name='ResponseStatus',
  full_name='ResponseStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RESPONSE_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESPONSE_STATUS_OK', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESPONSE_STATUS_UNEXPECTED_ERROR', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESPONSE_STATUS_INVALID_REQUEST', index=3, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23539,
  serialized_end=23683,
)
_sym_db.RegisterEnumDescriptor(_RESPONSESTATUS)

ResponseStatus = enum_type_wrapper.EnumTypeWrapper(_RESPONSESTATUS)
_PHOTOURLSTATUS = _descriptor.EnumDescriptor(
  name='PhotoUrlStatus',
  full_name='PhotoUrlStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PHOTO_URL_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PHOTO_URL_STATUS_PLACEHOLDER', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PHOTO_URL_STATUS_USER_PHOTO', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23685,
  serialized_end=23798,
)
_sym_db.RegisterEnumDescriptor(_PHOTOURLSTATUS)

PhotoUrlStatus = enum_type_wrapper.EnumTypeWrapper(_PHOTOURLSTATUS)
_GENDER = _descriptor.EnumDescriptor(
  name='Gender',
  full_name='Gender',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GENDER_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GENDER_MALE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GENDER_FEMALE', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23800,
  serialized_end=23864,
)
_sym_db.RegisterEnumDescriptor(_GENDER)

Gender = enum_type_wrapper.EnumTypeWrapper(_GENDER)
_PROFILETYPE = _descriptor.EnumDescriptor(
  name='ProfileType',
  full_name='ProfileType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PROFILE_TYPE_NONE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_TYPE_ES_USER', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23866,
  serialized_end=23928,
)
_sym_db.RegisterEnumDescriptor(_PROFILETYPE)

ProfileType = enum_type_wrapper.EnumTypeWrapper(_PROFILETYPE)
_CONFIGURATIONBITTYPE = _descriptor.EnumDescriptor(
  name='ConfigurationBitType',
  full_name='ConfigurationBitType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_1', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_2', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_3', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_4', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_5', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_6', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_7', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_8', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_9', index=9, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_10', index=10, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_11', index=11, number=11,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_12', index=12, number=12,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_13', index=13, number=13,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_14', index=14, number=14,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_15', index=15, number=15,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_16', index=16, number=16,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_17', index=17, number=17,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_18', index=18, number=18,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_19', index=19, number=19,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_20', index=20, number=20,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_21', index=21, number=21,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_22', index=22, number=22,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_23', index=23, number=23,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_24', index=24, number=24,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_25', index=25, number=25,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_26', index=26, number=26,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_27', index=27, number=27,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_28', index=28, number=28,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_29', index=29, number=29,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_30', index=30, number=30,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_31', index=31, number=31,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_32', index=32, number=32,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_33', index=33, number=33,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_DESKTOP_AUTO_EMOJI_CONVERSION_ENABLED', index=34, number=34,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_35', index=35, number=35,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_UNKNOWN_36', index=36, number=36,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONFIGURATION_BIT_TYPE_DESKTOP_COMPACT_MODE_ENABLED', index=37, number=38,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=23931,
  serialized_end=25468,
)
_sym_db.RegisterEnumDescriptor(_CONFIGURATIONBITTYPE)

ConfigurationBitType = enum_type_wrapper.EnumTypeWrapper(_CONFIGURATIONBITTYPE)
_RICHPRESENCETYPE = _descriptor.EnumDescriptor(
  name='RichPresenceType',
  full_name='RichPresenceType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_IN_CALL_STATE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_UNKNOWN_3', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_UNKNOWN_4', index=3, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_UNKNOWN_5', index=4, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_DEVICE', index=5, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RICH_PRESENCE_TYPE_LAST_SEEN', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=25471,
  serialized_end=25726,
)
_sym_db.RegisterEnumDescriptor(_RICHPRESENCETYPE)

RichPresenceType = enum_type_wrapper.EnumTypeWrapper(_RICHPRESENCETYPE)
_FIELDMASK = _descriptor.EnumDescriptor(
  name='FieldMask',
  full_name='FieldMask',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FIELD_MASK_REACHABLE', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIELD_MASK_AVAILABLE', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIELD_MASK_MOOD', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIELD_MASK_IN_CALL', index=3, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIELD_MASK_DEVICE', index=4, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIELD_MASK_LAST_SEEN', index=5, number=10,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=25729,
  serialized_end=25886,
)
_sym_db.RegisterEnumDescriptor(_FIELDMASK)

FieldMask = enum_type_wrapper.EnumTypeWrapper(_FIELDMASK)
_DELETETYPE = _descriptor.EnumDescriptor(
  name='DeleteType',
  full_name='DeleteType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DELETE_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DELETE_TYPE_UPPER_BOUND', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=25888,
  serialized_end=25954,
)
_sym_db.RegisterEnumDescriptor(_DELETETYPE)

DeleteType = enum_type_wrapper.EnumTypeWrapper(_DELETETYPE)
_SYNCFILTER = _descriptor.EnumDescriptor(
  name='SyncFilter',
  full_name='SyncFilter',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SYNC_FILTER_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYNC_FILTER_INBOX', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYNC_FILTER_ARCHIVED', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=25956,
  serialized_end=26042,
)
_sym_db.RegisterEnumDescriptor(_SYNCFILTER)

SyncFilter = enum_type_wrapper.EnumTypeWrapper(_SYNCFILTER)
_SOUNDSTATE = _descriptor.EnumDescriptor(
  name='SoundState',
  full_name='SoundState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SOUND_STATE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOUND_STATE_ON', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOUND_STATE_OFF', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26044,
  serialized_end=26122,
)
_sym_db.RegisterEnumDescriptor(_SOUNDSTATE)

SoundState = enum_type_wrapper.EnumTypeWrapper(_SOUNDSTATE)
_CALLERIDSETTINGSMASK = _descriptor.EnumDescriptor(
  name='CallerIdSettingsMask',
  full_name='CallerIdSettingsMask',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CALLER_ID_SETTINGS_MASK_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CALLER_ID_SETTINGS_MASK_PROVIDED', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26124,
  serialized_end=26221,
)
_sym_db.RegisterEnumDescriptor(_CALLERIDSETTINGSMASK)

CallerIdSettingsMask = enum_type_wrapper.EnumTypeWrapper(_CALLERIDSETTINGSMASK)
_PHONEVERIFICATIONSTATUS = _descriptor.EnumDescriptor(
  name='PhoneVerificationStatus',
  full_name='PhoneVerificationStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PHONE_VERIFICATION_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PHONE_VERIFICATION_STATUS_VERIFIED', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26223,
  serialized_end=26327,
)
_sym_db.RegisterEnumDescriptor(_PHONEVERIFICATIONSTATUS)

PhoneVerificationStatus = enum_type_wrapper.EnumTypeWrapper(_PHONEVERIFICATIONSTATUS)
_PHONEDISCOVERABILITYSTATUS = _descriptor.EnumDescriptor(
  name='PhoneDiscoverabilityStatus',
  full_name='PhoneDiscoverabilityStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PHONE_DISCOVERABILITY_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PHONE_DISCOVERABILITY_STATUS_OPTED_IN_BUT_NOT_DISCOVERABLE', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26330,
  serialized_end=26464,
)
_sym_db.RegisterEnumDescriptor(_PHONEDISCOVERABILITYSTATUS)

PhoneDiscoverabilityStatus = enum_type_wrapper.EnumTypeWrapper(_PHONEDISCOVERABILITYSTATUS)
_PHONEVALIDATIONRESULT = _descriptor.EnumDescriptor(
  name='PhoneValidationResult',
  full_name='PhoneValidationResult',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PHONE_VALIDATION_RESULT_IS_POSSIBLE', index=0, number=0,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26466,
  serialized_end=26530,
)
_sym_db.RegisterEnumDescriptor(_PHONEVALIDATIONRESULT)

PhoneValidationResult = enum_type_wrapper.EnumTypeWrapper(_PHONEVALIDATIONRESULT)
_OFFNETWORKADDRESSTYPE = _descriptor.EnumDescriptor(
  name='OffnetworkAddressType',
  full_name='OffnetworkAddressType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OFFNETWORK_ADDRESS_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFFNETWORK_ADDRESS_TYPE_EMAIL', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26532,
  serialized_end=26627,
)
_sym_db.RegisterEnumDescriptor(_OFFNETWORKADDRESSTYPE)

OffnetworkAddressType = enum_type_wrapper.EnumTypeWrapper(_OFFNETWORKADDRESSTYPE)
_GROUPLINKSHARINGSTATUS = _descriptor.EnumDescriptor(
  name='GroupLinkSharingStatus',
  full_name='GroupLinkSharingStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GROUP_LINK_SHARING_STATUS_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GROUP_LINK_SHARING_STATUS_ON', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GROUP_LINK_SHARING_STATUS_OFF', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=26630,
  serialized_end=26762,
)
_sym_db.RegisterEnumDescriptor(_GROUPLINKSHARINGSTATUS)

GroupLinkSharingStatus = enum_type_wrapper.EnumTypeWrapper(_GROUPLINKSHARINGSTATUS)
ACTIVE_CLIENT_STATE_NO_ACTIVE = 0
ACTIVE_CLIENT_STATE_IS_ACTIVE = 1
ACTIVE_CLIENT_STATE_OTHER_ACTIVE = 2
FOCUS_TYPE_UNKNOWN = 0
FOCUS_TYPE_FOCUSED = 1
FOCUS_TYPE_UNFOCUSED = 2
FOCUS_DEVICE_UNSPECIFIED = 0
FOCUS_DEVICE_DESKTOP = 20
FOCUS_DEVICE_MOBILE = 300
TYPING_TYPE_UNKNOWN = 0
TYPING_TYPE_STARTED = 1
TYPING_TYPE_PAUSED = 2
TYPING_TYPE_STOPPED = 3
CLIENT_PRESENCE_STATE_UNKNOWN = 0
CLIENT_PRESENCE_STATE_NONE = 1
CLIENT_PRESENCE_STATE_DESKTOP_IDLE = 30
CLIENT_PRESENCE_STATE_DESKTOP_ACTIVE = 40
NOTIFICATION_LEVEL_UNKNOWN = 0
NOTIFICATION_LEVEL_QUIET = 10
NOTIFICATION_LEVEL_RING = 30
SEGMENT_TYPE_TEXT = 0
SEGMENT_TYPE_LINE_BREAK = 1
SEGMENT_TYPE_LINK = 2
ITEM_TYPE_THING = 0
ITEM_TYPE_PLUS_PHOTO = 249
ITEM_TYPE_PLACE = 335
ITEM_TYPE_PLACE_V2 = 340
MEMBERSHIP_CHANGE_TYPE_JOIN = 1
MEMBERSHIP_CHANGE_TYPE_LEAVE = 2
HANGOUT_EVENT_TYPE_UNKNOWN = 0
HANGOUT_EVENT_TYPE_START = 1
HANGOUT_EVENT_TYPE_END = 2
HANGOUT_EVENT_TYPE_JOIN = 3
HANGOUT_EVENT_TYPE_LEAVE = 4
HANGOUT_EVENT_TYPE_COMING_SOON = 5
HANGOUT_EVENT_TYPE_ONGOING = 6
OFF_THE_RECORD_TOGGLE_UNKNOWN = 0
OFF_THE_RECORD_TOGGLE_ENABLED = 1
OFF_THE_RECORD_TOGGLE_DISABLED = 2
OFF_THE_RECORD_STATUS_UNKNOWN = 0
OFF_THE_RECORD_STATUS_OFF_THE_RECORD = 1
OFF_THE_RECORD_STATUS_ON_THE_RECORD = 2
SOURCE_TYPE_UNKNOWN = 0
EVENT_TYPE_UNKNOWN = 0
EVENT_TYPE_REGULAR_CHAT_MESSAGE = 1
EVENT_TYPE_SMS = 2
EVENT_TYPE_VOICEMAIL = 3
EVENT_TYPE_ADD_USER = 4
EVENT_TYPE_REMOVE_USER = 5
EVENT_TYPE_CONVERSATION_RENAME = 6
EVENT_TYPE_HANGOUT = 7
EVENT_TYPE_PHONE_CALL = 8
EVENT_TYPE_OTR_MODIFICATION = 9
EVENT_TYPE_PLAN_MUTATION = 10
EVENT_TYPE_MMS = 11
EVENT_TYPE_DEPRECATED_12 = 12
EVENT_TYPE_OBSERVED_EVENT = 13
EVENT_TYPE_GROUP_LINK_SHARING_MODIFICATION = 14
CONVERSATION_TYPE_UNKNOWN = 0
CONVERSATION_TYPE_ONE_TO_ONE = 1
CONVERSATION_TYPE_GROUP = 2
CONVERSATION_STATUS_UNKNOWN = 0
CONVERSATION_STATUS_INVITED = 1
CONVERSATION_STATUS_ACTIVE = 2
CONVERSATION_STATUS_LEFT = 3
CONVERSATION_VIEW_UNKNOWN = 0
CONVERSATION_VIEW_INBOX = 1
CONVERSATION_VIEW_ARCHIVED = 2
DELIVERY_MEDIUM_UNKNOWN = 0
DELIVERY_MEDIUM_BABEL = 1
DELIVERY_MEDIUM_GOOGLE_VOICE = 2
DELIVERY_MEDIUM_LOCAL_SMS = 3
INVITE_AFFINITY_UNKNOWN = 0
INVITE_AFFINITY_HIGH = 1
INVITE_AFFINITY_LOW = 2
PARTICIPANT_TYPE_UNKNOWN = 0
PARTICIPANT_TYPE_GAIA = 2
PARTICIPANT_TYPE_GOOGLE_VOICE = 3
INVITATION_STATUS_UNKNOWN = 0
INVITATION_STATUS_PENDING = 1
INVITATION_STATUS_ACCEPTED = 2
FORCE_HISTORY_UNKNOWN = 0
FORCE_HISTORY_NO = 1
NETWORK_TYPE_UNKNOWN = 0
NETWORK_TYPE_BABEL = 1
NETWORK_TYPE_GOOGLE_VOICE = 2
BLOCK_STATE_UNKNOWN = 0
BLOCK_STATE_BLOCK = 1
BLOCK_STATE_UNBLOCK = 2
REPLY_TO_INVITE_TYPE_UNKNOWN = 0
REPLY_TO_INVITE_TYPE_ACCEPT = 1
REPLY_TO_INVITE_TYPE_DECLINE = 2
CLIENT_ID_UNKNOWN = 0
CLIENT_ID_ANDROID = 1
CLIENT_ID_IOS = 2
CLIENT_ID_CHROME = 3
CLIENT_ID_WEB_GPLUS = 5
CLIENT_ID_WEB_GMAIL = 6
CLIENT_ID_ULTRAVIOLET = 13
CLIENT_ID_WEB_HANGOUTS = 44
BUILD_TYPE_UNKNOWN = 0
BUILD_TYPE_PRODUCTION_WEB = 1
BUILD_TYPE_PRODUCTION_APP = 3
RESPONSE_STATUS_UNKNOWN = 0
RESPONSE_STATUS_OK = 1
RESPONSE_STATUS_UNEXPECTED_ERROR = 3
RESPONSE_STATUS_INVALID_REQUEST = 4
PHOTO_URL_STATUS_UNKNOWN = 0
PHOTO_URL_STATUS_PLACEHOLDER = 1
PHOTO_URL_STATUS_USER_PHOTO = 2
GENDER_UNKNOWN = 0
GENDER_MALE = 1
GENDER_FEMALE = 2
PROFILE_TYPE_NONE = 0
PROFILE_TYPE_ES_USER = 1
CONFIGURATION_BIT_TYPE_UNKNOWN = 0
CONFIGURATION_BIT_TYPE_UNKNOWN_1 = 1
CONFIGURATION_BIT_TYPE_UNKNOWN_2 = 2
CONFIGURATION_BIT_TYPE_UNKNOWN_3 = 3
CONFIGURATION_BIT_TYPE_UNKNOWN_4 = 4
CONFIGURATION_BIT_TYPE_UNKNOWN_5 = 5
CONFIGURATION_BIT_TYPE_UNKNOWN_6 = 6
CONFIGURATION_BIT_TYPE_UNKNOWN_7 = 7
CONFIGURATION_BIT_TYPE_UNKNOWN_8 = 8
CONFIGURATION_BIT_TYPE_UNKNOWN_9 = 9
CONFIGURATION_BIT_TYPE_UNKNOWN_10 = 10
CONFIGURATION_BIT_TYPE_UNKNOWN_11 = 11
CONFIGURATION_BIT_TYPE_UNKNOWN_12 = 12
CONFIGURATION_BIT_TYPE_UNKNOWN_13 = 13
CONFIGURATION_BIT_TYPE_UNKNOWN_14 = 14
CONFIGURATION_BIT_TYPE_UNKNOWN_15 = 15
CONFIGURATION_BIT_TYPE_UNKNOWN_16 = 16
CONFIGURATION_BIT_TYPE_UNKNOWN_17 = 17
CONFIGURATION_BIT_TYPE_UNKNOWN_18 = 18
CONFIGURATION_BIT_TYPE_UNKNOWN_19 = 19
CONFIGURATION_BIT_TYPE_UNKNOWN_20 = 20
CONFIGURATION_BIT_TYPE_UNKNOWN_21 = 21
CONFIGURATION_BIT_TYPE_UNKNOWN_22 = 22
CONFIGURATION_BIT_TYPE_UNKNOWN_23 = 23
CONFIGURATION_BIT_TYPE_UNKNOWN_24 = 24
CONFIGURATION_BIT_TYPE_UNKNOWN_25 = 25
CONFIGURATION_BIT_TYPE_UNKNOWN_26 = 26
CONFIGURATION_BIT_TYPE_UNKNOWN_27 = 27
CONFIGURATION_BIT_TYPE_UNKNOWN_28 = 28
CONFIGURATION_BIT_TYPE_UNKNOWN_29 = 29
CONFIGURATION_BIT_TYPE_UNKNOWN_30 = 30
CONFIGURATION_BIT_TYPE_UNKNOWN_31 = 31
CONFIGURATION_BIT_TYPE_UNKNOWN_32 = 32
CONFIGURATION_BIT_TYPE_UNKNOWN_33 = 33
CONFIGURATION_BIT_TYPE_DESKTOP_AUTO_EMOJI_CONVERSION_ENABLED = 34
CONFIGURATION_BIT_TYPE_UNKNOWN_35 = 35
CONFIGURATION_BIT_TYPE_UNKNOWN_36 = 36
CONFIGURATION_BIT_TYPE_DESKTOP_COMPACT_MODE_ENABLED = 38
RICH_PRESENCE_TYPE_UNKNOWN = 0
RICH_PRESENCE_TYPE_IN_CALL_STATE = 1
RICH_PRESENCE_TYPE_UNKNOWN_3 = 3
RICH_PRESENCE_TYPE_UNKNOWN_4 = 4
RICH_PRESENCE_TYPE_UNKNOWN_5 = 5
RICH_PRESENCE_TYPE_DEVICE = 2
RICH_PRESENCE_TYPE_LAST_SEEN = 6
FIELD_MASK_REACHABLE = 1
FIELD_MASK_AVAILABLE = 2
FIELD_MASK_MOOD = 3
FIELD_MASK_IN_CALL = 6
FIELD_MASK_DEVICE = 7
FIELD_MASK_LAST_SEEN = 10
DELETE_TYPE_UNKNOWN = 0
DELETE_TYPE_UPPER_BOUND = 1
SYNC_FILTER_UNKNOWN = 0
SYNC_FILTER_INBOX = 1
SYNC_FILTER_ARCHIVED = 2
SOUND_STATE_UNKNOWN = 0
SOUND_STATE_ON = 1
SOUND_STATE_OFF = 2
CALLER_ID_SETTINGS_MASK_UNKNOWN = 0
CALLER_ID_SETTINGS_MASK_PROVIDED = 1
PHONE_VERIFICATION_STATUS_UNKNOWN = 0
PHONE_VERIFICATION_STATUS_VERIFIED = 1
PHONE_DISCOVERABILITY_STATUS_UNKNOWN = 0
PHONE_DISCOVERABILITY_STATUS_OPTED_IN_BUT_NOT_DISCOVERABLE = 2
PHONE_VALIDATION_RESULT_IS_POSSIBLE = 0
OFFNETWORK_ADDRESS_TYPE_UNKNOWN = 0
OFFNETWORK_ADDRESS_TYPE_EMAIL = 1
GROUP_LINK_SHARING_STATUS_UNKNOWN = 0
GROUP_LINK_SHARING_STATUS_ON = 1
GROUP_LINK_SHARING_STATUS_OFF = 2


_PLUSPHOTO_MEDIATYPE = _descriptor.EnumDescriptor(
  name='MediaType',
  full_name='PlusPhoto.MediaType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MEDIA_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEDIA_TYPE_PHOTO', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEDIA_TYPE_ANIMATED_PHOTO', index=2, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1474,
  serialized_end=1562,
)
_sym_db.RegisterEnumDescriptor(_PLUSPHOTO_MEDIATYPE)

_ENTITY_PASTHANGOUTSTATE = _descriptor.EnumDescriptor(
  name='PastHangoutState',
  full_name='Entity.PastHangoutState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PAST_HANGOUT_STATE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAST_HANGOUT_STATE_HAD_PAST_HANGOUT', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAST_HANGOUT_STATE_NO_PAST_HANGOUT', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=6278,
  serialized_end=6409,
)
_sym_db.RegisterEnumDescriptor(_ENTITY_PASTHANGOUTSTATE)


_DONOTDISTURBSETTING = _descriptor.Descriptor(
  name='DoNotDisturbSetting',
  full_name='DoNotDisturbSetting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='do_not_disturb', full_name='DoNotDisturbSetting.do_not_disturb', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='expiration_timestamp', full_name='DoNotDisturbSetting.expiration_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='DoNotDisturbSetting.version', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26,
  serialized_end=118,
)


_NOTIFICATIONSETTINGS = _descriptor.Descriptor(
  name='NotificationSettings',
  full_name='NotificationSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dnd_settings', full_name='NotificationSettings.dnd_settings', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=186,
)


_CONVERSATIONID = _descriptor.Descriptor(
  name='ConversationId',
  full_name='ConversationId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ConversationId.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=188,
  serialized_end=216,
)


_PARTICIPANTID = _descriptor.Descriptor(
  name='ParticipantId',
  full_name='ParticipantId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gaia_id', full_name='ParticipantId.gaia_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chat_id', full_name='ParticipantId.chat_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=218,
  serialized_end=267,
)


_DEVICESTATUS = _descriptor.Descriptor(
  name='DeviceStatus',
  full_name='DeviceStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mobile', full_name='DeviceStatus.mobile', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop', full_name='DeviceStatus.desktop', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tablet', full_name='DeviceStatus.tablet', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=269,
  serialized_end=332,
)


_LASTSEEN = _descriptor.Descriptor(
  name='LastSeen',
  full_name='LastSeen',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='last_seen_timestamp_usec', full_name='LastSeen.last_seen_timestamp_usec', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='usec_since_last_seen', full_name='LastSeen.usec_since_last_seen', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=334,
  serialized_end=408,
)


_PRESENCE = _descriptor.Descriptor(
  name='Presence',
  full_name='Presence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='reachable', full_name='Presence.reachable', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='available', full_name='Presence.available', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_status', full_name='Presence.device_status', index=2,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mood_message', full_name='Presence.mood_message', index=3,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_seen', full_name='Presence.last_seen', index=4,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=411,
  serialized_end=563,
)


_PRESENCERESULT = _descriptor.Descriptor(
  name='PresenceResult',
  full_name='PresenceResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='PresenceResult.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='presence', full_name='PresenceResult.presence', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=565,
  serialized_end=643,
)


_CLIENTIDENTIFIER = _descriptor.Descriptor(
  name='ClientIdentifier',
  full_name='ClientIdentifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='ClientIdentifier.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='header_id', full_name='ClientIdentifier.header_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=645,
  serialized_end=700,
)


_CLIENTPRESENCESTATE = _descriptor.Descriptor(
  name='ClientPresenceState',
  full_name='ClientPresenceState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='identifier', full_name='ClientPresenceState.identifier', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state', full_name='ClientPresenceState.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=702,
  serialized_end=803,
)


_USEREVENTSTATE = _descriptor.Descriptor(
  name='UserEventState',
  full_name='UserEventState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='UserEventState.user_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_generated_id', full_name='UserEventState.client_generated_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='notification_level', full_name='UserEventState.notification_level', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=805,
  serialized_end=931,
)


_FORMATTING = _descriptor.Descriptor(
  name='Formatting',
  full_name='Formatting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bold', full_name='Formatting.bold', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='italic', full_name='Formatting.italic', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='strikethrough', full_name='Formatting.strikethrough', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='underline', full_name='Formatting.underline', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=933,
  serialized_end=1017,
)


_LINKDATA = _descriptor.Descriptor(
  name='LinkData',
  full_name='LinkData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='link_target', full_name='LinkData.link_target', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1019,
  serialized_end=1050,
)


_SEGMENT = _descriptor.Descriptor(
  name='Segment',
  full_name='Segment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='Segment.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='text', full_name='Segment.text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='formatting', full_name='Segment.formatting', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='link_data', full_name='Segment.link_data', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1052,
  serialized_end=1166,
)


_PLUSPHOTO_THUMBNAIL = _descriptor.Descriptor(
  name='Thumbnail',
  full_name='PlusPhoto.Thumbnail',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='PlusPhoto.Thumbnail.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='image_url', full_name='PlusPhoto.Thumbnail.image_url', index=1,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='width_px', full_name='PlusPhoto.Thumbnail.width_px', index=2,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height_px', full_name='PlusPhoto.Thumbnail.height_px', index=3,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1392,
  serialized_end=1472,
)

_PLUSPHOTO = _descriptor.Descriptor(
  name='PlusPhoto',
  full_name='PlusPhoto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='thumbnail', full_name='PlusPhoto.thumbnail', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='owner_obfuscated_id', full_name='PlusPhoto.owner_obfuscated_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='album_id', full_name='PlusPhoto.album_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='photo_id', full_name='PlusPhoto.photo_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='url', full_name='PlusPhoto.url', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='original_content_url', full_name='PlusPhoto.original_content_url', index=5,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='media_type', full_name='PlusPhoto.media_type', index=6,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stream_id', full_name='PlusPhoto.stream_id', index=7,
      number=14, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_PLUSPHOTO_THUMBNAIL, ],
  enum_types=[
    _PLUSPHOTO_MEDIATYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1169,
  serialized_end=1562,
)


_PLACE = _descriptor.Descriptor(
  name='Place',
  full_name='Place',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='Place.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='Place.name', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='address', full_name='Place.address', index=2,
      number=24, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='geo', full_name='Place.geo', index=3,
      number=25, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='representative_image', full_name='Place.representative_image', index=4,
      number=185, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1565,
  serialized_end=1696,
)


_EMBEDITEM_POSTALADDRESS = _descriptor.Descriptor(
  name='PostalAddress',
  full_name='EmbedItem.PostalAddress',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='street_address', full_name='EmbedItem.PostalAddress.street_address', index=0,
      number=35, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1954,
  serialized_end=1993,
)

_EMBEDITEM_GEOCOORDINATES = _descriptor.Descriptor(
  name='GeoCoordinates',
  full_name='EmbedItem.GeoCoordinates',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='latitude', full_name='EmbedItem.GeoCoordinates.latitude', index=0,
      number=36, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='longitude', full_name='EmbedItem.GeoCoordinates.longitude', index=1,
      number=37, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1995,
  serialized_end=2048,
)

_EMBEDITEM_IMAGE = _descriptor.Descriptor(
  name='Image',
  full_name='EmbedItem.Image',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='EmbedItem.Image.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2050,
  serialized_end=2070,
)

_EMBEDITEM = _descriptor.Descriptor(
  name='EmbedItem',
  full_name='EmbedItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='EmbedItem.type', index=0,
      number=1, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='EmbedItem.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='plus_photo', full_name='EmbedItem.plus_photo', index=2,
      number=27639957, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='place', full_name='EmbedItem.place', index=3,
      number=35825640, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='postal_address', full_name='EmbedItem.postal_address', index=4,
      number=36003298, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='geo_coordinates', full_name='EmbedItem.geo_coordinates', index=5,
      number=36736749, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='image', full_name='EmbedItem.image', index=6,
      number=40265033, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_EMBEDITEM_POSTALADDRESS, _EMBEDITEM_GEOCOORDINATES, _EMBEDITEM_IMAGE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1699,
  serialized_end=2070,
)


_ATTACHMENT = _descriptor.Descriptor(
  name='Attachment',
  full_name='Attachment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='embed_item', full_name='Attachment.embed_item', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2072,
  serialized_end=2116,
)


_MESSAGECONTENT = _descriptor.Descriptor(
  name='MessageContent',
  full_name='MessageContent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='segment', full_name='MessageContent.segment', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attachment', full_name='MessageContent.attachment', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2118,
  serialized_end=2194,
)


_EVENTANNOTATION = _descriptor.Descriptor(
  name='EventAnnotation',
  full_name='EventAnnotation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='EventAnnotation.type', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='EventAnnotation.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2196,
  serialized_end=2242,
)


_CHATMESSAGE = _descriptor.Descriptor(
  name='ChatMessage',
  full_name='ChatMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='annotation', full_name='ChatMessage.annotation', index=0,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message_content', full_name='ChatMessage.message_content', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2244,
  serialized_end=2337,
)


_MEMBERSHIPCHANGE = _descriptor.Descriptor(
  name='MembershipChange',
  full_name='MembershipChange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='MembershipChange.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='participant_ids', full_name='MembershipChange.participant_ids', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2339,
  serialized_end=2435,
)


_CONVERSATIONRENAME = _descriptor.Descriptor(
  name='ConversationRename',
  full_name='ConversationRename',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='new_name', full_name='ConversationRename.new_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_name', full_name='ConversationRename.old_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2437,
  serialized_end=2493,
)


_HANGOUTEVENT = _descriptor.Descriptor(
  name='HangoutEvent',
  full_name='HangoutEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='event_type', full_name='HangoutEvent.event_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='HangoutEvent.participant_id', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2495,
  serialized_end=2588,
)


_OTRMODIFICATION = _descriptor.Descriptor(
  name='OTRModification',
  full_name='OTRModification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='old_otr_status', full_name='OTRModification.old_otr_status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_otr_status', full_name='OTRModification.new_otr_status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_otr_toggle', full_name='OTRModification.old_otr_toggle', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_otr_toggle', full_name='OTRModification.new_otr_toggle', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2591,
  serialized_end=2788,
)


_HASHMODIFIER = _descriptor.Descriptor(
  name='HashModifier',
  full_name='HashModifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='update_id', full_name='HashModifier.update_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash_diff', full_name='HashModifier.hash_diff', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='HashModifier.version', index=2,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2790,
  serialized_end=2859,
)


_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='Event.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sender_id', full_name='Event.sender_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='Event.timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='self_event_state', full_name='Event.self_event_state', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='source_type', full_name='Event.source_type', index=4,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chat_message', full_name='Event.chat_message', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='membership_change', full_name='Event.membership_change', index=6,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_rename', full_name='Event.conversation_rename', index=7,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hangout_event', full_name='Event.hangout_event', index=8,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_id', full_name='Event.event_id', index=9,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='expiration_timestamp', full_name='Event.expiration_timestamp', index=10,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='otr_modification', full_name='Event.otr_modification', index=11,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='advances_sort_timestamp', full_name='Event.advances_sort_timestamp', index=12,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='otr_status', full_name='Event.otr_status', index=13,
      number=16, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='persisted', full_name='Event.persisted', index=14,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='medium_type', full_name='Event.medium_type', index=15,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_type', full_name='Event.event_type', index=16,
      number=23, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_version', full_name='Event.event_version', index=17,
      number=24, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash_modifier', full_name='Event.hash_modifier', index=18,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='group_link_sharing_modification', full_name='Event.group_link_sharing_modification', index=19,
      number=31, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2862,
  serialized_end=3600,
)


_USERREADSTATE = _descriptor.Descriptor(
  name='UserReadState',
  full_name='UserReadState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='UserReadState.participant_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='latest_read_timestamp', full_name='UserReadState.latest_read_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3602,
  serialized_end=3688,
)


_DELIVERYMEDIUM = _descriptor.Descriptor(
  name='DeliveryMedium',
  full_name='DeliveryMedium',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='medium_type', full_name='DeliveryMedium.medium_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phone_number', full_name='DeliveryMedium.phone_number', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3690,
  serialized_end=3784,
)


_DELIVERYMEDIUMOPTION = _descriptor.Descriptor(
  name='DeliveryMediumOption',
  full_name='DeliveryMediumOption',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='delivery_medium', full_name='DeliveryMediumOption.delivery_medium', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='current_default', full_name='DeliveryMediumOption.current_default', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3786,
  serialized_end=3875,
)


_USERCONVERSATIONSTATE = _descriptor.Descriptor(
  name='UserConversationState',
  full_name='UserConversationState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_generated_id', full_name='UserConversationState.client_generated_id', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='self_read_state', full_name='UserConversationState.self_read_state', index=1,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='status', full_name='UserConversationState.status', index=2,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='notification_level', full_name='UserConversationState.notification_level', index=3,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='view', full_name='UserConversationState.view', index=4,
      number=10, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inviter_id', full_name='UserConversationState.inviter_id', index=5,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invite_timestamp', full_name='UserConversationState.invite_timestamp', index=6,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sort_timestamp', full_name='UserConversationState.sort_timestamp', index=7,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='active_timestamp', full_name='UserConversationState.active_timestamp', index=8,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invite_affinity', full_name='UserConversationState.invite_affinity', index=9,
      number=15, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delivery_medium_option', full_name='UserConversationState.delivery_medium_option', index=10,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3878,
  serialized_end=4302,
)


_CONVERSATIONPARTICIPANTDATA = _descriptor.Descriptor(
  name='ConversationParticipantData',
  full_name='ConversationParticipantData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ConversationParticipantData.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fallback_name', full_name='ConversationParticipantData.fallback_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invitation_status', full_name='ConversationParticipantData.invitation_status', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='participant_type', full_name='ConversationParticipantData.participant_type', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_invitation_status', full_name='ConversationParticipantData.new_invitation_status', index=4,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4305,
  serialized_end=4525,
)


_CONVERSATION = _descriptor.Descriptor(
  name='Conversation',
  full_name='Conversation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='Conversation.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='Conversation.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='Conversation.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='self_conversation_state', full_name='Conversation.self_conversation_state', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='read_state', full_name='Conversation.read_state', index=4,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='has_active_hangout', full_name='Conversation.has_active_hangout', index=5,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='otr_status', full_name='Conversation.otr_status', index=6,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='otr_toggle', full_name='Conversation.otr_toggle', index=7,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_history_supported', full_name='Conversation.conversation_history_supported', index=8,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='current_participant', full_name='Conversation.current_participant', index=9,
      number=13, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='participant_data', full_name='Conversation.participant_data', index=10,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='network_type', full_name='Conversation.network_type', index=11,
      number=18, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='force_history_state', full_name='Conversation.force_history_state', index=12,
      number=19, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='group_link_sharing_status', full_name='Conversation.group_link_sharing_status', index=13,
      number=22, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4528,
  serialized_end=5115,
)


_EASTEREGG = _descriptor.Descriptor(
  name='EasterEgg',
  full_name='EasterEgg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='EasterEgg.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5117,
  serialized_end=5145,
)


_BLOCKSTATECHANGE = _descriptor.Descriptor(
  name='BlockStateChange',
  full_name='BlockStateChange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='BlockStateChange.participant_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_block_state', full_name='BlockStateChange.new_block_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5147,
  serialized_end=5243,
)


_PHOTO = _descriptor.Descriptor(
  name='Photo',
  full_name='Photo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='photo_id', full_name='Photo.photo_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_albumless_source_photo', full_name='Photo.delete_albumless_source_photo', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='Photo.user_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_custom_user_id', full_name='Photo.is_custom_user_id', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5245,
  serialized_end=5353,
)


_EXISTINGMEDIA = _descriptor.Descriptor(
  name='ExistingMedia',
  full_name='ExistingMedia',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='photo', full_name='ExistingMedia.photo', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5355,
  serialized_end=5393,
)


_EVENTREQUESTHEADER = _descriptor.Descriptor(
  name='EventRequestHeader',
  full_name='EventRequestHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='EventRequestHeader.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_generated_id', full_name='EventRequestHeader.client_generated_id', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='expected_otr', full_name='EventRequestHeader.expected_otr', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delivery_medium', full_name='EventRequestHeader.delivery_medium', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_type', full_name='EventRequestHeader.event_type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5396,
  serialized_end=5604,
)


_CLIENTVERSION = _descriptor.Descriptor(
  name='ClientVersion',
  full_name='ClientVersion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_id', full_name='ClientVersion.client_id', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='build_type', full_name='ClientVersion.build_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='major_version', full_name='ClientVersion.major_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version_timestamp', full_name='ClientVersion.version_timestamp', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_os_version', full_name='ClientVersion.device_os_version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device_hardware', full_name='ClientVersion.device_hardware', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5607,
  serialized_end=5792,
)


_REQUESTHEADER = _descriptor.Descriptor(
  name='RequestHeader',
  full_name='RequestHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_version', full_name='RequestHeader.client_version', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_identifier', full_name='RequestHeader.client_identifier', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='language_code', full_name='RequestHeader.language_code', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5794,
  serialized_end=5918,
)


_RESPONSEHEADER = _descriptor.Descriptor(
  name='ResponseHeader',
  full_name='ResponseHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='ResponseHeader.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='error_description', full_name='ResponseHeader.error_description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug_url', full_name='ResponseHeader.debug_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='request_trace_id', full_name='ResponseHeader.request_trace_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='current_server_time', full_name='ResponseHeader.current_server_time', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5921,
  serialized_end=6071,
)


_ENTITY = _descriptor.Descriptor(
  name='Entity',
  full_name='Entity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Entity.id', index=0,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='presence', full_name='Entity.presence', index=1,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='properties', full_name='Entity.properties', index=2,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity_type', full_name='Entity.entity_type', index=3,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='had_past_hangout_state', full_name='Entity.had_past_hangout_state', index=4,
      number=16, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ENTITY_PASTHANGOUTSTATE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6074,
  serialized_end=6409,
)


_ENTITYPROPERTIES = _descriptor.Descriptor(
  name='EntityProperties',
  full_name='EntityProperties',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='EntityProperties.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='EntityProperties.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='EntityProperties.first_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='photo_url', full_name='EntityProperties.photo_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='email', full_name='EntityProperties.email', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phone', full_name='EntityProperties.phone', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='in_users_domain', full_name='EntityProperties.in_users_domain', index=6,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='gender', full_name='EntityProperties.gender', index=7,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='photo_url_status', full_name='EntityProperties.photo_url_status', index=8,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='canonical_email', full_name='EntityProperties.canonical_email', index=9,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6412,
  serialized_end=6667,
)


_CONVERSATIONSTATE = _descriptor.Descriptor(
  name='ConversationState',
  full_name='ConversationState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='ConversationState.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation', full_name='ConversationState.conversation', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event', full_name='ConversationState.event', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_continuation_token', full_name='ConversationState.event_continuation_token', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6670,
  serialized_end=6850,
)


_EVENTCONTINUATIONTOKEN = _descriptor.Descriptor(
  name='EventContinuationToken',
  full_name='EventContinuationToken',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='event_id', full_name='EventContinuationToken.event_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='storage_continuation_token', full_name='EventContinuationToken.storage_continuation_token', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_timestamp', full_name='EventContinuationToken.event_timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6852,
  serialized_end=6955,
)


_ENTITYLOOKUPSPEC = _descriptor.Descriptor(
  name='EntityLookupSpec',
  full_name='EntityLookupSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gaia_id', full_name='EntityLookupSpec.gaia_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='email', full_name='EntityLookupSpec.email', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phone', full_name='EntityLookupSpec.phone', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='create_offnetwork_gaia', full_name='EntityLookupSpec.create_offnetwork_gaia', index=3,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6957,
  serialized_end=7054,
)


_CONFIGURATIONBIT = _descriptor.Descriptor(
  name='ConfigurationBit',
  full_name='ConfigurationBit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='configuration_bit_type', full_name='ConfigurationBit.configuration_bit_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='ConfigurationBit.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7056,
  serialized_end=7144,
)


_RICHPRESENCESTATE = _descriptor.Descriptor(
  name='RichPresenceState',
  full_name='RichPresenceState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='get_rich_presence_enabled_state', full_name='RichPresenceState.get_rich_presence_enabled_state', index=0,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7146,
  serialized_end=7233,
)


_RICHPRESENCEENABLEDSTATE = _descriptor.Descriptor(
  name='RichPresenceEnabledState',
  full_name='RichPresenceEnabledState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='RichPresenceEnabledState.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='enabled', full_name='RichPresenceEnabledState.enabled', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7235,
  serialized_end=7311,
)


_DESKTOPOFFSETTING = _descriptor.Descriptor(
  name='DesktopOffSetting',
  full_name='DesktopOffSetting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='desktop_off', full_name='DesktopOffSetting.desktop_off', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7313,
  serialized_end=7353,
)


_DESKTOPOFFSTATE = _descriptor.Descriptor(
  name='DesktopOffState',
  full_name='DesktopOffState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='desktop_off', full_name='DesktopOffState.desktop_off', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='DesktopOffState.version', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7355,
  serialized_end=7410,
)


_DNDSETTING = _descriptor.Descriptor(
  name='DndSetting',
  full_name='DndSetting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='do_not_disturb', full_name='DndSetting.do_not_disturb', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeout_secs', full_name='DndSetting.timeout_secs', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7412,
  serialized_end=7470,
)


_PRESENCESTATESETTING = _descriptor.Descriptor(
  name='PresenceStateSetting',
  full_name='PresenceStateSetting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timeout_secs', full_name='PresenceStateSetting.timeout_secs', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='PresenceStateSetting.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7472,
  serialized_end=7556,
)


_MOODMESSAGE = _descriptor.Descriptor(
  name='MoodMessage',
  full_name='MoodMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mood_content', full_name='MoodMessage.mood_content', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7558,
  serialized_end=7607,
)


_MOODCONTENT = _descriptor.Descriptor(
  name='MoodContent',
  full_name='MoodContent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='segment', full_name='MoodContent.segment', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7609,
  serialized_end=7649,
)


_MOODSETTING = _descriptor.Descriptor(
  name='MoodSetting',
  full_name='MoodSetting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mood_message', full_name='MoodSetting.mood_message', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7651,
  serialized_end=7700,
)


_MOODSTATE = _descriptor.Descriptor(
  name='MoodState',
  full_name='MoodState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mood_setting', full_name='MoodState.mood_setting', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7702,
  serialized_end=7749,
)


_DELETEACTION = _descriptor.Descriptor(
  name='DeleteAction',
  full_name='DeleteAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='delete_action_timestamp', full_name='DeleteAction.delete_action_timestamp', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_upper_bound_timestamp', full_name='DeleteAction.delete_upper_bound_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_type', full_name='DeleteAction.delete_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7751,
  serialized_end=7870,
)


_INVITEEID = _descriptor.Descriptor(
  name='InviteeID',
  full_name='InviteeID',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gaia_id', full_name='InviteeID.gaia_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fallback_name', full_name='InviteeID.fallback_name', index=1,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7872,
  serialized_end=7923,
)


_COUNTRY = _descriptor.Descriptor(
  name='Country',
  full_name='Country',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='region_code', full_name='Country.region_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='country_code', full_name='Country.country_code', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7925,
  serialized_end=7977,
)


_DESKTOPSOUNDSETTING = _descriptor.Descriptor(
  name='DesktopSoundSetting',
  full_name='DesktopSoundSetting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='desktop_sound_state', full_name='DesktopSoundSetting.desktop_sound_state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_ring_sound_state', full_name='DesktopSoundSetting.desktop_ring_sound_state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7979,
  serialized_end=8089,
)


_PHONEDATA = _descriptor.Descriptor(
  name='PhoneData',
  full_name='PhoneData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='phone', full_name='PhoneData.phone', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='caller_id_settings_mask', full_name='PhoneData.caller_id_settings_mask', index=1,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8091,
  serialized_end=8181,
)


_PHONE = _descriptor.Descriptor(
  name='Phone',
  full_name='Phone',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='phone_number', full_name='Phone.phone_number', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='google_voice', full_name='Phone.google_voice', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='verification_status', full_name='Phone.verification_status', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='discoverable', full_name='Phone.discoverable', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='discoverability_status', full_name='Phone.discoverability_status', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='primary', full_name='Phone.primary', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8184,
  serialized_end=8404,
)


_I18NDATA = _descriptor.Descriptor(
  name='I18nData',
  full_name='I18nData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='national_number', full_name='I18nData.national_number', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='international_number', full_name='I18nData.international_number', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='country_code', full_name='I18nData.country_code', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='region_code', full_name='I18nData.region_code', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_valid', full_name='I18nData.is_valid', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validation_result', full_name='I18nData.validation_result', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8407,
  serialized_end=8584,
)


_PHONENUMBER = _descriptor.Descriptor(
  name='PhoneNumber',
  full_name='PhoneNumber',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='e164', full_name='PhoneNumber.e164', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='i18n_data', full_name='PhoneNumber.i18n_data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8586,
  serialized_end=8643,
)


_SUGGESTEDCONTACTGROUPHASH = _descriptor.Descriptor(
  name='SuggestedContactGroupHash',
  full_name='SuggestedContactGroupHash',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='max_results', full_name='SuggestedContactGroupHash.max_results', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='SuggestedContactGroupHash.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8645,
  serialized_end=8707,
)


_SUGGESTEDCONTACT = _descriptor.Descriptor(
  name='SuggestedContact',
  full_name='SuggestedContact',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entity', full_name='SuggestedContact.entity', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invitation_status', full_name='SuggestedContact.invitation_status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8709,
  serialized_end=8798,
)


_SUGGESTEDCONTACTGROUP = _descriptor.Descriptor(
  name='SuggestedContactGroup',
  full_name='SuggestedContactGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_matched', full_name='SuggestedContactGroup.hash_matched', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='SuggestedContactGroup.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='contact', full_name='SuggestedContactGroup.contact', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8800,
  serialized_end=8895,
)


_GROUPLINKSHARINGMODIFICATION = _descriptor.Descriptor(
  name='GroupLinkSharingModification',
  full_name='GroupLinkSharingModification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='new_status', full_name='GroupLinkSharingModification.new_status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8897,
  serialized_end=8972,
)


_STATEUPDATE = _descriptor.Descriptor(
  name='StateUpdate',
  full_name='StateUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state_update_header', full_name='StateUpdate.state_update_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation', full_name='StateUpdate.conversation', index=1,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_notification', full_name='StateUpdate.conversation_notification', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_notification', full_name='StateUpdate.event_notification', index=3,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='focus_notification', full_name='StateUpdate.focus_notification', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='typing_notification', full_name='StateUpdate.typing_notification', index=5,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='notification_level_notification', full_name='StateUpdate.notification_level_notification', index=6,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reply_to_invite_notification', full_name='StateUpdate.reply_to_invite_notification', index=7,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='watermark_notification', full_name='StateUpdate.watermark_notification', index=8,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='view_modification', full_name='StateUpdate.view_modification', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='easter_egg_notification', full_name='StateUpdate.easter_egg_notification', index=10,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='self_presence_notification', full_name='StateUpdate.self_presence_notification', index=11,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_notification', full_name='StateUpdate.delete_notification', index=12,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='presence_notification', full_name='StateUpdate.presence_notification', index=13,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='block_notification', full_name='StateUpdate.block_notification', index=14,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='notification_setting_notification', full_name='StateUpdate.notification_setting_notification', index=15,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rich_presence_enabled_state_notification', full_name='StateUpdate.rich_presence_enabled_state_notification', index=16,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='state_update', full_name='StateUpdate.state_update',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=8975,
  serialized_end=10049,
)


_STATEUPDATEHEADER = _descriptor.Descriptor(
  name='StateUpdateHeader',
  full_name='StateUpdateHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='active_client_state', full_name='StateUpdateHeader.active_client_state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='request_trace_id', full_name='StateUpdateHeader.request_trace_id', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='notification_settings', full_name='StateUpdateHeader.notification_settings', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='current_server_time', full_name='StateUpdateHeader.current_server_time', index=3,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10052,
  serialized_end=10229,
)


_BATCHUPDATE = _descriptor.Descriptor(
  name='BatchUpdate',
  full_name='BatchUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state_update', full_name='BatchUpdate.state_update', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10231,
  serialized_end=10280,
)


_CONVERSATIONNOTIFICATION = _descriptor.Descriptor(
  name='ConversationNotification',
  full_name='ConversationNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation', full_name='ConversationNotification.conversation', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10282,
  serialized_end=10345,
)


_EVENTNOTIFICATION = _descriptor.Descriptor(
  name='EventNotification',
  full_name='EventNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='event', full_name='EventNotification.event', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10347,
  serialized_end=10389,
)


_SETFOCUSNOTIFICATION = _descriptor.Descriptor(
  name='SetFocusNotification',
  full_name='SetFocusNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='SetFocusNotification.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sender_id', full_name='SetFocusNotification.sender_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SetFocusNotification.timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='SetFocusNotification.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='device', full_name='SetFocusNotification.device', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10392,
  serialized_end=10566,
)


_SETTYPINGNOTIFICATION = _descriptor.Descriptor(
  name='SetTypingNotification',
  full_name='SetTypingNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='SetTypingNotification.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sender_id', full_name='SetTypingNotification.sender_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SetTypingNotification.timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='SetTypingNotification.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10569,
  serialized_end=10715,
)


_SETCONVERSATIONNOTIFICATIONLEVELNOTIFICATION = _descriptor.Descriptor(
  name='SetConversationNotificationLevelNotification',
  full_name='SetConversationNotificationLevelNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='SetConversationNotificationLevelNotification.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='level', full_name='SetConversationNotificationLevelNotification.level', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SetConversationNotificationLevelNotification.timestamp', index=2,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10718,
  serialized_end=10860,
)


_REPLYTOINVITENOTIFICATION = _descriptor.Descriptor(
  name='ReplyToInviteNotification',
  full_name='ReplyToInviteNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='ReplyToInviteNotification.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='ReplyToInviteNotification.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10862,
  serialized_end=10965,
)


_WATERMARKNOTIFICATION = _descriptor.Descriptor(
  name='WatermarkNotification',
  full_name='WatermarkNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender_id', full_name='WatermarkNotification.sender_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='WatermarkNotification.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='latest_read_timestamp', full_name='WatermarkNotification.latest_read_timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10968,
  serialized_end=11099,
)


_CONVERSATIONVIEWMODIFICATION = _descriptor.Descriptor(
  name='ConversationViewModification',
  full_name='ConversationViewModification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='ConversationViewModification.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_view', full_name='ConversationViewModification.old_view', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_view', full_name='ConversationViewModification.new_view', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11102,
  serialized_end=11248,
)


_EASTEREGGNOTIFICATION = _descriptor.Descriptor(
  name='EasterEggNotification',
  full_name='EasterEggNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender_id', full_name='EasterEggNotification.sender_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='EasterEggNotification.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='easter_egg', full_name='EasterEggNotification.easter_egg', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11251,
  serialized_end=11383,
)


_SELFPRESENCENOTIFICATION = _descriptor.Descriptor(
  name='SelfPresenceNotification',
  full_name='SelfPresenceNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_presence_state', full_name='SelfPresenceNotification.client_presence_state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='do_not_disturb_setting', full_name='SelfPresenceNotification.do_not_disturb_setting', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_off_setting', full_name='SelfPresenceNotification.desktop_off_setting', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_off_state', full_name='SelfPresenceNotification.desktop_off_state', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mood_state', full_name='SelfPresenceNotification.mood_state', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11386,
  serialized_end=11645,
)


_DELETEACTIONNOTIFICATION = _descriptor.Descriptor(
  name='DeleteActionNotification',
  full_name='DeleteActionNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='DeleteActionNotification.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_action', full_name='DeleteActionNotification.delete_action', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11647,
  serialized_end=11753,
)


_PRESENCENOTIFICATION = _descriptor.Descriptor(
  name='PresenceNotification',
  full_name='PresenceNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='presence', full_name='PresenceNotification.presence', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11755,
  serialized_end=11812,
)


_BLOCKNOTIFICATION = _descriptor.Descriptor(
  name='BlockNotification',
  full_name='BlockNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_state_change', full_name='BlockNotification.block_state_change', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11814,
  serialized_end=11880,
)


_SETNOTIFICATIONSETTINGNOTIFICATION = _descriptor.Descriptor(
  name='SetNotificationSettingNotification',
  full_name='SetNotificationSettingNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='configuration_bit', full_name='SetNotificationSettingNotification.configuration_bit', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_sound_setting', full_name='SetNotificationSettingNotification.desktop_sound_setting', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11883,
  serialized_end=12018,
)


_RICHPRESENCEENABLEDSTATENOTIFICATION = _descriptor.Descriptor(
  name='RichPresenceEnabledStateNotification',
  full_name='RichPresenceEnabledStateNotification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rich_presence_enabled_state', full_name='RichPresenceEnabledStateNotification.rich_presence_enabled_state', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12020,
  serialized_end=12122,
)


_CONVERSATIONSPEC = _descriptor.Descriptor(
  name='ConversationSpec',
  full_name='ConversationSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='ConversationSpec.conversation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12124,
  serialized_end=12184,
)


_OFFNETWORKADDRESS = _descriptor.Descriptor(
  name='OffnetworkAddress',
  full_name='OffnetworkAddress',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='OffnetworkAddress.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='email', full_name='OffnetworkAddress.email', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12186,
  serialized_end=12258,
)


_ENTITYRESULT = _descriptor.Descriptor(
  name='EntityResult',
  full_name='EntityResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lookup_spec', full_name='EntityResult.lookup_spec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity', full_name='EntityResult.entity', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12260,
  serialized_end=12339,
)


_ADDUSERREQUEST = _descriptor.Descriptor(
  name='AddUserRequest',
  full_name='AddUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='AddUserRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invitee_id', full_name='AddUserRequest.invitee_id', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_request_header', full_name='AddUserRequest.event_request_header', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12342,
  serialized_end=12481,
)


_ADDUSERRESPONSE = _descriptor.Descriptor(
  name='AddUserResponse',
  full_name='AddUserResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='AddUserResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_event', full_name='AddUserResponse.created_event', index=1,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12483,
  serialized_end=12573,
)


_CREATECONVERSATIONREQUEST = _descriptor.Descriptor(
  name='CreateConversationRequest',
  full_name='CreateConversationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='CreateConversationRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='CreateConversationRequest.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_generated_id', full_name='CreateConversationRequest.client_generated_id', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='CreateConversationRequest.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invitee_id', full_name='CreateConversationRequest.invitee_id', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12576,
  serialized_end=12751,
)


_CREATECONVERSATIONRESPONSE = _descriptor.Descriptor(
  name='CreateConversationResponse',
  full_name='CreateConversationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='CreateConversationResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation', full_name='CreateConversationResponse.conversation', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_conversation_created', full_name='CreateConversationResponse.new_conversation_created', index=2,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12754,
  serialized_end=12895,
)


_DELETECONVERSATIONREQUEST = _descriptor.Descriptor(
  name='DeleteConversationRequest',
  full_name='DeleteConversationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='DeleteConversationRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='DeleteConversationRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_upper_bound_timestamp', full_name='DeleteConversationRequest.delete_upper_bound_timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12898,
  serialized_end=13045,
)


_DELETECONVERSATIONRESPONSE = _descriptor.Descriptor(
  name='DeleteConversationResponse',
  full_name='DeleteConversationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='DeleteConversationResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='delete_action', full_name='DeleteConversationResponse.delete_action', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13047,
  serialized_end=13155,
)


_EASTEREGGREQUEST = _descriptor.Descriptor(
  name='EasterEggRequest',
  full_name='EasterEggRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='EasterEggRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='EasterEggRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='easter_egg', full_name='EasterEggRequest.easter_egg', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13158,
  serialized_end=13290,
)


_EASTEREGGRESPONSE = _descriptor.Descriptor(
  name='EasterEggResponse',
  full_name='EasterEggResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='EasterEggResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='EasterEggResponse.timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13292,
  serialized_end=13372,
)


_GETCONVERSATIONREQUEST = _descriptor.Descriptor(
  name='GetConversationRequest',
  full_name='GetConversationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetConversationRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_spec', full_name='GetConversationRequest.conversation_spec', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='include_conversation_metadata', full_name='GetConversationRequest.include_conversation_metadata', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='include_event', full_name='GetConversationRequest.include_event', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_events_per_conversation', full_name='GetConversationRequest.max_events_per_conversation', index=4,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_continuation_token', full_name='GetConversationRequest.event_continuation_token', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13375,
  serialized_end=13643,
)


_GETCONVERSATIONRESPONSE = _descriptor.Descriptor(
  name='GetConversationResponse',
  full_name='GetConversationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='GetConversationResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_state', full_name='GetConversationResponse.conversation_state', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13645,
  serialized_end=13760,
)


_GETENTITYBYIDREQUEST = _descriptor.Descriptor(
  name='GetEntityByIdRequest',
  full_name='GetEntityByIdRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetEntityByIdRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='batch_lookup_spec', full_name='GetEntityByIdRequest.batch_lookup_spec', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13762,
  serialized_end=13870,
)


_GETENTITYBYIDRESPONSE = _descriptor.Descriptor(
  name='GetEntityByIdResponse',
  full_name='GetEntityByIdResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='GetEntityByIdResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity', full_name='GetEntityByIdResponse.entity', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity_result', full_name='GetEntityByIdResponse.entity_result', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13873,
  serialized_end=14001,
)


_GETGROUPCONVERSATIONURLREQUEST = _descriptor.Descriptor(
  name='GetGroupConversationUrlRequest',
  full_name='GetGroupConversationUrlRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetGroupConversationUrlRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='GetGroupConversationUrlRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14003,
  serialized_end=14117,
)


_GETGROUPCONVERSATIONURLRESPONSE = _descriptor.Descriptor(
  name='GetGroupConversationUrlResponse',
  full_name='GetGroupConversationUrlResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='GetGroupConversationUrlResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='group_conversation_url', full_name='GetGroupConversationUrlResponse.group_conversation_url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14119,
  serialized_end=14226,
)


_GETSUGGESTEDENTITIESREQUEST = _descriptor.Descriptor(
  name='GetSuggestedEntitiesRequest',
  full_name='GetSuggestedEntitiesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetSuggestedEntitiesRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_count', full_name='GetSuggestedEntitiesRequest.max_count', index=1,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='favorites', full_name='GetSuggestedEntitiesRequest.favorites', index=2,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='contacts_you_hangout_with', full_name='GetSuggestedEntitiesRequest.contacts_you_hangout_with', index=3,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='other_contacts_on_hangouts', full_name='GetSuggestedEntitiesRequest.other_contacts_on_hangouts', index=4,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='other_contacts', full_name='GetSuggestedEntitiesRequest.other_contacts', index=5,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dismissed_contacts', full_name='GetSuggestedEntitiesRequest.dismissed_contacts', index=6,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pinned_favorites', full_name='GetSuggestedEntitiesRequest.pinned_favorites', index=7,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14229,
  serialized_end=14653,
)


_GETSUGGESTEDENTITIESRESPONSE = _descriptor.Descriptor(
  name='GetSuggestedEntitiesResponse',
  full_name='GetSuggestedEntitiesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='GetSuggestedEntitiesResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity', full_name='GetSuggestedEntitiesResponse.entity', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='favorites', full_name='GetSuggestedEntitiesResponse.favorites', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='contacts_you_hangout_with', full_name='GetSuggestedEntitiesResponse.contacts_you_hangout_with', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='other_contacts_on_hangouts', full_name='GetSuggestedEntitiesResponse.other_contacts_on_hangouts', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='other_contacts', full_name='GetSuggestedEntitiesResponse.other_contacts', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dismissed_contacts', full_name='GetSuggestedEntitiesResponse.dismissed_contacts', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pinned_favorites', full_name='GetSuggestedEntitiesResponse.pinned_favorites', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14656,
  serialized_end=15065,
)


_GETSELFINFOREQUEST = _descriptor.Descriptor(
  name='GetSelfInfoRequest',
  full_name='GetSelfInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='GetSelfInfoRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15067,
  serialized_end=15127,
)


_GETSELFINFORESPONSE = _descriptor.Descriptor(
  name='GetSelfInfoResponse',
  full_name='GetSelfInfoResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='GetSelfInfoResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='self_entity', full_name='GetSelfInfoResponse.self_entity', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_known_minor', full_name='GetSelfInfoResponse.is_known_minor', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dnd_state', full_name='GetSelfInfoResponse.dnd_state', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_off_setting', full_name='GetSelfInfoResponse.desktop_off_setting', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phone_data', full_name='GetSelfInfoResponse.phone_data', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='configuration_bit', full_name='GetSelfInfoResponse.configuration_bit', index=6,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_off_state', full_name='GetSelfInfoResponse.desktop_off_state', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='google_plus_user', full_name='GetSelfInfoResponse.google_plus_user', index=8,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_sound_setting', full_name='GetSelfInfoResponse.desktop_sound_setting', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rich_presence_state', full_name='GetSelfInfoResponse.rich_presence_state', index=10,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='default_country', full_name='GetSelfInfoResponse.default_country', index=11,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15130,
  serialized_end=15623,
)


_QUERYPRESENCEREQUEST = _descriptor.Descriptor(
  name='QueryPresenceRequest',
  full_name='QueryPresenceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='QueryPresenceRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='QueryPresenceRequest.participant_id', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='field_mask', full_name='QueryPresenceRequest.field_mask', index=2,
      number=3, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15626,
  serialized_end=15760,
)


_QUERYPRESENCERESPONSE = _descriptor.Descriptor(
  name='QueryPresenceResponse',
  full_name='QueryPresenceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='QueryPresenceResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='presence_result', full_name='QueryPresenceResponse.presence_result', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15762,
  serialized_end=15869,
)


_REMOVEUSERREQUEST = _descriptor.Descriptor(
  name='RemoveUserRequest',
  full_name='RemoveUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='RemoveUserRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='RemoveUserRequest.participant_id', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_request_header', full_name='RemoveUserRequest.event_request_header', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15872,
  serialized_end=16022,
)


_REMOVEUSERRESPONSE = _descriptor.Descriptor(
  name='RemoveUserResponse',
  full_name='RemoveUserResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='RemoveUserResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_event', full_name='RemoveUserResponse.created_event', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16024,
  serialized_end=16117,
)


_RENAMECONVERSATIONREQUEST = _descriptor.Descriptor(
  name='RenameConversationRequest',
  full_name='RenameConversationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='RenameConversationRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='new_name', full_name='RenameConversationRequest.new_name', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_request_header', full_name='RenameConversationRequest.event_request_header', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16120,
  serialized_end=16256,
)


_RENAMECONVERSATIONRESPONSE = _descriptor.Descriptor(
  name='RenameConversationResponse',
  full_name='RenameConversationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='RenameConversationResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_event', full_name='RenameConversationResponse.created_event', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16258,
  serialized_end=16359,
)


_SEARCHENTITIESREQUEST = _descriptor.Descriptor(
  name='SearchEntitiesRequest',
  full_name='SearchEntitiesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SearchEntitiesRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='query', full_name='SearchEntitiesRequest.query', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_count', full_name='SearchEntitiesRequest.max_count', index=2,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16361,
  serialized_end=16458,
)


_SEARCHENTITIESRESPONSE = _descriptor.Descriptor(
  name='SearchEntitiesResponse',
  full_name='SearchEntitiesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SearchEntitiesResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity', full_name='SearchEntitiesResponse.entity', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16460,
  serialized_end=16551,
)


_LOCATION = _descriptor.Descriptor(
  name='Location',
  full_name='Location',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='place', full_name='Location.place', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16553,
  serialized_end=16586,
)


_SENDCHATMESSAGEREQUEST = _descriptor.Descriptor(
  name='SendChatMessageRequest',
  full_name='SendChatMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SendChatMessageRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='annotation', full_name='SendChatMessageRequest.annotation', index=1,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message_content', full_name='SendChatMessageRequest.message_content', index=2,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='existing_media', full_name='SendChatMessageRequest.existing_media', index=3,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_request_header', full_name='SendChatMessageRequest.event_request_header', index=4,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='SendChatMessageRequest.user_id', index=5,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='location', full_name='SendChatMessageRequest.location', index=6,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16589,
  serialized_end=16886,
)


_SENDCHATMESSAGERESPONSE = _descriptor.Descriptor(
  name='SendChatMessageResponse',
  full_name='SendChatMessageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SendChatMessageResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_event', full_name='SendChatMessageResponse.created_event', index=1,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16888,
  serialized_end=16986,
)


_MODIFYOTRSTATUSREQUEST = _descriptor.Descriptor(
  name='ModifyOTRStatusRequest',
  full_name='ModifyOTRStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='ModifyOTRStatusRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='otr_status', full_name='ModifyOTRStatusRequest.otr_status', index=1,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_request_header', full_name='ModifyOTRStatusRequest.event_request_header', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16989,
  serialized_end=17145,
)


_MODIFYOTRSTATUSRESPONSE = _descriptor.Descriptor(
  name='ModifyOTRStatusResponse',
  full_name='ModifyOTRStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='ModifyOTRStatusResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_event', full_name='ModifyOTRStatusResponse.created_event', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17147,
  serialized_end=17245,
)


_SENDOFFNETWORKINVITATIONREQUEST = _descriptor.Descriptor(
  name='SendOffnetworkInvitationRequest',
  full_name='SendOffnetworkInvitationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SendOffnetworkInvitationRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='invitee_address', full_name='SendOffnetworkInvitationRequest.invitee_address', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17247,
  serialized_end=17365,
)


_SENDOFFNETWORKINVITATIONRESPONSE = _descriptor.Descriptor(
  name='SendOffnetworkInvitationResponse',
  full_name='SendOffnetworkInvitationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SendOffnetworkInvitationResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17367,
  serialized_end=17443,
)


_SETACTIVECLIENTREQUEST = _descriptor.Descriptor(
  name='SetActiveClientRequest',
  full_name='SetActiveClientRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetActiveClientRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_active', full_name='SetActiveClientRequest.is_active', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='full_jid', full_name='SetActiveClientRequest.full_jid', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeout_secs', full_name='SetActiveClientRequest.timeout_secs', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17445,
  serialized_end=17568,
)


_SETACTIVECLIENTRESPONSE = _descriptor.Descriptor(
  name='SetActiveClientResponse',
  full_name='SetActiveClientResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetActiveClientResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17570,
  serialized_end=17637,
)


_SETCONVERSATIONLEVELREQUEST = _descriptor.Descriptor(
  name='SetConversationLevelRequest',
  full_name='SetConversationLevelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetConversationLevelRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17639,
  serialized_end=17708,
)


_SETCONVERSATIONLEVELRESPONSE = _descriptor.Descriptor(
  name='SetConversationLevelResponse',
  full_name='SetConversationLevelResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetConversationLevelResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17710,
  serialized_end=17782,
)


_SETCONVERSATIONNOTIFICATIONLEVELREQUEST = _descriptor.Descriptor(
  name='SetConversationNotificationLevelRequest',
  full_name='SetConversationNotificationLevelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetConversationNotificationLevelRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='SetConversationNotificationLevelRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='level', full_name='SetConversationNotificationLevelRequest.level', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17785,
  serialized_end=17943,
)


_SETCONVERSATIONNOTIFICATIONLEVELRESPONSE = _descriptor.Descriptor(
  name='SetConversationNotificationLevelResponse',
  full_name='SetConversationNotificationLevelResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetConversationNotificationLevelResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SetConversationNotificationLevelResponse.timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17945,
  serialized_end=18048,
)


_SETFOCUSREQUEST = _descriptor.Descriptor(
  name='SetFocusRequest',
  full_name='SetFocusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetFocusRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='SetFocusRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='SetFocusRequest.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeout_secs', full_name='SetFocusRequest.timeout_secs', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18051,
  serialized_end=18198,
)


_SETFOCUSRESPONSE = _descriptor.Descriptor(
  name='SetFocusResponse',
  full_name='SetFocusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetFocusResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SetFocusResponse.timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18200,
  serialized_end=18279,
)


_SETGROUPLINKSHARINGENABLEDREQUEST = _descriptor.Descriptor(
  name='SetGroupLinkSharingEnabledRequest',
  full_name='SetGroupLinkSharingEnabledRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetGroupLinkSharingEnabledRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='event_request_header', full_name='SetGroupLinkSharingEnabledRequest.event_request_header', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='group_link_sharing_status', full_name='SetGroupLinkSharingEnabledRequest.group_link_sharing_status', index=2,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18282,
  serialized_end=18468,
)


_SETGROUPLINKSHARINGENABLEDRESPONSE = _descriptor.Descriptor(
  name='SetGroupLinkSharingEnabledResponse',
  full_name='SetGroupLinkSharingEnabledResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetGroupLinkSharingEnabledResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created_event', full_name='SetGroupLinkSharingEnabledResponse.created_event', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='updated_conversation', full_name='SetGroupLinkSharingEnabledResponse.updated_conversation', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18471,
  serialized_end=18625,
)


_SETPRESENCEREQUEST = _descriptor.Descriptor(
  name='SetPresenceRequest',
  full_name='SetPresenceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetPresenceRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='presence_state_setting', full_name='SetPresenceRequest.presence_state_setting', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dnd_setting', full_name='SetPresenceRequest.dnd_setting', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desktop_off_setting', full_name='SetPresenceRequest.desktop_off_setting', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mood_setting', full_name='SetPresenceRequest.mood_setting', index=4,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18628,
  serialized_end=18862,
)


_SETPRESENCERESPONSE = _descriptor.Descriptor(
  name='SetPresenceResponse',
  full_name='SetPresenceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetPresenceResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18864,
  serialized_end=18927,
)


_SETTYPINGREQUEST = _descriptor.Descriptor(
  name='SetTypingRequest',
  full_name='SetTypingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SetTypingRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='SetTypingRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='SetTypingRequest.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18929,
  serialized_end=19056,
)


_SETTYPINGRESPONSE = _descriptor.Descriptor(
  name='SetTypingResponse',
  full_name='SetTypingResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SetTypingResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='SetTypingResponse.timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19058,
  serialized_end=19138,
)


_SYNCALLNEWEVENTSREQUEST = _descriptor.Descriptor(
  name='SyncAllNewEventsRequest',
  full_name='SyncAllNewEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SyncAllNewEventsRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_sync_timestamp', full_name='SyncAllNewEventsRequest.last_sync_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_response_size_bytes', full_name='SyncAllNewEventsRequest.max_response_size_bytes', index=2,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19140,
  serialized_end=19267,
)


_SYNCALLNEWEVENTSRESPONSE = _descriptor.Descriptor(
  name='SyncAllNewEventsResponse',
  full_name='SyncAllNewEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SyncAllNewEventsResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sync_timestamp', full_name='SyncAllNewEventsResponse.sync_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_state', full_name='SyncAllNewEventsResponse.conversation_state', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19270,
  serialized_end=19410,
)


_SYNCRECENTCONVERSATIONSREQUEST = _descriptor.Descriptor(
  name='SyncRecentConversationsRequest',
  full_name='SyncRecentConversationsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='SyncRecentConversationsRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_event_timestamp', full_name='SyncRecentConversationsRequest.last_event_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_conversations', full_name='SyncRecentConversationsRequest.max_conversations', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_events_per_conversation', full_name='SyncRecentConversationsRequest.max_events_per_conversation', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sync_filter', full_name='SyncRecentConversationsRequest.sync_filter', index=4,
      number=5, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19413,
  serialized_end=19613,
)


_SYNCRECENTCONVERSATIONSRESPONSE = _descriptor.Descriptor(
  name='SyncRecentConversationsResponse',
  full_name='SyncRecentConversationsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='SyncRecentConversationsResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sync_timestamp', full_name='SyncRecentConversationsResponse.sync_timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_state', full_name='SyncRecentConversationsResponse.conversation_state', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='continuation_end_timestamp', full_name='SyncRecentConversationsResponse.continuation_end_timestamp', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19616,
  serialized_end=19799,
)


_UPDATEWATERMARKREQUEST = _descriptor.Descriptor(
  name='UpdateWatermarkRequest',
  full_name='UpdateWatermarkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_header', full_name='UpdateWatermarkRequest.request_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conversation_id', full_name='UpdateWatermarkRequest.conversation_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_read_timestamp', full_name='UpdateWatermarkRequest.last_read_timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19802,
  serialized_end=19937,
)


_UPDATEWATERMARKRESPONSE = _descriptor.Descriptor(
  name='UpdateWatermarkResponse',
  full_name='UpdateWatermarkResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response_header', full_name='UpdateWatermarkResponse.response_header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19939,
  serialized_end=20006,
)

_NOTIFICATIONSETTINGS.fields_by_name['dnd_settings'].message_type = _DONOTDISTURBSETTING
_PRESENCE.fields_by_name['device_status'].message_type = _DEVICESTATUS
_PRESENCE.fields_by_name['mood_message'].message_type = _MOODMESSAGE
_PRESENCE.fields_by_name['last_seen'].message_type = _LASTSEEN
_PRESENCERESULT.fields_by_name['user_id'].message_type = _PARTICIPANTID
_PRESENCERESULT.fields_by_name['presence'].message_type = _PRESENCE
_CLIENTPRESENCESTATE.fields_by_name['identifier'].message_type = _CLIENTIDENTIFIER
_CLIENTPRESENCESTATE.fields_by_name['state'].enum_type = _CLIENTPRESENCESTATETYPE
_USEREVENTSTATE.fields_by_name['user_id'].message_type = _PARTICIPANTID
_USEREVENTSTATE.fields_by_name['notification_level'].enum_type = _NOTIFICATIONLEVEL
_SEGMENT.fields_by_name['type'].enum_type = _SEGMENTTYPE
_SEGMENT.fields_by_name['formatting'].message_type = _FORMATTING
_SEGMENT.fields_by_name['link_data'].message_type = _LINKDATA
_PLUSPHOTO_THUMBNAIL.containing_type = _PLUSPHOTO
_PLUSPHOTO.fields_by_name['thumbnail'].message_type = _PLUSPHOTO_THUMBNAIL
_PLUSPHOTO.fields_by_name['media_type'].enum_type = _PLUSPHOTO_MEDIATYPE
_PLUSPHOTO_MEDIATYPE.containing_type = _PLUSPHOTO
_PLACE.fields_by_name['address'].message_type = _EMBEDITEM
_PLACE.fields_by_name['geo'].message_type = _EMBEDITEM
_PLACE.fields_by_name['representative_image'].message_type = _EMBEDITEM
_EMBEDITEM_POSTALADDRESS.containing_type = _EMBEDITEM
_EMBEDITEM_GEOCOORDINATES.containing_type = _EMBEDITEM
_EMBEDITEM_IMAGE.containing_type = _EMBEDITEM
_EMBEDITEM.fields_by_name['type'].enum_type = _ITEMTYPE
_EMBEDITEM.fields_by_name['plus_photo'].message_type = _PLUSPHOTO
_EMBEDITEM.fields_by_name['place'].message_type = _PLACE
_EMBEDITEM.fields_by_name['postal_address'].message_type = _EMBEDITEM_POSTALADDRESS
_EMBEDITEM.fields_by_name['geo_coordinates'].message_type = _EMBEDITEM_GEOCOORDINATES
_EMBEDITEM.fields_by_name['image'].message_type = _EMBEDITEM_IMAGE
_ATTACHMENT.fields_by_name['embed_item'].message_type = _EMBEDITEM
_MESSAGECONTENT.fields_by_name['segment'].message_type = _SEGMENT
_MESSAGECONTENT.fields_by_name['attachment'].message_type = _ATTACHMENT
_CHATMESSAGE.fields_by_name['annotation'].message_type = _EVENTANNOTATION
_CHATMESSAGE.fields_by_name['message_content'].message_type = _MESSAGECONTENT
_MEMBERSHIPCHANGE.fields_by_name['type'].enum_type = _MEMBERSHIPCHANGETYPE
_MEMBERSHIPCHANGE.fields_by_name['participant_ids'].message_type = _PARTICIPANTID
_HANGOUTEVENT.fields_by_name['event_type'].enum_type = _HANGOUTEVENTTYPE
_HANGOUTEVENT.fields_by_name['participant_id'].message_type = _PARTICIPANTID
_OTRMODIFICATION.fields_by_name['old_otr_status'].enum_type = _OFFTHERECORDSTATUS
_OTRMODIFICATION.fields_by_name['new_otr_status'].enum_type = _OFFTHERECORDSTATUS
_OTRMODIFICATION.fields_by_name['old_otr_toggle'].enum_type = _OFFTHERECORDTOGGLE
_OTRMODIFICATION.fields_by_name['new_otr_toggle'].enum_type = _OFFTHERECORDTOGGLE
_EVENT.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_EVENT.fields_by_name['sender_id'].message_type = _PARTICIPANTID
_EVENT.fields_by_name['self_event_state'].message_type = _USEREVENTSTATE
_EVENT.fields_by_name['source_type'].enum_type = _SOURCETYPE
_EVENT.fields_by_name['chat_message'].message_type = _CHATMESSAGE
_EVENT.fields_by_name['membership_change'].message_type = _MEMBERSHIPCHANGE
_EVENT.fields_by_name['conversation_rename'].message_type = _CONVERSATIONRENAME
_EVENT.fields_by_name['hangout_event'].message_type = _HANGOUTEVENT
_EVENT.fields_by_name['otr_modification'].message_type = _OTRMODIFICATION
_EVENT.fields_by_name['otr_status'].enum_type = _OFFTHERECORDSTATUS
_EVENT.fields_by_name['medium_type'].message_type = _DELIVERYMEDIUM
_EVENT.fields_by_name['event_type'].enum_type = _EVENTTYPE
_EVENT.fields_by_name['hash_modifier'].message_type = _HASHMODIFIER
_EVENT.fields_by_name['group_link_sharing_modification'].message_type = _GROUPLINKSHARINGMODIFICATION
_USERREADSTATE.fields_by_name['participant_id'].message_type = _PARTICIPANTID
_DELIVERYMEDIUM.fields_by_name['medium_type'].enum_type = _DELIVERYMEDIUMTYPE
_DELIVERYMEDIUM.fields_by_name['phone_number'].message_type = _PHONENUMBER
_DELIVERYMEDIUMOPTION.fields_by_name['delivery_medium'].message_type = _DELIVERYMEDIUM
_USERCONVERSATIONSTATE.fields_by_name['self_read_state'].message_type = _USERREADSTATE
_USERCONVERSATIONSTATE.fields_by_name['status'].enum_type = _CONVERSATIONSTATUS
_USERCONVERSATIONSTATE.fields_by_name['notification_level'].enum_type = _NOTIFICATIONLEVEL
_USERCONVERSATIONSTATE.fields_by_name['view'].enum_type = _CONVERSATIONVIEW
_USERCONVERSATIONSTATE.fields_by_name['inviter_id'].message_type = _PARTICIPANTID
_USERCONVERSATIONSTATE.fields_by_name['invite_affinity'].enum_type = _INVITATIONAFFINITY
_USERCONVERSATIONSTATE.fields_by_name['delivery_medium_option'].message_type = _DELIVERYMEDIUMOPTION
_CONVERSATIONPARTICIPANTDATA.fields_by_name['id'].message_type = _PARTICIPANTID
_CONVERSATIONPARTICIPANTDATA.fields_by_name['invitation_status'].enum_type = _INVITATIONSTATUS
_CONVERSATIONPARTICIPANTDATA.fields_by_name['participant_type'].enum_type = _PARTICIPANTTYPE
_CONVERSATIONPARTICIPANTDATA.fields_by_name['new_invitation_status'].enum_type = _INVITATIONSTATUS
_CONVERSATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_CONVERSATION.fields_by_name['type'].enum_type = _CONVERSATIONTYPE
_CONVERSATION.fields_by_name['self_conversation_state'].message_type = _USERCONVERSATIONSTATE
_CONVERSATION.fields_by_name['read_state'].message_type = _USERREADSTATE
_CONVERSATION.fields_by_name['otr_status'].enum_type = _OFFTHERECORDSTATUS
_CONVERSATION.fields_by_name['otr_toggle'].enum_type = _OFFTHERECORDTOGGLE
_CONVERSATION.fields_by_name['current_participant'].message_type = _PARTICIPANTID
_CONVERSATION.fields_by_name['participant_data'].message_type = _CONVERSATIONPARTICIPANTDATA
_CONVERSATION.fields_by_name['network_type'].enum_type = _NETWORKTYPE
_CONVERSATION.fields_by_name['force_history_state'].enum_type = _FORCEHISTORY
_CONVERSATION.fields_by_name['group_link_sharing_status'].enum_type = _GROUPLINKSHARINGSTATUS
_BLOCKSTATECHANGE.fields_by_name['participant_id'].message_type = _PARTICIPANTID
_BLOCKSTATECHANGE.fields_by_name['new_block_state'].enum_type = _BLOCKSTATE
_EXISTINGMEDIA.fields_by_name['photo'].message_type = _PHOTO
_EVENTREQUESTHEADER.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_EVENTREQUESTHEADER.fields_by_name['expected_otr'].enum_type = _OFFTHERECORDSTATUS
_EVENTREQUESTHEADER.fields_by_name['delivery_medium'].message_type = _DELIVERYMEDIUM
_EVENTREQUESTHEADER.fields_by_name['event_type'].enum_type = _EVENTTYPE
_CLIENTVERSION.fields_by_name['client_id'].enum_type = _CLIENTID
_CLIENTVERSION.fields_by_name['build_type'].enum_type = _CLIENTBUILDTYPE
_REQUESTHEADER.fields_by_name['client_version'].message_type = _CLIENTVERSION
_REQUESTHEADER.fields_by_name['client_identifier'].message_type = _CLIENTIDENTIFIER
_RESPONSEHEADER.fields_by_name['status'].enum_type = _RESPONSESTATUS
_ENTITY.fields_by_name['id'].message_type = _PARTICIPANTID
_ENTITY.fields_by_name['presence'].message_type = _PRESENCE
_ENTITY.fields_by_name['properties'].message_type = _ENTITYPROPERTIES
_ENTITY.fields_by_name['entity_type'].enum_type = _PARTICIPANTTYPE
_ENTITY.fields_by_name['had_past_hangout_state'].enum_type = _ENTITY_PASTHANGOUTSTATE
_ENTITY_PASTHANGOUTSTATE.containing_type = _ENTITY
_ENTITYPROPERTIES.fields_by_name['type'].enum_type = _PROFILETYPE
_ENTITYPROPERTIES.fields_by_name['gender'].enum_type = _GENDER
_ENTITYPROPERTIES.fields_by_name['photo_url_status'].enum_type = _PHOTOURLSTATUS
_CONVERSATIONSTATE.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_CONVERSATIONSTATE.fields_by_name['conversation'].message_type = _CONVERSATION
_CONVERSATIONSTATE.fields_by_name['event'].message_type = _EVENT
_CONVERSATIONSTATE.fields_by_name['event_continuation_token'].message_type = _EVENTCONTINUATIONTOKEN
_CONFIGURATIONBIT.fields_by_name['configuration_bit_type'].enum_type = _CONFIGURATIONBITTYPE
_RICHPRESENCESTATE.fields_by_name['get_rich_presence_enabled_state'].message_type = _RICHPRESENCEENABLEDSTATE
_RICHPRESENCEENABLEDSTATE.fields_by_name['type'].enum_type = _RICHPRESENCETYPE
_PRESENCESTATESETTING.fields_by_name['type'].enum_type = _CLIENTPRESENCESTATETYPE
_MOODMESSAGE.fields_by_name['mood_content'].message_type = _MOODCONTENT
_MOODCONTENT.fields_by_name['segment'].message_type = _SEGMENT
_MOODSETTING.fields_by_name['mood_message'].message_type = _MOODMESSAGE
_MOODSTATE.fields_by_name['mood_setting'].message_type = _MOODSETTING
_DELETEACTION.fields_by_name['delete_type'].enum_type = _DELETETYPE
_DESKTOPSOUNDSETTING.fields_by_name['desktop_sound_state'].enum_type = _SOUNDSTATE
_DESKTOPSOUNDSETTING.fields_by_name['desktop_ring_sound_state'].enum_type = _SOUNDSTATE
_PHONEDATA.fields_by_name['phone'].message_type = _PHONE
_PHONEDATA.fields_by_name['caller_id_settings_mask'].enum_type = _CALLERIDSETTINGSMASK
_PHONE.fields_by_name['phone_number'].message_type = _PHONENUMBER
_PHONE.fields_by_name['verification_status'].enum_type = _PHONEVERIFICATIONSTATUS
_PHONE.fields_by_name['discoverability_status'].enum_type = _PHONEDISCOVERABILITYSTATUS
_I18NDATA.fields_by_name['validation_result'].enum_type = _PHONEVALIDATIONRESULT
_PHONENUMBER.fields_by_name['i18n_data'].message_type = _I18NDATA
_SUGGESTEDCONTACT.fields_by_name['entity'].message_type = _ENTITY
_SUGGESTEDCONTACT.fields_by_name['invitation_status'].enum_type = _INVITATIONSTATUS
_SUGGESTEDCONTACTGROUP.fields_by_name['contact'].message_type = _SUGGESTEDCONTACT
_GROUPLINKSHARINGMODIFICATION.fields_by_name['new_status'].enum_type = _GROUPLINKSHARINGSTATUS
_STATEUPDATE.fields_by_name['state_update_header'].message_type = _STATEUPDATEHEADER
_STATEUPDATE.fields_by_name['conversation'].message_type = _CONVERSATION
_STATEUPDATE.fields_by_name['conversation_notification'].message_type = _CONVERSATIONNOTIFICATION
_STATEUPDATE.fields_by_name['event_notification'].message_type = _EVENTNOTIFICATION
_STATEUPDATE.fields_by_name['focus_notification'].message_type = _SETFOCUSNOTIFICATION
_STATEUPDATE.fields_by_name['typing_notification'].message_type = _SETTYPINGNOTIFICATION
_STATEUPDATE.fields_by_name['notification_level_notification'].message_type = _SETCONVERSATIONNOTIFICATIONLEVELNOTIFICATION
_STATEUPDATE.fields_by_name['reply_to_invite_notification'].message_type = _REPLYTOINVITENOTIFICATION
_STATEUPDATE.fields_by_name['watermark_notification'].message_type = _WATERMARKNOTIFICATION
_STATEUPDATE.fields_by_name['view_modification'].message_type = _CONVERSATIONVIEWMODIFICATION
_STATEUPDATE.fields_by_name['easter_egg_notification'].message_type = _EASTEREGGNOTIFICATION
_STATEUPDATE.fields_by_name['self_presence_notification'].message_type = _SELFPRESENCENOTIFICATION
_STATEUPDATE.fields_by_name['delete_notification'].message_type = _DELETEACTIONNOTIFICATION
_STATEUPDATE.fields_by_name['presence_notification'].message_type = _PRESENCENOTIFICATION
_STATEUPDATE.fields_by_name['block_notification'].message_type = _BLOCKNOTIFICATION
_STATEUPDATE.fields_by_name['notification_setting_notification'].message_type = _SETNOTIFICATIONSETTINGNOTIFICATION
_STATEUPDATE.fields_by_name['rich_presence_enabled_state_notification'].message_type = _RICHPRESENCEENABLEDSTATENOTIFICATION
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['conversation_notification'])
_STATEUPDATE.fields_by_name['conversation_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['event_notification'])
_STATEUPDATE.fields_by_name['event_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['focus_notification'])
_STATEUPDATE.fields_by_name['focus_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['typing_notification'])
_STATEUPDATE.fields_by_name['typing_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['notification_level_notification'])
_STATEUPDATE.fields_by_name['notification_level_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['reply_to_invite_notification'])
_STATEUPDATE.fields_by_name['reply_to_invite_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['watermark_notification'])
_STATEUPDATE.fields_by_name['watermark_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['view_modification'])
_STATEUPDATE.fields_by_name['view_modification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['easter_egg_notification'])
_STATEUPDATE.fields_by_name['easter_egg_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['self_presence_notification'])
_STATEUPDATE.fields_by_name['self_presence_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['delete_notification'])
_STATEUPDATE.fields_by_name['delete_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['presence_notification'])
_STATEUPDATE.fields_by_name['presence_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['block_notification'])
_STATEUPDATE.fields_by_name['block_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['notification_setting_notification'])
_STATEUPDATE.fields_by_name['notification_setting_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATE.oneofs_by_name['state_update'].fields.append(
  _STATEUPDATE.fields_by_name['rich_presence_enabled_state_notification'])
_STATEUPDATE.fields_by_name['rich_presence_enabled_state_notification'].containing_oneof = _STATEUPDATE.oneofs_by_name['state_update']
_STATEUPDATEHEADER.fields_by_name['active_client_state'].enum_type = _ACTIVECLIENTSTATE
_STATEUPDATEHEADER.fields_by_name['notification_settings'].message_type = _NOTIFICATIONSETTINGS
_BATCHUPDATE.fields_by_name['state_update'].message_type = _STATEUPDATE
_CONVERSATIONNOTIFICATION.fields_by_name['conversation'].message_type = _CONVERSATION
_EVENTNOTIFICATION.fields_by_name['event'].message_type = _EVENT
_SETFOCUSNOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_SETFOCUSNOTIFICATION.fields_by_name['sender_id'].message_type = _PARTICIPANTID
_SETFOCUSNOTIFICATION.fields_by_name['type'].enum_type = _FOCUSTYPE
_SETFOCUSNOTIFICATION.fields_by_name['device'].enum_type = _FOCUSDEVICE
_SETTYPINGNOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_SETTYPINGNOTIFICATION.fields_by_name['sender_id'].message_type = _PARTICIPANTID
_SETTYPINGNOTIFICATION.fields_by_name['type'].enum_type = _TYPINGTYPE
_SETCONVERSATIONNOTIFICATIONLEVELNOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_SETCONVERSATIONNOTIFICATIONLEVELNOTIFICATION.fields_by_name['level'].enum_type = _NOTIFICATIONLEVEL
_REPLYTOINVITENOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_REPLYTOINVITENOTIFICATION.fields_by_name['type'].enum_type = _REPLYTOINVITETYPE
_WATERMARKNOTIFICATION.fields_by_name['sender_id'].message_type = _PARTICIPANTID
_WATERMARKNOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_CONVERSATIONVIEWMODIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_CONVERSATIONVIEWMODIFICATION.fields_by_name['old_view'].enum_type = _CONVERSATIONVIEW
_CONVERSATIONVIEWMODIFICATION.fields_by_name['new_view'].enum_type = _CONVERSATIONVIEW
_EASTEREGGNOTIFICATION.fields_by_name['sender_id'].message_type = _PARTICIPANTID
_EASTEREGGNOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_EASTEREGGNOTIFICATION.fields_by_name['easter_egg'].message_type = _EASTEREGG
_SELFPRESENCENOTIFICATION.fields_by_name['client_presence_state'].message_type = _CLIENTPRESENCESTATE
_SELFPRESENCENOTIFICATION.fields_by_name['do_not_disturb_setting'].message_type = _DONOTDISTURBSETTING
_SELFPRESENCENOTIFICATION.fields_by_name['desktop_off_setting'].message_type = _DESKTOPOFFSETTING
_SELFPRESENCENOTIFICATION.fields_by_name['desktop_off_state'].message_type = _DESKTOPOFFSTATE
_SELFPRESENCENOTIFICATION.fields_by_name['mood_state'].message_type = _MOODSTATE
_DELETEACTIONNOTIFICATION.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_DELETEACTIONNOTIFICATION.fields_by_name['delete_action'].message_type = _DELETEACTION
_PRESENCENOTIFICATION.fields_by_name['presence'].message_type = _PRESENCERESULT
_BLOCKNOTIFICATION.fields_by_name['block_state_change'].message_type = _BLOCKSTATECHANGE
_SETNOTIFICATIONSETTINGNOTIFICATION.fields_by_name['configuration_bit'].message_type = _CONFIGURATIONBIT
_SETNOTIFICATIONSETTINGNOTIFICATION.fields_by_name['desktop_sound_setting'].message_type = _DESKTOPSOUNDSETTING
_RICHPRESENCEENABLEDSTATENOTIFICATION.fields_by_name['rich_presence_enabled_state'].message_type = _RICHPRESENCEENABLEDSTATE
_CONVERSATIONSPEC.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_OFFNETWORKADDRESS.fields_by_name['type'].enum_type = _OFFNETWORKADDRESSTYPE
_ENTITYRESULT.fields_by_name['lookup_spec'].message_type = _ENTITYLOOKUPSPEC
_ENTITYRESULT.fields_by_name['entity'].message_type = _ENTITY
_ADDUSERREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_ADDUSERREQUEST.fields_by_name['invitee_id'].message_type = _INVITEEID
_ADDUSERREQUEST.fields_by_name['event_request_header'].message_type = _EVENTREQUESTHEADER
_ADDUSERRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_ADDUSERRESPONSE.fields_by_name['created_event'].message_type = _EVENT
_CREATECONVERSATIONREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_CREATECONVERSATIONREQUEST.fields_by_name['type'].enum_type = _CONVERSATIONTYPE
_CREATECONVERSATIONREQUEST.fields_by_name['invitee_id'].message_type = _INVITEEID
_CREATECONVERSATIONRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_CREATECONVERSATIONRESPONSE.fields_by_name['conversation'].message_type = _CONVERSATION
_DELETECONVERSATIONREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_DELETECONVERSATIONREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_DELETECONVERSATIONRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_DELETECONVERSATIONRESPONSE.fields_by_name['delete_action'].message_type = _DELETEACTION
_EASTEREGGREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_EASTEREGGREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_EASTEREGGREQUEST.fields_by_name['easter_egg'].message_type = _EASTEREGG
_EASTEREGGRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_GETCONVERSATIONREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETCONVERSATIONREQUEST.fields_by_name['conversation_spec'].message_type = _CONVERSATIONSPEC
_GETCONVERSATIONREQUEST.fields_by_name['event_continuation_token'].message_type = _EVENTCONTINUATIONTOKEN
_GETCONVERSATIONRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_GETCONVERSATIONRESPONSE.fields_by_name['conversation_state'].message_type = _CONVERSATIONSTATE
_GETENTITYBYIDREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETENTITYBYIDREQUEST.fields_by_name['batch_lookup_spec'].message_type = _ENTITYLOOKUPSPEC
_GETENTITYBYIDRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_GETENTITYBYIDRESPONSE.fields_by_name['entity'].message_type = _ENTITY
_GETENTITYBYIDRESPONSE.fields_by_name['entity_result'].message_type = _ENTITYRESULT
_GETGROUPCONVERSATIONURLREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETGROUPCONVERSATIONURLREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_GETGROUPCONVERSATIONURLRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['favorites'].message_type = _SUGGESTEDCONTACTGROUPHASH
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['contacts_you_hangout_with'].message_type = _SUGGESTEDCONTACTGROUPHASH
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['other_contacts_on_hangouts'].message_type = _SUGGESTEDCONTACTGROUPHASH
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['other_contacts'].message_type = _SUGGESTEDCONTACTGROUPHASH
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['dismissed_contacts'].message_type = _SUGGESTEDCONTACTGROUPHASH
_GETSUGGESTEDENTITIESREQUEST.fields_by_name['pinned_favorites'].message_type = _SUGGESTEDCONTACTGROUPHASH
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['entity'].message_type = _ENTITY
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['favorites'].message_type = _SUGGESTEDCONTACTGROUP
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['contacts_you_hangout_with'].message_type = _SUGGESTEDCONTACTGROUP
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['other_contacts_on_hangouts'].message_type = _SUGGESTEDCONTACTGROUP
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['other_contacts'].message_type = _SUGGESTEDCONTACTGROUP
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['dismissed_contacts'].message_type = _SUGGESTEDCONTACTGROUP
_GETSUGGESTEDENTITIESRESPONSE.fields_by_name['pinned_favorites'].message_type = _SUGGESTEDCONTACTGROUP
_GETSELFINFOREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_GETSELFINFORESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_GETSELFINFORESPONSE.fields_by_name['self_entity'].message_type = _ENTITY
_GETSELFINFORESPONSE.fields_by_name['dnd_state'].message_type = _DONOTDISTURBSETTING
_GETSELFINFORESPONSE.fields_by_name['desktop_off_setting'].message_type = _DESKTOPOFFSETTING
_GETSELFINFORESPONSE.fields_by_name['phone_data'].message_type = _PHONEDATA
_GETSELFINFORESPONSE.fields_by_name['configuration_bit'].message_type = _CONFIGURATIONBIT
_GETSELFINFORESPONSE.fields_by_name['desktop_off_state'].message_type = _DESKTOPOFFSTATE
_GETSELFINFORESPONSE.fields_by_name['desktop_sound_setting'].message_type = _DESKTOPSOUNDSETTING
_GETSELFINFORESPONSE.fields_by_name['rich_presence_state'].message_type = _RICHPRESENCESTATE
_GETSELFINFORESPONSE.fields_by_name['default_country'].message_type = _COUNTRY
_QUERYPRESENCEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_QUERYPRESENCEREQUEST.fields_by_name['participant_id'].message_type = _PARTICIPANTID
_QUERYPRESENCEREQUEST.fields_by_name['field_mask'].enum_type = _FIELDMASK
_QUERYPRESENCERESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_QUERYPRESENCERESPONSE.fields_by_name['presence_result'].message_type = _PRESENCERESULT
_REMOVEUSERREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_REMOVEUSERREQUEST.fields_by_name['participant_id'].message_type = _PARTICIPANTID
_REMOVEUSERREQUEST.fields_by_name['event_request_header'].message_type = _EVENTREQUESTHEADER
_REMOVEUSERRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_REMOVEUSERRESPONSE.fields_by_name['created_event'].message_type = _EVENT
_RENAMECONVERSATIONREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_RENAMECONVERSATIONREQUEST.fields_by_name['event_request_header'].message_type = _EVENTREQUESTHEADER
_RENAMECONVERSATIONRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_RENAMECONVERSATIONRESPONSE.fields_by_name['created_event'].message_type = _EVENT
_SEARCHENTITIESREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SEARCHENTITIESRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SEARCHENTITIESRESPONSE.fields_by_name['entity'].message_type = _ENTITY
_LOCATION.fields_by_name['place'].message_type = _PLACE
_SENDCHATMESSAGEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SENDCHATMESSAGEREQUEST.fields_by_name['annotation'].message_type = _EVENTANNOTATION
_SENDCHATMESSAGEREQUEST.fields_by_name['message_content'].message_type = _MESSAGECONTENT
_SENDCHATMESSAGEREQUEST.fields_by_name['existing_media'].message_type = _EXISTINGMEDIA
_SENDCHATMESSAGEREQUEST.fields_by_name['event_request_header'].message_type = _EVENTREQUESTHEADER
_SENDCHATMESSAGEREQUEST.fields_by_name['user_id'].message_type = _PARTICIPANTID
_SENDCHATMESSAGEREQUEST.fields_by_name['location'].message_type = _LOCATION
_SENDCHATMESSAGERESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SENDCHATMESSAGERESPONSE.fields_by_name['created_event'].message_type = _EVENT
_MODIFYOTRSTATUSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_MODIFYOTRSTATUSREQUEST.fields_by_name['otr_status'].enum_type = _OFFTHERECORDSTATUS
_MODIFYOTRSTATUSREQUEST.fields_by_name['event_request_header'].message_type = _EVENTREQUESTHEADER
_MODIFYOTRSTATUSRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_MODIFYOTRSTATUSRESPONSE.fields_by_name['created_event'].message_type = _EVENT
_SENDOFFNETWORKINVITATIONREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SENDOFFNETWORKINVITATIONREQUEST.fields_by_name['invitee_address'].message_type = _OFFNETWORKADDRESS
_SENDOFFNETWORKINVITATIONRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETACTIVECLIENTREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETACTIVECLIENTRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETCONVERSATIONLEVELREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETCONVERSATIONLEVELRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETCONVERSATIONNOTIFICATIONLEVELREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETCONVERSATIONNOTIFICATIONLEVELREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_SETCONVERSATIONNOTIFICATIONLEVELREQUEST.fields_by_name['level'].enum_type = _NOTIFICATIONLEVEL
_SETCONVERSATIONNOTIFICATIONLEVELRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETFOCUSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETFOCUSREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_SETFOCUSREQUEST.fields_by_name['type'].enum_type = _FOCUSTYPE
_SETFOCUSRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETGROUPLINKSHARINGENABLEDREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETGROUPLINKSHARINGENABLEDREQUEST.fields_by_name['event_request_header'].message_type = _EVENTREQUESTHEADER
_SETGROUPLINKSHARINGENABLEDREQUEST.fields_by_name['group_link_sharing_status'].enum_type = _GROUPLINKSHARINGSTATUS
_SETGROUPLINKSHARINGENABLEDRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETGROUPLINKSHARINGENABLEDRESPONSE.fields_by_name['created_event'].message_type = _EVENT
_SETGROUPLINKSHARINGENABLEDRESPONSE.fields_by_name['updated_conversation'].message_type = _CONVERSATION
_SETPRESENCEREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETPRESENCEREQUEST.fields_by_name['presence_state_setting'].message_type = _PRESENCESTATESETTING
_SETPRESENCEREQUEST.fields_by_name['dnd_setting'].message_type = _DNDSETTING
_SETPRESENCEREQUEST.fields_by_name['desktop_off_setting'].message_type = _DESKTOPOFFSETTING
_SETPRESENCEREQUEST.fields_by_name['mood_setting'].message_type = _MOODSETTING
_SETPRESENCERESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SETTYPINGREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SETTYPINGREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_SETTYPINGREQUEST.fields_by_name['type'].enum_type = _TYPINGTYPE
_SETTYPINGRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SYNCALLNEWEVENTSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SYNCALLNEWEVENTSRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SYNCALLNEWEVENTSRESPONSE.fields_by_name['conversation_state'].message_type = _CONVERSATIONSTATE
_SYNCRECENTCONVERSATIONSREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_SYNCRECENTCONVERSATIONSREQUEST.fields_by_name['sync_filter'].enum_type = _SYNCFILTER
_SYNCRECENTCONVERSATIONSRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
_SYNCRECENTCONVERSATIONSRESPONSE.fields_by_name['conversation_state'].message_type = _CONVERSATIONSTATE
_UPDATEWATERMARKREQUEST.fields_by_name['request_header'].message_type = _REQUESTHEADER
_UPDATEWATERMARKREQUEST.fields_by_name['conversation_id'].message_type = _CONVERSATIONID
_UPDATEWATERMARKRESPONSE.fields_by_name['response_header'].message_type = _RESPONSEHEADER
DESCRIPTOR.message_types_by_name['DoNotDisturbSetting'] = _DONOTDISTURBSETTING
DESCRIPTOR.message_types_by_name['NotificationSettings'] = _NOTIFICATIONSETTINGS
DESCRIPTOR.message_types_by_name['ConversationId'] = _CONVERSATIONID
DESCRIPTOR.message_types_by_name['ParticipantId'] = _PARTICIPANTID
DESCRIPTOR.message_types_by_name['DeviceStatus'] = _DEVICESTATUS
DESCRIPTOR.message_types_by_name['LastSeen'] = _LASTSEEN
DESCRIPTOR.message_types_by_name['Presence'] = _PRESENCE
DESCRIPTOR.message_types_by_name['PresenceResult'] = _PRESENCERESULT
DESCRIPTOR.message_types_by_name['ClientIdentifier'] = _CLIENTIDENTIFIER
DESCRIPTOR.message_types_by_name['ClientPresenceState'] = _CLIENTPRESENCESTATE
DESCRIPTOR.message_types_by_name['UserEventState'] = _USEREVENTSTATE
DESCRIPTOR.message_types_by_name['Formatting'] = _FORMATTING
DESCRIPTOR.message_types_by_name['LinkData'] = _LINKDATA
DESCRIPTOR.message_types_by_name['Segment'] = _SEGMENT
DESCRIPTOR.message_types_by_name['PlusPhoto'] = _PLUSPHOTO
DESCRIPTOR.message_types_by_name['Place'] = _PLACE
DESCRIPTOR.message_types_by_name['EmbedItem'] = _EMBEDITEM
DESCRIPTOR.message_types_by_name['Attachment'] = _ATTACHMENT
DESCRIPTOR.message_types_by_name['MessageContent'] = _MESSAGECONTENT
DESCRIPTOR.message_types_by_name['EventAnnotation'] = _EVENTANNOTATION
DESCRIPTOR.message_types_by_name['ChatMessage'] = _CHATMESSAGE
DESCRIPTOR.message_types_by_name['MembershipChange'] = _MEMBERSHIPCHANGE
DESCRIPTOR.message_types_by_name['ConversationRename'] = _CONVERSATIONRENAME
DESCRIPTOR.message_types_by_name['HangoutEvent'] = _HANGOUTEVENT
DESCRIPTOR.message_types_by_name['OTRModification'] = _OTRMODIFICATION
DESCRIPTOR.message_types_by_name['HashModifier'] = _HASHMODIFIER
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['UserReadState'] = _USERREADSTATE
DESCRIPTOR.message_types_by_name['DeliveryMedium'] = _DELIVERYMEDIUM
DESCRIPTOR.message_types_by_name['DeliveryMediumOption'] = _DELIVERYMEDIUMOPTION
DESCRIPTOR.message_types_by_name['UserConversationState'] = _USERCONVERSATIONSTATE
DESCRIPTOR.message_types_by_name['ConversationParticipantData'] = _CONVERSATIONPARTICIPANTDATA
DESCRIPTOR.message_types_by_name['Conversation'] = _CONVERSATION
DESCRIPTOR.message_types_by_name['EasterEgg'] = _EASTEREGG
DESCRIPTOR.message_types_by_name['BlockStateChange'] = _BLOCKSTATECHANGE
DESCRIPTOR.message_types_by_name['Photo'] = _PHOTO
DESCRIPTOR.message_types_by_name['ExistingMedia'] = _EXISTINGMEDIA
DESCRIPTOR.message_types_by_name['EventRequestHeader'] = _EVENTREQUESTHEADER
DESCRIPTOR.message_types_by_name['ClientVersion'] = _CLIENTVERSION
DESCRIPTOR.message_types_by_name['RequestHeader'] = _REQUESTHEADER
DESCRIPTOR.message_types_by_name['ResponseHeader'] = _RESPONSEHEADER
DESCRIPTOR.message_types_by_name['Entity'] = _ENTITY
DESCRIPTOR.message_types_by_name['EntityProperties'] = _ENTITYPROPERTIES
DESCRIPTOR.message_types_by_name['ConversationState'] = _CONVERSATIONSTATE
DESCRIPTOR.message_types_by_name['EventContinuationToken'] = _EVENTCONTINUATIONTOKEN
DESCRIPTOR.message_types_by_name['EntityLookupSpec'] = _ENTITYLOOKUPSPEC
DESCRIPTOR.message_types_by_name['ConfigurationBit'] = _CONFIGURATIONBIT
DESCRIPTOR.message_types_by_name['RichPresenceState'] = _RICHPRESENCESTATE
DESCRIPTOR.message_types_by_name['RichPresenceEnabledState'] = _RICHPRESENCEENABLEDSTATE
DESCRIPTOR.message_types_by_name['DesktopOffSetting'] = _DESKTOPOFFSETTING
DESCRIPTOR.message_types_by_name['DesktopOffState'] = _DESKTOPOFFSTATE
DESCRIPTOR.message_types_by_name['DndSetting'] = _DNDSETTING
DESCRIPTOR.message_types_by_name['PresenceStateSetting'] = _PRESENCESTATESETTING
DESCRIPTOR.message_types_by_name['MoodMessage'] = _MOODMESSAGE
DESCRIPTOR.message_types_by_name['MoodContent'] = _MOODCONTENT
DESCRIPTOR.message_types_by_name['MoodSetting'] = _MOODSETTING
DESCRIPTOR.message_types_by_name['MoodState'] = _MOODSTATE
DESCRIPTOR.message_types_by_name['DeleteAction'] = _DELETEACTION
DESCRIPTOR.message_types_by_name['InviteeID'] = _INVITEEID
DESCRIPTOR.message_types_by_name['Country'] = _COUNTRY
DESCRIPTOR.message_types_by_name['DesktopSoundSetting'] = _DESKTOPSOUNDSETTING
DESCRIPTOR.message_types_by_name['PhoneData'] = _PHONEDATA
DESCRIPTOR.message_types_by_name['Phone'] = _PHONE
DESCRIPTOR.message_types_by_name['I18nData'] = _I18NDATA
DESCRIPTOR.message_types_by_name['PhoneNumber'] = _PHONENUMBER
DESCRIPTOR.message_types_by_name['SuggestedContactGroupHash'] = _SUGGESTEDCONTACTGROUPHASH
DESCRIPTOR.message_types_by_name['SuggestedContact'] = _SUGGESTEDCONTACT
DESCRIPTOR.message_types_by_name['SuggestedContactGroup'] = _SUGGESTEDCONTACTGROUP
DESCRIPTOR.message_types_by_name['GroupLinkSharingModification'] = _GROUPLINKSHARINGMODIFICATION
DESCRIPTOR.message_types_by_name['StateUpdate'] = _STATEUPDATE
DESCRIPTOR.message_types_by_name['StateUpdateHeader'] = _STATEUPDATEHEADER
DESCRIPTOR.message_types_by_name['BatchUpdate'] = _BATCHUPDATE
DESCRIPTOR.message_types_by_name['ConversationNotification'] = _CONVERSATIONNOTIFICATION
DESCRIPTOR.message_types_by_name['EventNotification'] = _EVENTNOTIFICATION
DESCRIPTOR.message_types_by_name['SetFocusNotification'] = _SETFOCUSNOTIFICATION
DESCRIPTOR.message_types_by_name['SetTypingNotification'] = _SETTYPINGNOTIFICATION
DESCRIPTOR.message_types_by_name['SetConversationNotificationLevelNotification'] = _SETCONVERSATIONNOTIFICATIONLEVELNOTIFICATION
DESCRIPTOR.message_types_by_name['ReplyToInviteNotification'] = _REPLYTOINVITENOTIFICATION
DESCRIPTOR.message_types_by_name['WatermarkNotification'] = _WATERMARKNOTIFICATION
DESCRIPTOR.message_types_by_name['ConversationViewModification'] = _CONVERSATIONVIEWMODIFICATION
DESCRIPTOR.message_types_by_name['EasterEggNotification'] = _EASTEREGGNOTIFICATION
DESCRIPTOR.message_types_by_name['SelfPresenceNotification'] = _SELFPRESENCENOTIFICATION
DESCRIPTOR.message_types_by_name['DeleteActionNotification'] = _DELETEACTIONNOTIFICATION
DESCRIPTOR.message_types_by_name['PresenceNotification'] = _PRESENCENOTIFICATION
DESCRIPTOR.message_types_by_name['BlockNotification'] = _BLOCKNOTIFICATION
DESCRIPTOR.message_types_by_name['SetNotificationSettingNotification'] = _SETNOTIFICATIONSETTINGNOTIFICATION
DESCRIPTOR.message_types_by_name['RichPresenceEnabledStateNotification'] = _RICHPRESENCEENABLEDSTATENOTIFICATION
DESCRIPTOR.message_types_by_name['ConversationSpec'] = _CONVERSATIONSPEC
DESCRIPTOR.message_types_by_name['OffnetworkAddress'] = _OFFNETWORKADDRESS
DESCRIPTOR.message_types_by_name['EntityResult'] = _ENTITYRESULT
DESCRIPTOR.message_types_by_name['AddUserRequest'] = _ADDUSERREQUEST
DESCRIPTOR.message_types_by_name['AddUserResponse'] = _ADDUSERRESPONSE
DESCRIPTOR.message_types_by_name['CreateConversationRequest'] = _CREATECONVERSATIONREQUEST
DESCRIPTOR.message_types_by_name['CreateConversationResponse'] = _CREATECONVERSATIONRESPONSE
DESCRIPTOR.message_types_by_name['DeleteConversationRequest'] = _DELETECONVERSATIONREQUEST
DESCRIPTOR.message_types_by_name['DeleteConversationResponse'] = _DELETECONVERSATIONRESPONSE
DESCRIPTOR.message_types_by_name['EasterEggRequest'] = _EASTEREGGREQUEST
DESCRIPTOR.message_types_by_name['EasterEggResponse'] = _EASTEREGGRESPONSE
DESCRIPTOR.message_types_by_name['GetConversationRequest'] = _GETCONVERSATIONREQUEST
DESCRIPTOR.message_types_by_name['GetConversationResponse'] = _GETCONVERSATIONRESPONSE
DESCRIPTOR.message_types_by_name['GetEntityByIdRequest'] = _GETENTITYBYIDREQUEST
DESCRIPTOR.message_types_by_name['GetEntityByIdResponse'] = _GETENTITYBYIDRESPONSE
DESCRIPTOR.message_types_by_name['GetGroupConversationUrlRequest'] = _GETGROUPCONVERSATIONURLREQUEST
DESCRIPTOR.message_types_by_name['GetGroupConversationUrlResponse'] = _GETGROUPCONVERSATIONURLRESPONSE
DESCRIPTOR.message_types_by_name['GetSuggestedEntitiesRequest'] = _GETSUGGESTEDENTITIESREQUEST
DESCRIPTOR.message_types_by_name['GetSuggestedEntitiesResponse'] = _GETSUGGESTEDENTITIESRESPONSE
DESCRIPTOR.message_types_by_name['GetSelfInfoRequest'] = _GETSELFINFOREQUEST
DESCRIPTOR.message_types_by_name['GetSelfInfoResponse'] = _GETSELFINFORESPONSE
DESCRIPTOR.message_types_by_name['QueryPresenceRequest'] = _QUERYPRESENCEREQUEST
DESCRIPTOR.message_types_by_name['QueryPresenceResponse'] = _QUERYPRESENCERESPONSE
DESCRIPTOR.message_types_by_name['RemoveUserRequest'] = _REMOVEUSERREQUEST
DESCRIPTOR.message_types_by_name['RemoveUserResponse'] = _REMOVEUSERRESPONSE
DESCRIPTOR.message_types_by_name['RenameConversationRequest'] = _RENAMECONVERSATIONREQUEST
DESCRIPTOR.message_types_by_name['RenameConversationResponse'] = _RENAMECONVERSATIONRESPONSE
DESCRIPTOR.message_types_by_name['SearchEntitiesRequest'] = _SEARCHENTITIESREQUEST
DESCRIPTOR.message_types_by_name['SearchEntitiesResponse'] = _SEARCHENTITIESRESPONSE
DESCRIPTOR.message_types_by_name['Location'] = _LOCATION
DESCRIPTOR.message_types_by_name['SendChatMessageRequest'] = _SENDCHATMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['SendChatMessageResponse'] = _SENDCHATMESSAGERESPONSE
DESCRIPTOR.message_types_by_name['ModifyOTRStatusRequest'] = _MODIFYOTRSTATUSREQUEST
DESCRIPTOR.message_types_by_name['ModifyOTRStatusResponse'] = _MODIFYOTRSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['SendOffnetworkInvitationRequest'] = _SENDOFFNETWORKINVITATIONREQUEST
DESCRIPTOR.message_types_by_name['SendOffnetworkInvitationResponse'] = _SENDOFFNETWORKINVITATIONRESPONSE
DESCRIPTOR.message_types_by_name['SetActiveClientRequest'] = _SETACTIVECLIENTREQUEST
DESCRIPTOR.message_types_by_name['SetActiveClientResponse'] = _SETACTIVECLIENTRESPONSE
DESCRIPTOR.message_types_by_name['SetConversationLevelRequest'] = _SETCONVERSATIONLEVELREQUEST
DESCRIPTOR.message_types_by_name['SetConversationLevelResponse'] = _SETCONVERSATIONLEVELRESPONSE
DESCRIPTOR.message_types_by_name['SetConversationNotificationLevelRequest'] = _SETCONVERSATIONNOTIFICATIONLEVELREQUEST
DESCRIPTOR.message_types_by_name['SetConversationNotificationLevelResponse'] = _SETCONVERSATIONNOTIFICATIONLEVELRESPONSE
DESCRIPTOR.message_types_by_name['SetFocusRequest'] = _SETFOCUSREQUEST
DESCRIPTOR.message_types_by_name['SetFocusResponse'] = _SETFOCUSRESPONSE
DESCRIPTOR.message_types_by_name['SetGroupLinkSharingEnabledRequest'] = _SETGROUPLINKSHARINGENABLEDREQUEST
DESCRIPTOR.message_types_by_name['SetGroupLinkSharingEnabledResponse'] = _SETGROUPLINKSHARINGENABLEDRESPONSE
DESCRIPTOR.message_types_by_name['SetPresenceRequest'] = _SETPRESENCEREQUEST
DESCRIPTOR.message_types_by_name['SetPresenceResponse'] = _SETPRESENCERESPONSE
DESCRIPTOR.message_types_by_name['SetTypingRequest'] = _SETTYPINGREQUEST
DESCRIPTOR.message_types_by_name['SetTypingResponse'] = _SETTYPINGRESPONSE
DESCRIPTOR.message_types_by_name['SyncAllNewEventsRequest'] = _SYNCALLNEWEVENTSREQUEST
DESCRIPTOR.message_types_by_name['SyncAllNewEventsResponse'] = _SYNCALLNEWEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['SyncRecentConversationsRequest'] = _SYNCRECENTCONVERSATIONSREQUEST
DESCRIPTOR.message_types_by_name['SyncRecentConversationsResponse'] = _SYNCRECENTCONVERSATIONSRESPONSE
DESCRIPTOR.message_types_by_name['UpdateWatermarkRequest'] = _UPDATEWATERMARKREQUEST
DESCRIPTOR.message_types_by_name['UpdateWatermarkResponse'] = _UPDATEWATERMARKRESPONSE
DESCRIPTOR.enum_types_by_name['ActiveClientState'] = _ACTIVECLIENTSTATE
DESCRIPTOR.enum_types_by_name['FocusType'] = _FOCUSTYPE
DESCRIPTOR.enum_types_by_name['FocusDevice'] = _FOCUSDEVICE
DESCRIPTOR.enum_types_by_name['TypingType'] = _TYPINGTYPE
DESCRIPTOR.enum_types_by_name['ClientPresenceStateType'] = _CLIENTPRESENCESTATETYPE
DESCRIPTOR.enum_types_by_name['NotificationLevel'] = _NOTIFICATIONLEVEL
DESCRIPTOR.enum_types_by_name['SegmentType'] = _SEGMENTTYPE
DESCRIPTOR.enum_types_by_name['ItemType'] = _ITEMTYPE
DESCRIPTOR.enum_types_by_name['MembershipChangeType'] = _MEMBERSHIPCHANGETYPE
DESCRIPTOR.enum_types_by_name['HangoutEventType'] = _HANGOUTEVENTTYPE
DESCRIPTOR.enum_types_by_name['OffTheRecordToggle'] = _OFFTHERECORDTOGGLE
DESCRIPTOR.enum_types_by_name['OffTheRecordStatus'] = _OFFTHERECORDSTATUS
DESCRIPTOR.enum_types_by_name['SourceType'] = _SOURCETYPE
DESCRIPTOR.enum_types_by_name['EventType'] = _EVENTTYPE
DESCRIPTOR.enum_types_by_name['ConversationType'] = _CONVERSATIONTYPE
DESCRIPTOR.enum_types_by_name['ConversationStatus'] = _CONVERSATIONSTATUS
DESCRIPTOR.enum_types_by_name['ConversationView'] = _CONVERSATIONVIEW
DESCRIPTOR.enum_types_by_name['DeliveryMediumType'] = _DELIVERYMEDIUMTYPE
DESCRIPTOR.enum_types_by_name['InvitationAffinity'] = _INVITATIONAFFINITY
DESCRIPTOR.enum_types_by_name['ParticipantType'] = _PARTICIPANTTYPE
DESCRIPTOR.enum_types_by_name['InvitationStatus'] = _INVITATIONSTATUS
DESCRIPTOR.enum_types_by_name['ForceHistory'] = _FORCEHISTORY
DESCRIPTOR.enum_types_by_name['NetworkType'] = _NETWORKTYPE
DESCRIPTOR.enum_types_by_name['BlockState'] = _BLOCKSTATE
DESCRIPTOR.enum_types_by_name['ReplyToInviteType'] = _REPLYTOINVITETYPE
DESCRIPTOR.enum_types_by_name['ClientId'] = _CLIENTID
DESCRIPTOR.enum_types_by_name['ClientBuildType'] = _CLIENTBUILDTYPE
DESCRIPTOR.enum_types_by_name['ResponseStatus'] = _RESPONSESTATUS
DESCRIPTOR.enum_types_by_name['PhotoUrlStatus'] = _PHOTOURLSTATUS
DESCRIPTOR.enum_types_by_name['Gender'] = _GENDER
DESCRIPTOR.enum_types_by_name['ProfileType'] = _PROFILETYPE
DESCRIPTOR.enum_types_by_name['ConfigurationBitType'] = _CONFIGURATIONBITTYPE
DESCRIPTOR.enum_types_by_name['RichPresenceType'] = _RICHPRESENCETYPE
DESCRIPTOR.enum_types_by_name['FieldMask'] = _FIELDMASK
DESCRIPTOR.enum_types_by_name['DeleteType'] = _DELETETYPE
DESCRIPTOR.enum_types_by_name['SyncFilter'] = _SYNCFILTER
DESCRIPTOR.enum_types_by_name['SoundState'] = _SOUNDSTATE
DESCRIPTOR.enum_types_by_name['CallerIdSettingsMask'] = _CALLERIDSETTINGSMASK
DESCRIPTOR.enum_types_by_name['PhoneVerificationStatus'] = _PHONEVERIFICATIONSTATUS
DESCRIPTOR.enum_types_by_name['PhoneDiscoverabilityStatus'] = _PHONEDISCOVERABILITYSTATUS
DESCRIPTOR.enum_types_by_name['PhoneValidationResult'] = _PHONEVALIDATIONRESULT
DESCRIPTOR.enum_types_by_name['OffnetworkAddressType'] = _OFFNETWORKADDRESSTYPE
DESCRIPTOR.enum_types_by_name['GroupLinkSharingStatus'] = _GROUPLINKSHARINGSTATUS

DoNotDisturbSetting = _reflection.GeneratedProtocolMessageType('DoNotDisturbSetting', (_message.Message,), dict(
  DESCRIPTOR = _DONOTDISTURBSETTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DoNotDisturbSetting)
  ))
_sym_db.RegisterMessage(DoNotDisturbSetting)

NotificationSettings = _reflection.GeneratedProtocolMessageType('NotificationSettings', (_message.Message,), dict(
  DESCRIPTOR = _NOTIFICATIONSETTINGS,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:NotificationSettings)
  ))
_sym_db.RegisterMessage(NotificationSettings)

ConversationId = _reflection.GeneratedProtocolMessageType('ConversationId', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONID,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationId)
  ))
_sym_db.RegisterMessage(ConversationId)

ParticipantId = _reflection.GeneratedProtocolMessageType('ParticipantId', (_message.Message,), dict(
  DESCRIPTOR = _PARTICIPANTID,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ParticipantId)
  ))
_sym_db.RegisterMessage(ParticipantId)

DeviceStatus = _reflection.GeneratedProtocolMessageType('DeviceStatus', (_message.Message,), dict(
  DESCRIPTOR = _DEVICESTATUS,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeviceStatus)
  ))
_sym_db.RegisterMessage(DeviceStatus)

LastSeen = _reflection.GeneratedProtocolMessageType('LastSeen', (_message.Message,), dict(
  DESCRIPTOR = _LASTSEEN,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:LastSeen)
  ))
_sym_db.RegisterMessage(LastSeen)

Presence = _reflection.GeneratedProtocolMessageType('Presence', (_message.Message,), dict(
  DESCRIPTOR = _PRESENCE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Presence)
  ))
_sym_db.RegisterMessage(Presence)

PresenceResult = _reflection.GeneratedProtocolMessageType('PresenceResult', (_message.Message,), dict(
  DESCRIPTOR = _PRESENCERESULT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:PresenceResult)
  ))
_sym_db.RegisterMessage(PresenceResult)

ClientIdentifier = _reflection.GeneratedProtocolMessageType('ClientIdentifier', (_message.Message,), dict(
  DESCRIPTOR = _CLIENTIDENTIFIER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ClientIdentifier)
  ))
_sym_db.RegisterMessage(ClientIdentifier)

ClientPresenceState = _reflection.GeneratedProtocolMessageType('ClientPresenceState', (_message.Message,), dict(
  DESCRIPTOR = _CLIENTPRESENCESTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ClientPresenceState)
  ))
_sym_db.RegisterMessage(ClientPresenceState)

UserEventState = _reflection.GeneratedProtocolMessageType('UserEventState', (_message.Message,), dict(
  DESCRIPTOR = _USEREVENTSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:UserEventState)
  ))
_sym_db.RegisterMessage(UserEventState)

Formatting = _reflection.GeneratedProtocolMessageType('Formatting', (_message.Message,), dict(
  DESCRIPTOR = _FORMATTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Formatting)
  ))
_sym_db.RegisterMessage(Formatting)

LinkData = _reflection.GeneratedProtocolMessageType('LinkData', (_message.Message,), dict(
  DESCRIPTOR = _LINKDATA,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:LinkData)
  ))
_sym_db.RegisterMessage(LinkData)

Segment = _reflection.GeneratedProtocolMessageType('Segment', (_message.Message,), dict(
  DESCRIPTOR = _SEGMENT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Segment)
  ))
_sym_db.RegisterMessage(Segment)

PlusPhoto = _reflection.GeneratedProtocolMessageType('PlusPhoto', (_message.Message,), dict(

  Thumbnail = _reflection.GeneratedProtocolMessageType('Thumbnail', (_message.Message,), dict(
    DESCRIPTOR = _PLUSPHOTO_THUMBNAIL,
    __module__ = 'hangups.hangouts_pb2'
    # @@protoc_insertion_point(class_scope:PlusPhoto.Thumbnail)
    ))
  ,
  DESCRIPTOR = _PLUSPHOTO,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:PlusPhoto)
  ))
_sym_db.RegisterMessage(PlusPhoto)
_sym_db.RegisterMessage(PlusPhoto.Thumbnail)

Place = _reflection.GeneratedProtocolMessageType('Place', (_message.Message,), dict(
  DESCRIPTOR = _PLACE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Place)
  ))
_sym_db.RegisterMessage(Place)

EmbedItem = _reflection.GeneratedProtocolMessageType('EmbedItem', (_message.Message,), dict(

  PostalAddress = _reflection.GeneratedProtocolMessageType('PostalAddress', (_message.Message,), dict(
    DESCRIPTOR = _EMBEDITEM_POSTALADDRESS,
    __module__ = 'hangups.hangouts_pb2'
    # @@protoc_insertion_point(class_scope:EmbedItem.PostalAddress)
    ))
  ,

  GeoCoordinates = _reflection.GeneratedProtocolMessageType('GeoCoordinates', (_message.Message,), dict(
    DESCRIPTOR = _EMBEDITEM_GEOCOORDINATES,
    __module__ = 'hangups.hangouts_pb2'
    # @@protoc_insertion_point(class_scope:EmbedItem.GeoCoordinates)
    ))
  ,

  Image = _reflection.GeneratedProtocolMessageType('Image', (_message.Message,), dict(
    DESCRIPTOR = _EMBEDITEM_IMAGE,
    __module__ = 'hangups.hangouts_pb2'
    # @@protoc_insertion_point(class_scope:EmbedItem.Image)
    ))
  ,
  DESCRIPTOR = _EMBEDITEM,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EmbedItem)
  ))
_sym_db.RegisterMessage(EmbedItem)
_sym_db.RegisterMessage(EmbedItem.PostalAddress)
_sym_db.RegisterMessage(EmbedItem.GeoCoordinates)
_sym_db.RegisterMessage(EmbedItem.Image)

Attachment = _reflection.GeneratedProtocolMessageType('Attachment', (_message.Message,), dict(
  DESCRIPTOR = _ATTACHMENT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Attachment)
  ))
_sym_db.RegisterMessage(Attachment)

MessageContent = _reflection.GeneratedProtocolMessageType('MessageContent', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGECONTENT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:MessageContent)
  ))
_sym_db.RegisterMessage(MessageContent)

EventAnnotation = _reflection.GeneratedProtocolMessageType('EventAnnotation', (_message.Message,), dict(
  DESCRIPTOR = _EVENTANNOTATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EventAnnotation)
  ))
_sym_db.RegisterMessage(EventAnnotation)

ChatMessage = _reflection.GeneratedProtocolMessageType('ChatMessage', (_message.Message,), dict(
  DESCRIPTOR = _CHATMESSAGE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ChatMessage)
  ))
_sym_db.RegisterMessage(ChatMessage)

MembershipChange = _reflection.GeneratedProtocolMessageType('MembershipChange', (_message.Message,), dict(
  DESCRIPTOR = _MEMBERSHIPCHANGE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:MembershipChange)
  ))
_sym_db.RegisterMessage(MembershipChange)

ConversationRename = _reflection.GeneratedProtocolMessageType('ConversationRename', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONRENAME,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationRename)
  ))
_sym_db.RegisterMessage(ConversationRename)

HangoutEvent = _reflection.GeneratedProtocolMessageType('HangoutEvent', (_message.Message,), dict(
  DESCRIPTOR = _HANGOUTEVENT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:HangoutEvent)
  ))
_sym_db.RegisterMessage(HangoutEvent)

OTRModification = _reflection.GeneratedProtocolMessageType('OTRModification', (_message.Message,), dict(
  DESCRIPTOR = _OTRMODIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:OTRModification)
  ))
_sym_db.RegisterMessage(OTRModification)

HashModifier = _reflection.GeneratedProtocolMessageType('HashModifier', (_message.Message,), dict(
  DESCRIPTOR = _HASHMODIFIER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:HashModifier)
  ))
_sym_db.RegisterMessage(HashModifier)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), dict(
  DESCRIPTOR = _EVENT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Event)
  ))
_sym_db.RegisterMessage(Event)

UserReadState = _reflection.GeneratedProtocolMessageType('UserReadState', (_message.Message,), dict(
  DESCRIPTOR = _USERREADSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:UserReadState)
  ))
_sym_db.RegisterMessage(UserReadState)

DeliveryMedium = _reflection.GeneratedProtocolMessageType('DeliveryMedium', (_message.Message,), dict(
  DESCRIPTOR = _DELIVERYMEDIUM,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeliveryMedium)
  ))
_sym_db.RegisterMessage(DeliveryMedium)

DeliveryMediumOption = _reflection.GeneratedProtocolMessageType('DeliveryMediumOption', (_message.Message,), dict(
  DESCRIPTOR = _DELIVERYMEDIUMOPTION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeliveryMediumOption)
  ))
_sym_db.RegisterMessage(DeliveryMediumOption)

UserConversationState = _reflection.GeneratedProtocolMessageType('UserConversationState', (_message.Message,), dict(
  DESCRIPTOR = _USERCONVERSATIONSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:UserConversationState)
  ))
_sym_db.RegisterMessage(UserConversationState)

ConversationParticipantData = _reflection.GeneratedProtocolMessageType('ConversationParticipantData', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONPARTICIPANTDATA,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationParticipantData)
  ))
_sym_db.RegisterMessage(ConversationParticipantData)

Conversation = _reflection.GeneratedProtocolMessageType('Conversation', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Conversation)
  ))
_sym_db.RegisterMessage(Conversation)

EasterEgg = _reflection.GeneratedProtocolMessageType('EasterEgg', (_message.Message,), dict(
  DESCRIPTOR = _EASTEREGG,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EasterEgg)
  ))
_sym_db.RegisterMessage(EasterEgg)

BlockStateChange = _reflection.GeneratedProtocolMessageType('BlockStateChange', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKSTATECHANGE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:BlockStateChange)
  ))
_sym_db.RegisterMessage(BlockStateChange)

Photo = _reflection.GeneratedProtocolMessageType('Photo', (_message.Message,), dict(
  DESCRIPTOR = _PHOTO,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Photo)
  ))
_sym_db.RegisterMessage(Photo)

ExistingMedia = _reflection.GeneratedProtocolMessageType('ExistingMedia', (_message.Message,), dict(
  DESCRIPTOR = _EXISTINGMEDIA,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ExistingMedia)
  ))
_sym_db.RegisterMessage(ExistingMedia)

EventRequestHeader = _reflection.GeneratedProtocolMessageType('EventRequestHeader', (_message.Message,), dict(
  DESCRIPTOR = _EVENTREQUESTHEADER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EventRequestHeader)
  ))
_sym_db.RegisterMessage(EventRequestHeader)

ClientVersion = _reflection.GeneratedProtocolMessageType('ClientVersion', (_message.Message,), dict(
  DESCRIPTOR = _CLIENTVERSION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ClientVersion)
  ))
_sym_db.RegisterMessage(ClientVersion)

RequestHeader = _reflection.GeneratedProtocolMessageType('RequestHeader', (_message.Message,), dict(
  DESCRIPTOR = _REQUESTHEADER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RequestHeader)
  ))
_sym_db.RegisterMessage(RequestHeader)

ResponseHeader = _reflection.GeneratedProtocolMessageType('ResponseHeader', (_message.Message,), dict(
  DESCRIPTOR = _RESPONSEHEADER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ResponseHeader)
  ))
_sym_db.RegisterMessage(ResponseHeader)

Entity = _reflection.GeneratedProtocolMessageType('Entity', (_message.Message,), dict(
  DESCRIPTOR = _ENTITY,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Entity)
  ))
_sym_db.RegisterMessage(Entity)

EntityProperties = _reflection.GeneratedProtocolMessageType('EntityProperties', (_message.Message,), dict(
  DESCRIPTOR = _ENTITYPROPERTIES,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EntityProperties)
  ))
_sym_db.RegisterMessage(EntityProperties)

ConversationState = _reflection.GeneratedProtocolMessageType('ConversationState', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationState)
  ))
_sym_db.RegisterMessage(ConversationState)

EventContinuationToken = _reflection.GeneratedProtocolMessageType('EventContinuationToken', (_message.Message,), dict(
  DESCRIPTOR = _EVENTCONTINUATIONTOKEN,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EventContinuationToken)
  ))
_sym_db.RegisterMessage(EventContinuationToken)

EntityLookupSpec = _reflection.GeneratedProtocolMessageType('EntityLookupSpec', (_message.Message,), dict(
  DESCRIPTOR = _ENTITYLOOKUPSPEC,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EntityLookupSpec)
  ))
_sym_db.RegisterMessage(EntityLookupSpec)

ConfigurationBit = _reflection.GeneratedProtocolMessageType('ConfigurationBit', (_message.Message,), dict(
  DESCRIPTOR = _CONFIGURATIONBIT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConfigurationBit)
  ))
_sym_db.RegisterMessage(ConfigurationBit)

RichPresenceState = _reflection.GeneratedProtocolMessageType('RichPresenceState', (_message.Message,), dict(
  DESCRIPTOR = _RICHPRESENCESTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RichPresenceState)
  ))
_sym_db.RegisterMessage(RichPresenceState)

RichPresenceEnabledState = _reflection.GeneratedProtocolMessageType('RichPresenceEnabledState', (_message.Message,), dict(
  DESCRIPTOR = _RICHPRESENCEENABLEDSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RichPresenceEnabledState)
  ))
_sym_db.RegisterMessage(RichPresenceEnabledState)

DesktopOffSetting = _reflection.GeneratedProtocolMessageType('DesktopOffSetting', (_message.Message,), dict(
  DESCRIPTOR = _DESKTOPOFFSETTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DesktopOffSetting)
  ))
_sym_db.RegisterMessage(DesktopOffSetting)

DesktopOffState = _reflection.GeneratedProtocolMessageType('DesktopOffState', (_message.Message,), dict(
  DESCRIPTOR = _DESKTOPOFFSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DesktopOffState)
  ))
_sym_db.RegisterMessage(DesktopOffState)

DndSetting = _reflection.GeneratedProtocolMessageType('DndSetting', (_message.Message,), dict(
  DESCRIPTOR = _DNDSETTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DndSetting)
  ))
_sym_db.RegisterMessage(DndSetting)

PresenceStateSetting = _reflection.GeneratedProtocolMessageType('PresenceStateSetting', (_message.Message,), dict(
  DESCRIPTOR = _PRESENCESTATESETTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:PresenceStateSetting)
  ))
_sym_db.RegisterMessage(PresenceStateSetting)

MoodMessage = _reflection.GeneratedProtocolMessageType('MoodMessage', (_message.Message,), dict(
  DESCRIPTOR = _MOODMESSAGE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:MoodMessage)
  ))
_sym_db.RegisterMessage(MoodMessage)

MoodContent = _reflection.GeneratedProtocolMessageType('MoodContent', (_message.Message,), dict(
  DESCRIPTOR = _MOODCONTENT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:MoodContent)
  ))
_sym_db.RegisterMessage(MoodContent)

MoodSetting = _reflection.GeneratedProtocolMessageType('MoodSetting', (_message.Message,), dict(
  DESCRIPTOR = _MOODSETTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:MoodSetting)
  ))
_sym_db.RegisterMessage(MoodSetting)

MoodState = _reflection.GeneratedProtocolMessageType('MoodState', (_message.Message,), dict(
  DESCRIPTOR = _MOODSTATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:MoodState)
  ))
_sym_db.RegisterMessage(MoodState)

DeleteAction = _reflection.GeneratedProtocolMessageType('DeleteAction', (_message.Message,), dict(
  DESCRIPTOR = _DELETEACTION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeleteAction)
  ))
_sym_db.RegisterMessage(DeleteAction)

InviteeID = _reflection.GeneratedProtocolMessageType('InviteeID', (_message.Message,), dict(
  DESCRIPTOR = _INVITEEID,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:InviteeID)
  ))
_sym_db.RegisterMessage(InviteeID)

Country = _reflection.GeneratedProtocolMessageType('Country', (_message.Message,), dict(
  DESCRIPTOR = _COUNTRY,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Country)
  ))
_sym_db.RegisterMessage(Country)

DesktopSoundSetting = _reflection.GeneratedProtocolMessageType('DesktopSoundSetting', (_message.Message,), dict(
  DESCRIPTOR = _DESKTOPSOUNDSETTING,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DesktopSoundSetting)
  ))
_sym_db.RegisterMessage(DesktopSoundSetting)

PhoneData = _reflection.GeneratedProtocolMessageType('PhoneData', (_message.Message,), dict(
  DESCRIPTOR = _PHONEDATA,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:PhoneData)
  ))
_sym_db.RegisterMessage(PhoneData)

Phone = _reflection.GeneratedProtocolMessageType('Phone', (_message.Message,), dict(
  DESCRIPTOR = _PHONE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Phone)
  ))
_sym_db.RegisterMessage(Phone)

I18nData = _reflection.GeneratedProtocolMessageType('I18nData', (_message.Message,), dict(
  DESCRIPTOR = _I18NDATA,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:I18nData)
  ))
_sym_db.RegisterMessage(I18nData)

PhoneNumber = _reflection.GeneratedProtocolMessageType('PhoneNumber', (_message.Message,), dict(
  DESCRIPTOR = _PHONENUMBER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:PhoneNumber)
  ))
_sym_db.RegisterMessage(PhoneNumber)

SuggestedContactGroupHash = _reflection.GeneratedProtocolMessageType('SuggestedContactGroupHash', (_message.Message,), dict(
  DESCRIPTOR = _SUGGESTEDCONTACTGROUPHASH,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SuggestedContactGroupHash)
  ))
_sym_db.RegisterMessage(SuggestedContactGroupHash)

SuggestedContact = _reflection.GeneratedProtocolMessageType('SuggestedContact', (_message.Message,), dict(
  DESCRIPTOR = _SUGGESTEDCONTACT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SuggestedContact)
  ))
_sym_db.RegisterMessage(SuggestedContact)

SuggestedContactGroup = _reflection.GeneratedProtocolMessageType('SuggestedContactGroup', (_message.Message,), dict(
  DESCRIPTOR = _SUGGESTEDCONTACTGROUP,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SuggestedContactGroup)
  ))
_sym_db.RegisterMessage(SuggestedContactGroup)

GroupLinkSharingModification = _reflection.GeneratedProtocolMessageType('GroupLinkSharingModification', (_message.Message,), dict(
  DESCRIPTOR = _GROUPLINKSHARINGMODIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GroupLinkSharingModification)
  ))
_sym_db.RegisterMessage(GroupLinkSharingModification)

StateUpdate = _reflection.GeneratedProtocolMessageType('StateUpdate', (_message.Message,), dict(
  DESCRIPTOR = _STATEUPDATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:StateUpdate)
  ))
_sym_db.RegisterMessage(StateUpdate)

StateUpdateHeader = _reflection.GeneratedProtocolMessageType('StateUpdateHeader', (_message.Message,), dict(
  DESCRIPTOR = _STATEUPDATEHEADER,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:StateUpdateHeader)
  ))
_sym_db.RegisterMessage(StateUpdateHeader)

BatchUpdate = _reflection.GeneratedProtocolMessageType('BatchUpdate', (_message.Message,), dict(
  DESCRIPTOR = _BATCHUPDATE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:BatchUpdate)
  ))
_sym_db.RegisterMessage(BatchUpdate)

ConversationNotification = _reflection.GeneratedProtocolMessageType('ConversationNotification', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationNotification)
  ))
_sym_db.RegisterMessage(ConversationNotification)

EventNotification = _reflection.GeneratedProtocolMessageType('EventNotification', (_message.Message,), dict(
  DESCRIPTOR = _EVENTNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EventNotification)
  ))
_sym_db.RegisterMessage(EventNotification)

SetFocusNotification = _reflection.GeneratedProtocolMessageType('SetFocusNotification', (_message.Message,), dict(
  DESCRIPTOR = _SETFOCUSNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetFocusNotification)
  ))
_sym_db.RegisterMessage(SetFocusNotification)

SetTypingNotification = _reflection.GeneratedProtocolMessageType('SetTypingNotification', (_message.Message,), dict(
  DESCRIPTOR = _SETTYPINGNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetTypingNotification)
  ))
_sym_db.RegisterMessage(SetTypingNotification)

SetConversationNotificationLevelNotification = _reflection.GeneratedProtocolMessageType('SetConversationNotificationLevelNotification', (_message.Message,), dict(
  DESCRIPTOR = _SETCONVERSATIONNOTIFICATIONLEVELNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetConversationNotificationLevelNotification)
  ))
_sym_db.RegisterMessage(SetConversationNotificationLevelNotification)

ReplyToInviteNotification = _reflection.GeneratedProtocolMessageType('ReplyToInviteNotification', (_message.Message,), dict(
  DESCRIPTOR = _REPLYTOINVITENOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ReplyToInviteNotification)
  ))
_sym_db.RegisterMessage(ReplyToInviteNotification)

WatermarkNotification = _reflection.GeneratedProtocolMessageType('WatermarkNotification', (_message.Message,), dict(
  DESCRIPTOR = _WATERMARKNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:WatermarkNotification)
  ))
_sym_db.RegisterMessage(WatermarkNotification)

ConversationViewModification = _reflection.GeneratedProtocolMessageType('ConversationViewModification', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONVIEWMODIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationViewModification)
  ))
_sym_db.RegisterMessage(ConversationViewModification)

EasterEggNotification = _reflection.GeneratedProtocolMessageType('EasterEggNotification', (_message.Message,), dict(
  DESCRIPTOR = _EASTEREGGNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EasterEggNotification)
  ))
_sym_db.RegisterMessage(EasterEggNotification)

SelfPresenceNotification = _reflection.GeneratedProtocolMessageType('SelfPresenceNotification', (_message.Message,), dict(
  DESCRIPTOR = _SELFPRESENCENOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SelfPresenceNotification)
  ))
_sym_db.RegisterMessage(SelfPresenceNotification)

DeleteActionNotification = _reflection.GeneratedProtocolMessageType('DeleteActionNotification', (_message.Message,), dict(
  DESCRIPTOR = _DELETEACTIONNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeleteActionNotification)
  ))
_sym_db.RegisterMessage(DeleteActionNotification)

PresenceNotification = _reflection.GeneratedProtocolMessageType('PresenceNotification', (_message.Message,), dict(
  DESCRIPTOR = _PRESENCENOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:PresenceNotification)
  ))
_sym_db.RegisterMessage(PresenceNotification)

BlockNotification = _reflection.GeneratedProtocolMessageType('BlockNotification', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:BlockNotification)
  ))
_sym_db.RegisterMessage(BlockNotification)

SetNotificationSettingNotification = _reflection.GeneratedProtocolMessageType('SetNotificationSettingNotification', (_message.Message,), dict(
  DESCRIPTOR = _SETNOTIFICATIONSETTINGNOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetNotificationSettingNotification)
  ))
_sym_db.RegisterMessage(SetNotificationSettingNotification)

RichPresenceEnabledStateNotification = _reflection.GeneratedProtocolMessageType('RichPresenceEnabledStateNotification', (_message.Message,), dict(
  DESCRIPTOR = _RICHPRESENCEENABLEDSTATENOTIFICATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RichPresenceEnabledStateNotification)
  ))
_sym_db.RegisterMessage(RichPresenceEnabledStateNotification)

ConversationSpec = _reflection.GeneratedProtocolMessageType('ConversationSpec', (_message.Message,), dict(
  DESCRIPTOR = _CONVERSATIONSPEC,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ConversationSpec)
  ))
_sym_db.RegisterMessage(ConversationSpec)

OffnetworkAddress = _reflection.GeneratedProtocolMessageType('OffnetworkAddress', (_message.Message,), dict(
  DESCRIPTOR = _OFFNETWORKADDRESS,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:OffnetworkAddress)
  ))
_sym_db.RegisterMessage(OffnetworkAddress)

EntityResult = _reflection.GeneratedProtocolMessageType('EntityResult', (_message.Message,), dict(
  DESCRIPTOR = _ENTITYRESULT,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EntityResult)
  ))
_sym_db.RegisterMessage(EntityResult)

AddUserRequest = _reflection.GeneratedProtocolMessageType('AddUserRequest', (_message.Message,), dict(
  DESCRIPTOR = _ADDUSERREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:AddUserRequest)
  ))
_sym_db.RegisterMessage(AddUserRequest)

AddUserResponse = _reflection.GeneratedProtocolMessageType('AddUserResponse', (_message.Message,), dict(
  DESCRIPTOR = _ADDUSERRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:AddUserResponse)
  ))
_sym_db.RegisterMessage(AddUserResponse)

CreateConversationRequest = _reflection.GeneratedProtocolMessageType('CreateConversationRequest', (_message.Message,), dict(
  DESCRIPTOR = _CREATECONVERSATIONREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:CreateConversationRequest)
  ))
_sym_db.RegisterMessage(CreateConversationRequest)

CreateConversationResponse = _reflection.GeneratedProtocolMessageType('CreateConversationResponse', (_message.Message,), dict(
  DESCRIPTOR = _CREATECONVERSATIONRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:CreateConversationResponse)
  ))
_sym_db.RegisterMessage(CreateConversationResponse)

DeleteConversationRequest = _reflection.GeneratedProtocolMessageType('DeleteConversationRequest', (_message.Message,), dict(
  DESCRIPTOR = _DELETECONVERSATIONREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeleteConversationRequest)
  ))
_sym_db.RegisterMessage(DeleteConversationRequest)

DeleteConversationResponse = _reflection.GeneratedProtocolMessageType('DeleteConversationResponse', (_message.Message,), dict(
  DESCRIPTOR = _DELETECONVERSATIONRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:DeleteConversationResponse)
  ))
_sym_db.RegisterMessage(DeleteConversationResponse)

EasterEggRequest = _reflection.GeneratedProtocolMessageType('EasterEggRequest', (_message.Message,), dict(
  DESCRIPTOR = _EASTEREGGREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EasterEggRequest)
  ))
_sym_db.RegisterMessage(EasterEggRequest)

EasterEggResponse = _reflection.GeneratedProtocolMessageType('EasterEggResponse', (_message.Message,), dict(
  DESCRIPTOR = _EASTEREGGRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:EasterEggResponse)
  ))
_sym_db.RegisterMessage(EasterEggResponse)

GetConversationRequest = _reflection.GeneratedProtocolMessageType('GetConversationRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETCONVERSATIONREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetConversationRequest)
  ))
_sym_db.RegisterMessage(GetConversationRequest)

GetConversationResponse = _reflection.GeneratedProtocolMessageType('GetConversationResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETCONVERSATIONRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetConversationResponse)
  ))
_sym_db.RegisterMessage(GetConversationResponse)

GetEntityByIdRequest = _reflection.GeneratedProtocolMessageType('GetEntityByIdRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETENTITYBYIDREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetEntityByIdRequest)
  ))
_sym_db.RegisterMessage(GetEntityByIdRequest)

GetEntityByIdResponse = _reflection.GeneratedProtocolMessageType('GetEntityByIdResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETENTITYBYIDRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetEntityByIdResponse)
  ))
_sym_db.RegisterMessage(GetEntityByIdResponse)

GetGroupConversationUrlRequest = _reflection.GeneratedProtocolMessageType('GetGroupConversationUrlRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETGROUPCONVERSATIONURLREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetGroupConversationUrlRequest)
  ))
_sym_db.RegisterMessage(GetGroupConversationUrlRequest)

GetGroupConversationUrlResponse = _reflection.GeneratedProtocolMessageType('GetGroupConversationUrlResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETGROUPCONVERSATIONURLRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetGroupConversationUrlResponse)
  ))
_sym_db.RegisterMessage(GetGroupConversationUrlResponse)

GetSuggestedEntitiesRequest = _reflection.GeneratedProtocolMessageType('GetSuggestedEntitiesRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETSUGGESTEDENTITIESREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetSuggestedEntitiesRequest)
  ))
_sym_db.RegisterMessage(GetSuggestedEntitiesRequest)

GetSuggestedEntitiesResponse = _reflection.GeneratedProtocolMessageType('GetSuggestedEntitiesResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETSUGGESTEDENTITIESRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetSuggestedEntitiesResponse)
  ))
_sym_db.RegisterMessage(GetSuggestedEntitiesResponse)

GetSelfInfoRequest = _reflection.GeneratedProtocolMessageType('GetSelfInfoRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETSELFINFOREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetSelfInfoRequest)
  ))
_sym_db.RegisterMessage(GetSelfInfoRequest)

GetSelfInfoResponse = _reflection.GeneratedProtocolMessageType('GetSelfInfoResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETSELFINFORESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:GetSelfInfoResponse)
  ))
_sym_db.RegisterMessage(GetSelfInfoResponse)

QueryPresenceRequest = _reflection.GeneratedProtocolMessageType('QueryPresenceRequest', (_message.Message,), dict(
  DESCRIPTOR = _QUERYPRESENCEREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:QueryPresenceRequest)
  ))
_sym_db.RegisterMessage(QueryPresenceRequest)

QueryPresenceResponse = _reflection.GeneratedProtocolMessageType('QueryPresenceResponse', (_message.Message,), dict(
  DESCRIPTOR = _QUERYPRESENCERESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:QueryPresenceResponse)
  ))
_sym_db.RegisterMessage(QueryPresenceResponse)

RemoveUserRequest = _reflection.GeneratedProtocolMessageType('RemoveUserRequest', (_message.Message,), dict(
  DESCRIPTOR = _REMOVEUSERREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RemoveUserRequest)
  ))
_sym_db.RegisterMessage(RemoveUserRequest)

RemoveUserResponse = _reflection.GeneratedProtocolMessageType('RemoveUserResponse', (_message.Message,), dict(
  DESCRIPTOR = _REMOVEUSERRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RemoveUserResponse)
  ))
_sym_db.RegisterMessage(RemoveUserResponse)

RenameConversationRequest = _reflection.GeneratedProtocolMessageType('RenameConversationRequest', (_message.Message,), dict(
  DESCRIPTOR = _RENAMECONVERSATIONREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RenameConversationRequest)
  ))
_sym_db.RegisterMessage(RenameConversationRequest)

RenameConversationResponse = _reflection.GeneratedProtocolMessageType('RenameConversationResponse', (_message.Message,), dict(
  DESCRIPTOR = _RENAMECONVERSATIONRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:RenameConversationResponse)
  ))
_sym_db.RegisterMessage(RenameConversationResponse)

SearchEntitiesRequest = _reflection.GeneratedProtocolMessageType('SearchEntitiesRequest', (_message.Message,), dict(
  DESCRIPTOR = _SEARCHENTITIESREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SearchEntitiesRequest)
  ))
_sym_db.RegisterMessage(SearchEntitiesRequest)

SearchEntitiesResponse = _reflection.GeneratedProtocolMessageType('SearchEntitiesResponse', (_message.Message,), dict(
  DESCRIPTOR = _SEARCHENTITIESRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SearchEntitiesResponse)
  ))
_sym_db.RegisterMessage(SearchEntitiesResponse)

Location = _reflection.GeneratedProtocolMessageType('Location', (_message.Message,), dict(
  DESCRIPTOR = _LOCATION,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:Location)
  ))
_sym_db.RegisterMessage(Location)

SendChatMessageRequest = _reflection.GeneratedProtocolMessageType('SendChatMessageRequest', (_message.Message,), dict(
  DESCRIPTOR = _SENDCHATMESSAGEREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SendChatMessageRequest)
  ))
_sym_db.RegisterMessage(SendChatMessageRequest)

SendChatMessageResponse = _reflection.GeneratedProtocolMessageType('SendChatMessageResponse', (_message.Message,), dict(
  DESCRIPTOR = _SENDCHATMESSAGERESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SendChatMessageResponse)
  ))
_sym_db.RegisterMessage(SendChatMessageResponse)

ModifyOTRStatusRequest = _reflection.GeneratedProtocolMessageType('ModifyOTRStatusRequest', (_message.Message,), dict(
  DESCRIPTOR = _MODIFYOTRSTATUSREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ModifyOTRStatusRequest)
  ))
_sym_db.RegisterMessage(ModifyOTRStatusRequest)

ModifyOTRStatusResponse = _reflection.GeneratedProtocolMessageType('ModifyOTRStatusResponse', (_message.Message,), dict(
  DESCRIPTOR = _MODIFYOTRSTATUSRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:ModifyOTRStatusResponse)
  ))
_sym_db.RegisterMessage(ModifyOTRStatusResponse)

SendOffnetworkInvitationRequest = _reflection.GeneratedProtocolMessageType('SendOffnetworkInvitationRequest', (_message.Message,), dict(
  DESCRIPTOR = _SENDOFFNETWORKINVITATIONREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SendOffnetworkInvitationRequest)
  ))
_sym_db.RegisterMessage(SendOffnetworkInvitationRequest)

SendOffnetworkInvitationResponse = _reflection.GeneratedProtocolMessageType('SendOffnetworkInvitationResponse', (_message.Message,), dict(
  DESCRIPTOR = _SENDOFFNETWORKINVITATIONRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SendOffnetworkInvitationResponse)
  ))
_sym_db.RegisterMessage(SendOffnetworkInvitationResponse)

SetActiveClientRequest = _reflection.GeneratedProtocolMessageType('SetActiveClientRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETACTIVECLIENTREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetActiveClientRequest)
  ))
_sym_db.RegisterMessage(SetActiveClientRequest)

SetActiveClientResponse = _reflection.GeneratedProtocolMessageType('SetActiveClientResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETACTIVECLIENTRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetActiveClientResponse)
  ))
_sym_db.RegisterMessage(SetActiveClientResponse)

SetConversationLevelRequest = _reflection.GeneratedProtocolMessageType('SetConversationLevelRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETCONVERSATIONLEVELREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetConversationLevelRequest)
  ))
_sym_db.RegisterMessage(SetConversationLevelRequest)

SetConversationLevelResponse = _reflection.GeneratedProtocolMessageType('SetConversationLevelResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETCONVERSATIONLEVELRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetConversationLevelResponse)
  ))
_sym_db.RegisterMessage(SetConversationLevelResponse)

SetConversationNotificationLevelRequest = _reflection.GeneratedProtocolMessageType('SetConversationNotificationLevelRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETCONVERSATIONNOTIFICATIONLEVELREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetConversationNotificationLevelRequest)
  ))
_sym_db.RegisterMessage(SetConversationNotificationLevelRequest)

SetConversationNotificationLevelResponse = _reflection.GeneratedProtocolMessageType('SetConversationNotificationLevelResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETCONVERSATIONNOTIFICATIONLEVELRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetConversationNotificationLevelResponse)
  ))
_sym_db.RegisterMessage(SetConversationNotificationLevelResponse)

SetFocusRequest = _reflection.GeneratedProtocolMessageType('SetFocusRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETFOCUSREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetFocusRequest)
  ))
_sym_db.RegisterMessage(SetFocusRequest)

SetFocusResponse = _reflection.GeneratedProtocolMessageType('SetFocusResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETFOCUSRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetFocusResponse)
  ))
_sym_db.RegisterMessage(SetFocusResponse)

SetGroupLinkSharingEnabledRequest = _reflection.GeneratedProtocolMessageType('SetGroupLinkSharingEnabledRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETGROUPLINKSHARINGENABLEDREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetGroupLinkSharingEnabledRequest)
  ))
_sym_db.RegisterMessage(SetGroupLinkSharingEnabledRequest)

SetGroupLinkSharingEnabledResponse = _reflection.GeneratedProtocolMessageType('SetGroupLinkSharingEnabledResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETGROUPLINKSHARINGENABLEDRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetGroupLinkSharingEnabledResponse)
  ))
_sym_db.RegisterMessage(SetGroupLinkSharingEnabledResponse)

SetPresenceRequest = _reflection.GeneratedProtocolMessageType('SetPresenceRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETPRESENCEREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetPresenceRequest)
  ))
_sym_db.RegisterMessage(SetPresenceRequest)

SetPresenceResponse = _reflection.GeneratedProtocolMessageType('SetPresenceResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETPRESENCERESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetPresenceResponse)
  ))
_sym_db.RegisterMessage(SetPresenceResponse)

SetTypingRequest = _reflection.GeneratedProtocolMessageType('SetTypingRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETTYPINGREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetTypingRequest)
  ))
_sym_db.RegisterMessage(SetTypingRequest)

SetTypingResponse = _reflection.GeneratedProtocolMessageType('SetTypingResponse', (_message.Message,), dict(
  DESCRIPTOR = _SETTYPINGRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SetTypingResponse)
  ))
_sym_db.RegisterMessage(SetTypingResponse)

SyncAllNewEventsRequest = _reflection.GeneratedProtocolMessageType('SyncAllNewEventsRequest', (_message.Message,), dict(
  DESCRIPTOR = _SYNCALLNEWEVENTSREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SyncAllNewEventsRequest)
  ))
_sym_db.RegisterMessage(SyncAllNewEventsRequest)

SyncAllNewEventsResponse = _reflection.GeneratedProtocolMessageType('SyncAllNewEventsResponse', (_message.Message,), dict(
  DESCRIPTOR = _SYNCALLNEWEVENTSRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SyncAllNewEventsResponse)
  ))
_sym_db.RegisterMessage(SyncAllNewEventsResponse)

SyncRecentConversationsRequest = _reflection.GeneratedProtocolMessageType('SyncRecentConversationsRequest', (_message.Message,), dict(
  DESCRIPTOR = _SYNCRECENTCONVERSATIONSREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SyncRecentConversationsRequest)
  ))
_sym_db.RegisterMessage(SyncRecentConversationsRequest)

SyncRecentConversationsResponse = _reflection.GeneratedProtocolMessageType('SyncRecentConversationsResponse', (_message.Message,), dict(
  DESCRIPTOR = _SYNCRECENTCONVERSATIONSRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:SyncRecentConversationsResponse)
  ))
_sym_db.RegisterMessage(SyncRecentConversationsResponse)

UpdateWatermarkRequest = _reflection.GeneratedProtocolMessageType('UpdateWatermarkRequest', (_message.Message,), dict(
  DESCRIPTOR = _UPDATEWATERMARKREQUEST,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:UpdateWatermarkRequest)
  ))
_sym_db.RegisterMessage(UpdateWatermarkRequest)

UpdateWatermarkResponse = _reflection.GeneratedProtocolMessageType('UpdateWatermarkResponse', (_message.Message,), dict(
  DESCRIPTOR = _UPDATEWATERMARKRESPONSE,
  __module__ = 'hangups.hangouts_pb2'
  # @@protoc_insertion_point(class_scope:UpdateWatermarkResponse)
  ))
_sym_db.RegisterMessage(UpdateWatermarkResponse)


# @@protoc_insertion_point(module_scope)
